// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include "nvgl/nvgl.h"
#include "nvutil/RCObject.h"
#include "nvsg/Texture.h"
#include <map>
#include <string>
#include <algorithm>
#include <utility>

namespace nvgl
{
class GLFunctions;
class GLObjects;

// common device context helper function for Windows and Linux
#if defined(_WIN32)
# define MAKE_CURRENT(DC,RC) WGLMakeCurrent(DC,RC)
# define GET_CURRENT_DC()    WGLGetCurrentDC()
# define GET_CURRENT_RC()    WGLGetCurrentContext()
#elif defined(LINUX)
# define MAKE_CURRENT(DC,RC) GLXMakeCurrent(glContext->getDisplay(),DC,RC)
# define GET_CURRENT_RC()    GLXGetCurrentContext()
# define GET_CURRENT_DC()    GLXGetCurrentDrawable()
#else
# error Undefined Operating System!  
#endif

//! Wrapper for OpenGL functions to provide error handling, caching, and extension handling per context.
class GLContext
{
  friend NVSG_API void GLAccum( GLenum op, GLfloat value );
  friend NVSG_API void GLActiveTexture( GLenum texture );
  friend NVSG_API void GLAlphaFunc( GLenum func, GLclampf ref );
  friend NVSG_API GLboolean GLAreTexturesResident( GLsizei n, const GLuint *textures, GLboolean *residences );
  friend NVSG_API void GLArrayElement( GLint i );
  friend NVSG_API void GLBegin( GLenum mode );
  friend NVSG_API void GLBeginOcclusionQuery( GLuint id );
  friend NVSG_API void GLBindBufferARB(GLenum target, GLuint id);
  friend NVSG_API void GLBufferDataARB( GLenum target, GLsizeiptrARB size, const GLvoid *data, GLenum usage );
  friend NVSG_API void GLBufferSubDataARB( GLenum target, GLintptrARB offset, GLsizeiptrARB size, const GLvoid *data );
  friend NVSG_API void GLDeleteBuffersARB( GLsizei n, const GLuint *buffers );
  friend NVSG_API void GLGenBuffersARB( GLsizei n, GLuint *buffers );
  friend NVSG_API GLvoid* GLMapBufferARB( GLenum target, GLenum access );
  friend NVSG_API GLboolean GLUnmapBufferARB( GLenum target );
  friend NVSG_API GLboolean GLIsBufferARB( GLuint name );
  friend NVSG_API void GLBindProgram( GLenum target, GLuint id );
  friend NVSG_API void GLBitmap( GLsizei width, GLsizei height, GLfloat xorig, GLfloat yorig, GLfloat xmove, GLfloat ymove, const GLubyte *bitmap );
  friend NVSG_API void GLBlendFunc( GLenum sfactor, GLenum dfactor );
  friend NVSG_API void GLBindTexture( GLenum target, GLuint texture );
  friend NVSG_API void GLCallList( GLuint list );
  friend NVSG_API void GLCallLists( GLsizei n, GLenum type, const GLvoid *lists );
  friend NVSG_API void GLClear( GLbitfield mask );
  friend NVSG_API void GLClearAccum( GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha );
  friend NVSG_API void GLClearColor( GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha );
  friend NVSG_API void GLClearDepth( GLclampd depth );
  friend NVSG_API void GLClearIndex( GLfloat c );
  friend NVSG_API void GLClearStencil( GLint s );
  friend NVSG_API void GLClientActiveTexture( GLenum texture );
  friend NVSG_API void GLClipPlane( GLenum plane, const GLdouble *equation );
  friend NVSG_API void GLColor3b( GLbyte red, GLbyte green, GLbyte blue );
  friend NVSG_API void GLColor3d( GLdouble red, GLdouble green, GLdouble blue );
  friend NVSG_API void GLColor3f( GLfloat red, GLfloat green, GLfloat blue );
  friend NVSG_API void GLColor3fv( const GLfloat * v );
  friend NVSG_API void GLColor3i( GLint red, GLint green, GLint blue );
  friend NVSG_API void GLColor3s( GLshort red, GLshort green, GLshort blue );
  friend NVSG_API void GLColor3ub( GLubyte red, GLubyte green, GLubyte blue );
  friend NVSG_API void GLColor3ui( GLuint red, GLuint green, GLuint blue );
  friend NVSG_API void GLColor3us( GLushort red, GLushort green, GLushort blue );
  friend NVSG_API void GLColor4b( GLbyte red, GLbyte green, GLbyte blue, GLbyte alpha );
  friend NVSG_API void GLColor4d( GLdouble red, GLdouble green, GLdouble blue, GLdouble alpha );
  friend NVSG_API void GLColor4f( GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha );
  friend NVSG_API void GLColor4i( GLint red, GLint green, GLint blue, GLint alpha );
  friend NVSG_API void GLColor4s( GLshort red, GLshort green, GLshort blue, GLshort alpha );
  friend NVSG_API void GLColor4ub( GLubyte red, GLubyte green, GLubyte blue, GLubyte alpha );
  friend NVSG_API void GLColor4ui( GLuint red, GLuint green, GLuint blue, GLuint alpha );
  friend NVSG_API void GLColor4us( GLushort red, GLushort green, GLushort blue, GLushort alpha );
  friend NVSG_API void GLColorMask( GLboolean red, GLboolean green, GLboolean blue, GLboolean alpha );
  friend NVSG_API void GLColorMaterial( GLenum face, GLenum mode );
  friend NVSG_API void GLColorPointer( GLint size, GLenum type, GLsizei stride, const GLvoid *pointer );
  friend NVSG_API void GLCompressedTexImage1D( GLenum target, int level, GLenum internalformat, GLsizei width, int border, GLsizei imageSize, const void *data );
  friend NVSG_API void GLCompressedTexImage2D( GLenum target, int level, GLenum internalformat, GLsizei width, GLsizei height, int border, GLsizei imageSize, const void *data );
  friend NVSG_API void GLCompressedTexImage3D( GLenum target, int level, GLenum internalformat, GLsizei width, GLsizei height, GLsizei depth, int border, GLsizei imageSize, const void *data );
  friend NVSG_API void GLCopyPixels( GLint x, GLint y, GLsizei width, GLsizei height, GLenum type );
  friend NVSG_API void GLCopyTexImage1D( GLenum target, GLint level, GLenum internalFormat, GLint x, GLint y, GLsizei width, GLint border );
  friend NVSG_API void GLCopyTexImage2D( GLenum target, GLint level, GLenum internalFormat, GLint x, GLint y, GLsizei width, GLsizei height, GLint border );
  friend NVSG_API void GLCopyTexSubImage1D( GLenum target, GLint level, GLint xoffset, GLint x, GLint y, GLsizei width );
  friend NVSG_API void GLCopyTexSubImage2D( GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint x, GLint y, GLsizei width, GLsizei height );
  friend NVSG_API void GLCullFace( GLenum mode );
  friend NVSG_API void GLDeleteLists( GLuint list, GLsizei range );
  friend NVSG_API void GLDeleteOcclusionQueries( GLsizei n, const GLuint *ids );
  friend NVSG_API void GLDeletePrograms( GLsizei n, const GLuint *ids );
  friend NVSG_API void GLDeleteQueries( GLsizei n, const GLuint *ids );
  friend NVSG_API void GLDeleteTextures( GLsizei n, const GLuint *textures );
  friend NVSG_API void GLDepthFunc( GLenum func );
  friend NVSG_API void GLDepthMask( GLboolean flag );
  friend NVSG_API void GLDepthRange( GLclampd znear, GLclampd zfar );
  friend NVSG_API void GLDisable( GLenum cap );
  friend NVSG_API void GLDisableClientState( GLenum array );
  friend NVSG_API void GLDrawArrays( GLenum mode, GLint first, GLsizei count );
  friend NVSG_API void GLDrawBuffer( GLenum mode );
  friend NVSG_API void GLDrawElements( GLenum mode, GLsizei count, GLenum type, const GLvoid *indices );
  friend NVSG_API void GLMultiDrawElements( GLenum mode, const GLsizei *count, GLenum type, const GLvoid* *indices, GLsizei primcount );
  friend NVSG_API void GLDrawPixels( GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels );
  friend NVSG_API void GLDrawRangeElements( GLenum mode, GLuint start, GLuint end, GLsizei count, GLenum type, const GLvoid *indices );
  friend NVSG_API void GLEdgeFlagPointer( GLsizei stride, const GLboolean *pointer );
  friend NVSG_API void GLEdgeFlag( GLboolean flag );
  friend NVSG_API void GLEdgeFlagv( const GLboolean *flag );
  friend NVSG_API void GLEnable( GLenum cap );
  friend NVSG_API void GLEnableClientState( GLenum array );
  friend NVSG_API void GLEnd( void );
  friend NVSG_API void GLEndList( void );
  friend NVSG_API void GLEndOcclusionQuery( void );
  friend NVSG_API void GLEvalCoord1d( GLdouble u );
  friend NVSG_API void GLEvalCoord1f( GLfloat u );
  friend NVSG_API void GLEvalCoord2d( GLdouble u, GLdouble v );
  friend NVSG_API void GLEvalCoord2f( GLfloat u, GLfloat v );
  friend NVSG_API void GLEvalMesh1( GLenum mode, GLint i1, GLint i2 );
  friend NVSG_API void GLEvalMesh2( GLenum mode, GLint i1, GLint i2, GLint j1, GLint j2 );
  friend NVSG_API void GLEvalPoint1( GLint i );
  friend NVSG_API void GLEvalPoint2( GLint i, GLint j );
  friend NVSG_API void GLFeedbackBuffer( GLsizei size, GLenum type, GLfloat *buffer );
  friend NVSG_API void GLFinish( void );
  friend NVSG_API void GLFlush( void );
  friend NVSG_API void GLFogCoordPointer( GLenum type, GLsizei stride, const GLvoid *pointer );
  friend NVSG_API void GLFogf( GLenum pname, GLfloat param );
  friend NVSG_API void GLFogfv( GLenum pname, const GLfloat *params );
  friend NVSG_API void GLFogi( GLenum pname, GLint param );
  friend NVSG_API void GLFogiv( GLenum pname, const GLint *params );
  friend NVSG_API void GLFrontFace( GLenum mode );
  friend NVSG_API void GLFrustum( GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble zNear, GLdouble zFar );
  friend NVSG_API GLuint GLGenLists( GLsizei range );
  friend NVSG_API void GLGenOcclusionQueries( GLsizei n, GLuint *ids );
  friend NVSG_API void GLGenPrograms( GLsizei n, GLuint *ids );
  friend NVSG_API void GLGenQueries( GLsizei n, GLuint *ids );
  friend NVSG_API void GLGenTextures( GLsizei n, GLuint *textures );
  friend NVSG_API void GLGetBooleanv( GLenum pname, GLboolean *params );
  friend NVSG_API void GLGetDoublev( GLenum pname, GLdouble *params );
  friend NVSG_API void GLGetFloatv( GLenum pname, GLfloat *params );
  friend NVSG_API void GLGetIntegerv( GLenum pname, GLint *params );
  friend NVSG_API void GLGetClipPlane( GLenum plane, GLdouble *equation );
  friend NVSG_API void GLGetLightfv( GLenum light, GLenum pname, GLfloat *params );
  friend NVSG_API void GLGetLightiv( GLenum light, GLenum pname, GLint *params );
  friend NVSG_API void GLGetMapdv( GLenum target, GLenum query, GLdouble *v );
  friend NVSG_API void GLGetMapfv( GLenum target, GLenum query, GLfloat *v );
  friend NVSG_API void GLGetMapiv( GLenum target, GLenum query, GLint *v );
  friend NVSG_API void GLGetMaterialfv( GLenum face, GLenum pname, GLfloat *params );
  friend NVSG_API void GLGetMaterialiv( GLenum face, GLenum pname, GLint *params );
  friend NVSG_API void GLGetOcclusionQueryuiv( GLuint id, GLenum pname, GLuint *params );
  friend NVSG_API void GLGetPixelMapfv( GLenum map, GLfloat *values );
  friend NVSG_API void GLGetPixelMapuiv( GLenum map, GLuint *values );
  friend NVSG_API void GLGetPixelMapusv( GLenum map, GLushort *values );
  friend NVSG_API void GLGetPointerv( GLenum pname, GLvoid* *params );
  friend NVSG_API void GLGetPolygonStipple( GLubyte *mask );
  friend NVSG_API const GLubyte * GLGetString( GLenum name );
  friend NVSG_API void GLGetTexEnvfv( GLenum target, GLenum pname, GLfloat *params );
  friend NVSG_API void GLGetTexEnviv( GLenum target, GLenum pname, GLint *params );
  friend NVSG_API void GLGetTexGendv( GLenum coord, GLenum pname, GLdouble *params );
  friend NVSG_API void GLGetTexGenfv( GLenum coord, GLenum pname, GLfloat *params );
  friend NVSG_API void GLGetTexGeniv( GLenum coord, GLenum pname, GLint *params );
  friend NVSG_API void GLGetTexImage( GLenum target, GLint level, GLenum format, GLenum type, GLvoid *pixels );
  friend NVSG_API void GLGetTexLevelParameterfv( GLenum target, GLint level, GLenum pname, GLfloat *params );
  friend NVSG_API void GLGetTexLevelParameteriv( GLenum target, GLint level, GLenum pname, GLint *params );
  friend NVSG_API void GLGetTexParameterfv( GLenum target, GLenum pname, GLfloat *params );
  friend NVSG_API void GLGetTexParameteriv( GLenum target, GLenum pname, GLint *params );
  friend NVSG_API void GLHint( GLenum target, GLenum mode );
  friend NVSG_API void GLIndexd( GLdouble c );
  friend NVSG_API void GLIndexf( GLfloat c );
  friend NVSG_API void GLIndexi( GLint c );
  friend NVSG_API void GLIndexMask( GLuint mask );
  friend NVSG_API void GLIndexPointer( GLenum type, GLsizei stride, const GLvoid *pointer );
  friend NVSG_API void GLIndexs( GLshort c );
  friend NVSG_API void GLIndexub( GLubyte c );
  friend NVSG_API void GLInitNames( void );
  friend NVSG_API void GLInterleavedArrays( GLenum format, GLsizei stride, const GLvoid *pointer );
  friend NVSG_API GLboolean GLIsEnabled( GLenum cap );
  friend NVSG_API GLboolean GLIsList( GLuint list );
  friend NVSG_API GLboolean GLIsOcclusionQuery( GLuint id );
  friend NVSG_API GLboolean GLIsProgram( GLuint id );
  friend NVSG_API GLboolean GLIsQuery( GLuint id );
  friend NVSG_API GLboolean GLIsTexture( GLuint texture );
  friend NVSG_API void GLLightf( GLenum light, GLenum pname, GLfloat param );
  friend NVSG_API void GLLightfv( GLenum light, GLenum pname, const GLfloat *params );
  friend NVSG_API void GLLighti( GLenum light, GLenum pname, GLint param );
  friend NVSG_API void GLLightiv( GLenum light, GLenum pname, GLint *params );
  friend NVSG_API void GLLightModelf( GLenum pname, GLfloat param );
  friend NVSG_API void GLLightModelfv( GLenum pname, const GLfloat *params );
  friend NVSG_API void GLLightModeli( GLenum pname, GLint param );
  friend NVSG_API void GLLightModeliv( GLenum pname, const GLint *params );
  friend NVSG_API void GLLineStipple( GLint factor, GLushort pattern );
  friend NVSG_API void GLLineWidth( GLfloat width );
  friend NVSG_API void GLListBase( GLuint base );
  friend NVSG_API void GLLoadIdentity( void );
  friend NVSG_API void GLLoadMatrixd( const GLdouble *m );
  friend NVSG_API void GLLoadMatrixf( const GLfloat *m );
  friend NVSG_API void GLLoadName( GLuint name );
  friend NVSG_API void GLLoadProgram( GLenum target, GLuint id, GLsizei len, const GLubyte *program );
  friend NVSG_API void GLLogicOp( GLenum opcode );
  friend NVSG_API void GLMap1d( GLenum target, GLdouble u1, GLdouble u2, GLint stride, GLint order, const GLdouble *points );
  friend NVSG_API void GLMap1f( GLenum target, GLfloat u1, GLfloat u2, GLint stride, GLint order, const GLfloat *points );
  friend NVSG_API void GLMap2d( GLenum target, GLdouble u1, GLdouble u2, GLint ustride, GLint uorder, GLdouble v1, GLdouble v2, GLint vstride, GLint vorder, const GLdouble *points );
  friend NVSG_API void GLMap2f( GLenum target, GLfloat u1, GLfloat u2, GLint ustride, GLint uorder, GLfloat v1, GLfloat v2, GLint vstride, GLint vorder, const GLfloat *points );
  friend NVSG_API void GLMapGrid1d( GLint un, GLdouble u1, GLdouble u2 );
  friend NVSG_API void GLMapGrid1f( GLint un, GLfloat u1, GLfloat u2 );
  friend NVSG_API void GLMapGrid2d( GLint un, GLdouble u1, GLdouble u2, GLint vn, GLdouble v1, GLdouble v2 );
  friend NVSG_API void GLMapGrid2f( GLint un, GLfloat u1, GLfloat u2, GLint vn, GLfloat v1, GLfloat v2 );
  friend NVSG_API void GLMaterialf( GLenum face, GLenum pname, const GLfloat param );
  friend NVSG_API void GLMaterialfv( GLenum face, GLenum pname, const GLfloat *params );
  friend NVSG_API void GLMateriali( GLenum face, GLenum pname, const GLint param );
  friend NVSG_API void GLMaterialiv( GLenum face, GLenum pname, const GLint *params );
  friend NVSG_API void GLMatrixMode( GLenum mode );
  friend NVSG_API void GLMultiTexCoord2f( GLenum texture, GLfloat tx, GLfloat ty );
  friend NVSG_API void GLMultMatrixd( const GLdouble *m );
  friend NVSG_API void GLMultMatrixf( const GLfloat *m );
  friend NVSG_API void GLNewList( GLuint list, GLenum mode );
  friend NVSG_API void GLNormal3b( GLbyte nx, GLbyte ny, GLbyte nz );
  friend NVSG_API void GLNormal3bv( const GLbyte *v );
  friend NVSG_API void GLNormal3d( GLdouble nx, GLdouble ny, GLdouble nz );
  friend NVSG_API void GLNormal3dv( const GLdouble *v );
  friend NVSG_API void GLNormal3f( GLfloat nx, GLfloat ny, GLfloat nz );
  friend NVSG_API void GLNormal3fv( const GLfloat *v );
  friend NVSG_API void GLNormal3i( GLint nx, GLint ny, GLint nz );
  friend NVSG_API void GLNormal3iv( const GLint *v );
  friend NVSG_API void GLNormal3s( GLshort nx, GLshort ny, GLshort nz );
  friend NVSG_API void GLNormal3sv( const GLshort *v );
  friend NVSG_API void GLNormalPointer( GLenum type, GLsizei stride, const GLvoid *pointer );
  friend NVSG_API void GLOrtho( GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble zNear, GLdouble zFar );
  friend NVSG_API void GLPassThrough( GLfloat token );
  friend NVSG_API void GLPixelMapfv( GLenum map, GLsizei mapsize, const GLfloat *values );
  friend NVSG_API void GLPixelMapuiv( GLenum map, GLsizei mapsize, const GLuint *values );
  friend NVSG_API void GLPixelMapusv( GLenum map, GLsizei mapsize, const GLushort *values );
  friend NVSG_API void GLPixelStoref( GLenum pname, GLfloat param );
  friend NVSG_API void GLPixelStorei( GLenum pname, GLint param );
  friend NVSG_API void GLPixelTransferf( GLenum pname, GLfloat param );
  friend NVSG_API void GLPixelTransferi( GLenum pname, GLint param );
  friend NVSG_API void GLPixelZoom( GLfloat xfactor, GLfloat yfactor );
  friend NVSG_API void GLPointSize( GLfloat size );
  friend NVSG_API void GLPolygonMode( GLenum face, GLenum mode );
  friend NVSG_API void GLPolygonOffset( GLfloat factor, GLfloat units );
  friend NVSG_API void GLPolygonStipple( const GLubyte *mask );
  friend NVSG_API void GLPopAttrib( void );
  friend NVSG_API void GLPopClientAttrib( void );
  friend NVSG_API void GLPopMatrix( void );
  friend NVSG_API void GLPopName( void );
  friend NVSG_API void GLPrioritizeTextures( GLsizei n, const GLuint *textures, const GLclampf *priorities );
  friend NVSG_API void GLProgramNamedParameter4f( GLuint id, GLsizei len, const GLubyte *name, GLfloat x, GLfloat y, GLfloat z, GLfloat w );
  friend NVSG_API void GLPushAttrib( GLbitfield mask );
  friend NVSG_API void GLPushClientAttrib( GLbitfield mask );
  friend NVSG_API void GLPushMatrix( void );
  friend NVSG_API void GLPushName( GLuint name );
  friend NVSG_API void GLRasterPos2d( GLdouble x, GLdouble y );
  friend NVSG_API void GLRasterPos2f( GLfloat x, GLfloat y );
  friend NVSG_API void GLRasterPos2i( GLint x, GLint y );
  friend NVSG_API void GLRasterPos2s( GLshort x, GLshort y );
  friend NVSG_API void GLRasterPos3d( GLdouble x, GLdouble y, GLdouble z );
  friend NVSG_API void GLRasterPos3f( GLfloat x, GLfloat y, GLfloat z );
  friend NVSG_API void GLRasterPos3i( GLint x, GLint y, GLint z );
  friend NVSG_API void GLRasterPos3s( GLshort x, GLshort y, GLshort z );
  friend NVSG_API void GLRasterPos4d( GLdouble x, GLdouble y, GLdouble z, GLdouble w );
  friend NVSG_API void GLRasterPos4f( GLfloat x, GLfloat y, GLfloat z, GLfloat w );
  friend NVSG_API void GLRasterPos4i( GLint x, GLint y, GLint z, GLint w );
  friend NVSG_API void GLRasterPos4s( GLshort x, GLshort y, GLshort z, GLshort w );
  friend NVSG_API void GLReadBuffer( GLenum mode );
  friend NVSG_API void GLReadPixels( GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid *pixels );
  friend NVSG_API void GLRectd( GLdouble x1, GLdouble y1, GLdouble x2, GLdouble y2 );
  friend NVSG_API void GLRectf( GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2 );
  friend NVSG_API void GLRecti( GLint x1, GLint y1, GLint x2, GLint y2 );
  friend NVSG_API void GLRects( GLshort x1, GLshort y1, GLshort x2, GLshort y2 );
  friend NVSG_API void GLRenderMode( GLenum mode );
  friend NVSG_API void GLRotated( GLdouble angle, GLdouble x, GLdouble y, GLdouble z );
  friend NVSG_API void GLRotatef( GLfloat angle, GLfloat x, GLfloat y, GLfloat z );
  friend NVSG_API void GLScaled( GLdouble x, GLdouble y, GLdouble z );
  friend NVSG_API void GLScalef( GLfloat x, GLfloat y, GLfloat z );
  friend NVSG_API void GLScissor( GLint x, GLint y, GLsizei width, GLsizei height );
  friend NVSG_API void GLSecondaryColorPointer( GLint size, GLenum type, GLsizei stride, const GLvoid *pointer );
  friend NVSG_API void GLSelectBuffer( GLsizei size, GLuint *buffer );
  friend NVSG_API void GLSetWindowStereoMode( GLboolean displayMode );
  friend NVSG_API void GLShadeModel( GLenum mode );
  friend NVSG_API void GLStencilFunc( GLenum func, GLint ref, GLuint mask );
  friend NVSG_API void GLStencilMask( GLuint mask );
  friend NVSG_API void GLStencilOp( GLenum fail, GLenum zfail, GLenum zpass );
  friend NVSG_API void GLTexCoord1d( GLdouble s );
  friend NVSG_API void GLTexCoord1f( GLfloat s );
  friend NVSG_API void GLTexCoord1i( GLint s );
  friend NVSG_API void GLTexCoord1s( GLshort s );
  friend NVSG_API void GLTexCoord2d( GLdouble s, GLdouble t );
  friend NVSG_API void GLTexCoord2f( GLfloat s, GLfloat t );
  friend NVSG_API void GLTexCoord2fv( const GLfloat *v );
  friend NVSG_API void GLTexCoord2i( GLint s, GLint t );
  friend NVSG_API void GLTexCoord2s( GLshort s, GLshort t );
  friend NVSG_API void GLTexCoord3d( GLdouble s, GLdouble t, GLdouble r );
  friend NVSG_API void GLTexCoord3f( GLfloat s, GLfloat t, GLfloat r );
  friend NVSG_API void GLTexCoord3i( GLint s, GLint t, GLint r );
  friend NVSG_API void GLTexCoord3s( GLshort s, GLshort t, GLshort r );
  friend NVSG_API void GLTexCoord4d( GLdouble s, GLdouble t, GLdouble r, GLdouble q );
  friend NVSG_API void GLTexCoord4f( GLfloat s, GLfloat t, GLfloat r, GLfloat q );
  friend NVSG_API void GLTexCoord4i( GLint s, GLint t, GLint r, GLint q );
  friend NVSG_API void GLTexCoord4s( GLshort s, GLshort t, GLshort r, GLshort q );
  friend NVSG_API void GLTexCoordPointer( GLint size, GLenum type, GLsizei stride, const GLvoid *pointer );
  friend NVSG_API void GLTexEnvf( GLenum target, GLenum pname, GLfloat param );
  friend NVSG_API void GLTexEnvfv( GLenum target, GLenum pname, const GLfloat *params );
  friend NVSG_API void GLTexEnvi( GLenum target, GLenum pname, GLint param );
  friend NVSG_API void GLTexEnviv( GLenum target, GLenum pname, const GLint *params );
  friend NVSG_API void GLTexGend( GLenum coord, GLenum pname, GLdouble param );
  friend NVSG_API void GLTexGenf( GLenum coord, GLenum pname, GLfloat param );
  friend NVSG_API void GLTexGenfv( GLenum coord, GLenum pname, const GLfloat *param );
  friend NVSG_API void GLTexGeni( GLenum coord, GLenum pname, GLint param );
  friend NVSG_API void GLTexImage1D( GLenum target, GLint level, GLint internalformat, GLsizei width, GLint border, GLenum format, GLenum type, const GLvoid *pixels );
  friend NVSG_API void GLTexImage2D( GLenum target, GLint level, GLint internalformat, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const GLvoid *pixels );
  friend NVSG_API void GLTexImage3D( GLenum target, int level, GLenum internalformat, GLsizei width, GLsizei height, GLsizei depth, int border, GLenum format, GLenum type, const void* pixels);
  friend NVSG_API void GLTexParameterf( GLenum target, GLenum pname, GLfloat param );
  friend NVSG_API void GLTexParameteri( GLenum target, GLenum pname, GLint param );
  friend NVSG_API void GLTexParameterfv( GLenum target, GLenum pname, const GLfloat *params );
  friend NVSG_API void GLTexParameteriv( GLenum target, GLenum pname, const GLint *params );
  friend NVSG_API void GLTexSubImage1D( GLenum target, GLint level, GLint xoffset, GLsizei width, GLenum format, GLenum type, const GLvoid *pixels );
  friend NVSG_API void GLTexSubImage2D( GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels );
  friend NVSG_API void GLTranslated( GLdouble x, GLdouble y, GLdouble z );
  friend NVSG_API void GLTranslatef( GLfloat x, GLfloat y, GLfloat z );
  friend NVSG_API void GLVertex2d( GLdouble x, GLdouble y );
  friend NVSG_API void GLVertex2f( GLfloat x, GLfloat y );
  friend NVSG_API void GLVertex2i( GLint x, GLint y );
  friend NVSG_API void GLVertex2s( GLshort x, GLshort y );
  friend NVSG_API void GLVertex3d( GLdouble x, GLdouble y, GLdouble z );
  friend NVSG_API void GLVertex3f( GLfloat x, GLfloat y, GLfloat z );
  friend NVSG_API void GLVertex3fv( const GLfloat *v );
  friend NVSG_API void GLVertex3i( GLint x, GLint y, GLint z );
  friend NVSG_API void GLVertex3s( GLshort x, GLshort y, GLshort z );
  friend NVSG_API void GLVertex4d( GLdouble x, GLdouble y, GLdouble z, GLdouble w );
  friend NVSG_API void GLVertex4f( GLfloat x, GLfloat y, GLfloat z, GLfloat w );
  friend NVSG_API void GLVertex4i( GLint x, GLint y, GLint z, GLint w );
  friend NVSG_API void GLVertex4s( GLshort x, GLshort y, GLshort z, GLshort w );
  friend NVSG_API void GLVertexPointer( GLint size, GLenum type, GLsizei stride, const GLvoid *pointer );
  friend NVSG_API void GLViewport( GLint x, GLint y, GLsizei width, GLsizei height );
#if defined(_WIN32)
  friend NVSG_API BOOL WGLBindTexImage( HPBUFFERARB hPbuffer, int iBuffer );
  friend NVSG_API BOOL WGLChoosePixelFormat( HDC hdc, const int *piAttribIList, const FLOAT *pfAttribFList, UINT nMaxFormats, int *piFormats, UINT *nNumFormats );
  friend NVSG_API HGLRC WGLCreateContext( HDC hdc );
  friend NVSG_API HPBUFFERARB WGLCreatePbuffer( HDC hdc, int pixelFormat, int width, int height, const int *attribList );
  friend NVSG_API BOOL WGLDeleteContext( HGLRC hglrc );
  friend NVSG_API BOOL WGLDestroyPbuffer( HPBUFFERARB hPbuffer );
  friend NVSG_API HGLRC WGLGetCurrentContext(void);
  friend NVSG_API HDC WGLGetCurrentDC(void);
  friend NVSG_API BOOL WGLMakeCurrent( HDC hdc, HGLRC hglrc );
  friend NVSG_API HDC WGLGetPbufferDC( HPBUFFERARB hPbuffer );
  friend NVSG_API BOOL WGLQueryPbuffer( HPBUFFERARB hPbuffer, int attribute, int *value );
  friend NVSG_API int WGLReleasePbufferDC( HPBUFFERARB hPbuffer, HDC hdc );
  friend NVSG_API BOOL WGLReleaseTexImage( HPBUFFERARB hPbuffer, int iBuffer );
  friend NVSG_API BOOL WGLShareLists( HGLRC hglrc1, HGLRC hglrc2 );
  friend NVSG_API BOOL WGLJoinSwapGroupNV(HDC hDC, GLuint group);
  friend NVSG_API BOOL WGLBindSwapBarrierNV(GLuint group, GLuint barrier);
  friend NVSG_API BOOL WGLQuerySwapGroupNV(HDC hDC, GLuint * group, GLuint * barrier);
  friend NVSG_API BOOL WGLQueryMaxSwapGroupsNV(HDC hDC, GLuint * maxGroups, GLuint * maxBarriers);
  friend NVSG_API BOOL WGLQueryFrameCountNV(HDC hDC, GLuint * count);
  friend NVSG_API BOOL WGLResetFrameCountNV(HDC hDC);
  // routines to handle bypassing of WGLCreateContext and WGLDeleteContext  
  friend NVSG_API void WGLAttachContext(HDC, HGLRC);
  friend NVSG_API void WGLDetachContext(HGLRC);
#elif defined(LINUX)
  friend NVSG_API GLXContext GLXCreateContext(Display *dpy, XVisualInfo *vis, GLXContext shareList, Bool direct);
  friend NVSG_API void GLXDestroyContext(Display *dpy, GLXContext ctx);
  friend NVSG_API GLXContext GLXGetCurrentContext();
  friend NVSG_API GLXDrawable GLXGetCurrentDrawable();
  friend NVSG_API Bool GLXMakeCurrent(Display *dpy, GLXDrawable drawable, GLXContext ctx);
  // routines to handle bypassing of GLXCreateContext and GLXDestroyContext  
  friend NVSG_API void GLXAttachContext(Display*, GLXContext);
  friend NVSG_API void GLXDetachContext(GLXContext);
#endif
  // some NVSG extensions

  // CgFXSetDevice extension
  friend NVSG_API GLboolean GLCgFXSetDevice();
  friend NVSG_API void GLCgFXFreeDevice();

  // Texture lookup extension
  friend NVSG_API void GLAddTextureFileForLookup(const std::string& file, const nvsg::Texture * tex);
  friend NVSG_API void GLRemoveTextureFileFromLookup(const std::string& file);
  friend NVSG_API const nvsg::Texture * GLLookupTexture(const std::string& file);

  // query hw capacity for texcoords arrays and texture image units
  friend NVSG_API GLint GLGetMaxTexCoords();
  friend NVSG_API GLint GLGetMaxTexUnits();

  // extension query
  //! Test if a given extension is available on the current context.
  /** \note Not all exported extensions might be available.
    * \returns  \c true, if all functions of \a extension are available, otherwise \c false  */
  friend NVSG_API GLboolean GLIsExtensionAvailable(const std::string& extension);
  //! Test if a given extension is exported on the current context.
  /** \note Not all exported extensions might be available.
    * \returns  \c true, if \a extension is exported, otherwise \c false  */
  friend NVSG_API GLboolean GLIsExtensionExported(const std::string& extension);

public:  // public interface
#if defined(LINUX)
  inline Display* getDisplay() { return m_dpy; }
#endif

private: // private interface

#if defined(_WIN32)
  GLContext(HDC hdc, HGLRC hglrc);  
#elif defined(LINUX)
  GLContext(Display *dpy, GLXContext ctx);  
#endif
  ~GLContext();

#if defined(_WIN32)
  void init(); // init function pointers and caches
#elif defined(LINUX)
  void init(GLXDrawable drw);
#endif
  void exit(); // GL objects cleanup

  // Share sharable GL objects with the GLContext referenced through \a source.
  void share(const GLContext& source);

  // helper
  bool isTexture(GLuint name) const;
  bool isVBO(GLuint name)     const;
  bool isProgram(GLuint name) const;
  bool isDisplayList(GLuint name) const;
  bool isOcclusionQuery(GLuint name) const;
  bool isQuery(GLuint name) const;
  void addTexture(const std::string& file, const nvsg::Texture* texture);
  void removeTexture(const std::string& file);
  const nvsg::Texture * lookupTexture(const std::string& file) const;


public: // public data
  GLFunctions * m_glFunctions;    //!<  Pointer to OpenGL functions wrapper object

  // glFrontFaceCache
  GLenum m_fontFaceMode;

private: // private data

#if defined(_WIN32)
  HDC           m_hdc;          //!< Handle to the corresponding device context
  HGLRC         m_hglrc;        //!< Handle of the corresponding GL render context
#elif defined(LINUX)  
  static Display * m_dpy;       //!< Handle to corresponding X display
  GLXContext       m_ctx;       //!< Handle to corresponding X render context
  GLXDrawable      m_drawable;  //!< Handle to corresponding X drawable
#endif

  GLObjects   * m_glObjects;    //!<  Handling of shareable and non-shareable objects
  GLint         m_maxTexCoords; //!<  Maximal number of supported client texture slots on this HW
  GLint         m_maxTexUnits;  //!<  Maximal number of supported texture slots on this HW
#if defined( _DEBUG )
  bool          m_inBeginEndLoop;  //!<  \c true, when inside a glBegin/glEnd loop, otherwise \c false
#endif

  nvutil::RCObject * m_CgFXDevice; //!< CgFX device - unique per context but might be shared between multiple clients
};

} // namespace nvgl

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// inline implementations
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

// need defines now with implementing
#include "nvgl/GLFunctions.h"
#include "nvgl/GLObjects.h"

namespace nvgl 
{

//! Replacement of glAccum
/** This function asserts on the result before returning.  */
inline void  GLAccum( GLenum op, GLfloat value )
{
  glAccum( op, value );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glActiveTexture
/** This function stores the texture to activate for subsequent texture operations.  */
inline  void  GLActiveTexture( GLenum texture )
{
  glContext->m_glFunctions->glActiveTexture(texture);
}

//! Replacement of glAlphaFunc
/** This function caches the alpha function settings and asserts on the result before returning.  */
inline  void  GLAlphaFunc( GLenum func, GLclampf ref )
{
  glAlphaFunc( func, ref );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glAreTexturesResident
/** This function asserts on the result before returning.  */
inline  GLboolean GLAreTexturesResident( GLsizei n, const GLuint *textures, GLboolean *residences )
{
  GLboolean ok = glAreTexturesResident( n, textures, residences );
  ASSERT_GL_NO_ERROR;
  return( ok );
}

//! Replacement of glArrayElement
/** This function asserts on the result before returning.  */
inline  void  GLArrayElement( GLint i )
{
  glArrayElement( i );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glBegin
/** This function asserts on the result before returning.  */
inline  void  GLBegin( GLenum mode )
{
  glBegin( mode );
#if defined( _DEBUG )
  glContext->m_inBeginEndLoop = true;
#endif
}

//! Replacement of glBeginOcclusionQueryNV
/** This function hides the extension character of glBeginOcclusionQueryNV and asserts on the result before returning.  */
inline  void  GLBeginOcclusionQuery( GLuint id )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glBeginOcclusionQueryNV );
  glContext->m_glFunctions->glBeginOcclusionQueryNV( id );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glBeginQueryARB
/** This function hides the extension character of glBeginQueryARB and asserts on the result before returning. */
inline  void  GLBeginQuery( GLenum target, GLuint id )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glBeginQueryARB );
  glContext->m_glFunctions->glBeginQueryARB( target, id );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glBindBufferARB
/** This function hides the extension character of glBindBufferARB and asserts on the result before returning.  */
inline  void  GLBindBufferARB(GLenum target, GLuint id)
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glBindBufferARB );
  glContext->m_glFunctions->glBindBufferARB( target, id );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glBufferDataARB
/** This function hides the extension character of glBufferDataARB and asserts on the result before returning.  */
inline  void  GLBufferDataARB( GLenum target, GLsizeiptrARB size, const GLvoid *data, GLenum usage )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glBufferDataARB );
  glContext->m_glFunctions->glBufferDataARB( target, size, data, usage );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glBufferSubDataARB
/** This function hides the extension character of glBufferSubDataARB and asserts on the result before returning.  */
inline  void  GLBufferSubDataARB( GLenum target, GLintptrARB offset, GLsizeiptrARB size, const GLvoid *data )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glBufferSubDataARB );
  glContext->m_glFunctions->glBufferSubDataARB( target, offset, size, data );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glDeleteBuffersARB
/** This function hides the extension character of glDeleteBuffersARB and asserts on the result before returning.  */
inline  void  GLDeleteBuffersARB(GLsizei n, const GLuint *buffers)
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glDeleteBuffersARB );
  glContext->m_glFunctions->glDeleteBuffersARB( n, buffers );
  glContext->m_glObjects->removeBuffers(n, buffers);
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGenBuffersARB
/** This function hides the extension character of glGenBuffersARB and asserts on the result before returning.  */
inline  void  GLGenBuffersARB(GLsizei n, GLuint *buffers)
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glGenBuffersARB );
  glContext->m_glFunctions->glGenBuffersARB( n, buffers );
  glContext->m_glObjects->addBuffers(n, buffers);
  ASSERT_GL_NO_ERROR;
}

//void GLAPI glGetBufferParameterivARB (GLenum target, GLenum pname, GLint *params);
//void GLAPI glGetBufferPointervARB (GLenum target, GLenum pname, GLvoid* *params);
//void GLAPI glGetBufferSubDataARB (GLenum target, GLintptrARB offset, GLsizeiptrARB size, GLvoid *data);

//! Replacement of glMapBufferARB
/** This function hides the extension character of glMapBufferARB and asserts on the result before returning.  */
inline  GLvoid*  GLMapBufferARB( GLenum target, GLenum access )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glMapBufferARB );
  GLvoid * p = glContext->m_glFunctions->glMapBufferARB( target, access );
  ASSERT_GL_NO_ERROR;
  return p;
}

//! Replacement of glUnmapBufferARB
/** This function hides the extension character of glUnmapBufferARB and asserts on the result before returning.  */
inline  GLboolean  GLUnmapBufferARB( GLenum target )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glUnmapBufferARB );
  GLboolean b = glContext->m_glFunctions->glUnmapBufferARB( target );
  ASSERT_GL_NO_ERROR;
  return b;
}

//! Replacement of glIsBufferARB
/** This function tests if for the current render context, name denotes a valid vertex buffer object. */ 
inline  GLboolean  GLIsBufferARB( GLuint name )
{
  __ASSERT(glContext->m_glFunctions && glContext->m_glFunctions->glIsBufferARB);
  __ASSERT(!!glContext->isVBO(name)==!!glContext->m_glFunctions->glIsBufferARB(name));
  return glContext->isVBO(name);
}

//! Replacement of glBindProgramNV
/** This function hides the extension character of glBindProgramNV and asserts on the result before returning.  */
inline  void  GLBindProgram( GLenum target, GLuint id )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glBindProgramNV );
  glContext->m_glFunctions->glBindProgramNV( target, id );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glBitmap
/** This function asserts on the result before returning.  */
inline  void  GLBitmap( GLsizei width, GLsizei height, GLfloat xorig, GLfloat yorig, GLfloat xmove, GLfloat ymove, const GLubyte *bitmap )
{
  glBitmap( width, height, xorig, yorig, xmove, ymove, bitmap );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glBlendFunc
/** This function caches the blend func settings and asserts on the result before returning.  */
inline  void  GLBlendFunc( GLenum sfactor, GLenum dfactor )
{
  glBlendFunc( sfactor, dfactor );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glBindTexture
/** This function activates the texture unit if necessary, calls glBindTexture and asserts on the result before returning.  */
inline  void  GLBindTexture( GLenum target, GLuint texture )
{
  glBindTexture( target, texture );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glCallList
/** This function asserts on the result before returning.  */
inline  void  GLCallList( GLuint list )
{
  glCallList( list );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glCallLists
/** This function asserts on the result before returning.  */
inline  void  GLCallLists( GLsizei n, GLenum type, const GLvoid *lists )
{
  glCallLists( n, type, lists );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glClear
/** This function asserts on the result before returning.  */
inline  void  GLClear( GLbitfield mask )
{
  glClear( mask );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glClearAccum
/** This function asserts on the result before returning.  */
inline  void  GLClearAccum( GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha )
{
  glClearAccum( red, green, blue, alpha );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glClearColor
/** This function caches the clear color settings and asserts on the result before returning.  */
inline  void  GLClearColor( GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha )
{
  glClearColor(red, green, blue, alpha);
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glClearDepth
/** This function caches the clear depth setting and asserts on the result before returning.  */
inline  void  GLClearDepth( GLclampd depth )
{
  glClearDepth( depth );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glClearIndex
/** This function asserts on the result before returning.  */
inline  void  GLClearIndex( GLfloat c )
{
  glClearIndex( c );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glClearStencil
/** This function asserts on the result before returning.  */
inline  void  GLClearStencil( GLint s )
{
  glClearStencil( s );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glClientActiveTexture
/** This functions stores the next active texture unit. */
inline  void  GLClientActiveTexture( GLenum texture )
{
  glContext->m_glFunctions->glClientActiveTexture( texture );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glClipPlane
/** This function caches the clip plane settings and asserts on the result before returning.  */
inline  void  GLClipPlane( GLenum plane, const GLdouble *equation )
{
  glClipPlane(plane, equation);
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glColor3b
/** This function asserts on the result before returning.  */
inline  void  GLColor3b( GLbyte red, GLbyte green, GLbyte blue )
{
  glColor3b( red, green, blue );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glColor3d
/** This function asserts on the result before returning.  */
inline  void  GLColor3d( GLdouble red, GLdouble green, GLdouble blue )
{
  glColor3d( red, green, blue );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glColor3f
/** This function asserts on the result before returning.  */
inline  void  GLColor3f( GLfloat red, GLfloat green, GLfloat blue )
{
  glColor3f( red, green, blue );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glColor3fv
/** This function asserts on the result before returning.  */
inline  void  GLColor3fv( const GLfloat * v )
{
  glColor3fv( v );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glColor3i
/** This function asserts on the result before returning.  */
inline  void  GLColor3i( GLint red, GLint green, GLint blue )
{
  glColor3i( red, green, blue );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glColor3s
/** This function asserts on the result before returning.  */
inline  void  GLColor3s( GLshort red, GLshort green, GLshort blue )
{
  glColor3s( red, green, blue );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glColor3ub
/** This function asserts on the result before returning.  */
inline  void  GLColor3ub( GLubyte red, GLubyte green, GLubyte blue )
{
  glColor3ub( red, green, blue );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glColor3ui
/** This function asserts on the result before returning.  */
inline  void  GLColor3ui( GLuint red, GLuint green, GLuint blue )
{
  glColor3ui( red, green, blue );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glColor3us
/** This function asserts on the result before returning.  */
inline  void  GLColor3us( GLushort red, GLushort green, GLushort blue )
{
  glColor3us( red, green, blue );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glColor4b
/** This function asserts on the result before returning.  */
inline  void  GLColor4b( GLbyte red, GLbyte green, GLbyte blue, GLbyte alpha )
{
  glColor4b( red, green, blue, alpha );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glColor4d
/** This function asserts on the result before returning.  */
inline  void  GLColor4d( GLdouble red, GLdouble green, GLdouble blue, GLdouble alpha )
{
  glColor4d( red, green, blue, alpha );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glColor4f
/** This function asserts on the result before returning.  */
inline  void  GLColor4f( GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha )
{
  glColor4f( red, green, blue, alpha );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glColor4i
/** This function asserts on the result before returning.  */
inline  void  GLColor4i( GLint red, GLint green, GLint blue, GLint alpha )
{
  glColor4i( red, green, blue, alpha );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glColor4s
/** This function asserts on the result before returning.  */
inline  void  GLColor4s( GLshort red, GLshort green, GLshort blue, GLshort alpha )
{
  glColor4s( red, green, blue, alpha );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glColor4ub
/** This function asserts on the result before returning.  */
inline  void  GLColor4ub( GLubyte red, GLubyte green, GLubyte blue, GLubyte alpha )
{
  glColor4ub( red, green, blue, alpha );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glColor4ui
/** This function asserts on the result before returning.  */
inline  void  GLColor4ui( GLuint red, GLuint green, GLuint blue, GLuint alpha )
{
  glColor4ui( red, green, blue, alpha );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glColor4us
/** This function asserts on the result before returning.  */
inline  void  GLColor4us( GLushort red, GLushort green, GLushort blue, GLushort alpha )
{
  glColor4us( red, green, blue, alpha );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glColorMask
/** This function asserts on the result before returning.  */
inline  void  GLColorMask( GLboolean red, GLboolean green, GLboolean blue, GLboolean alpha )
{
  glColorMask( red, green, blue, alpha );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glColorMaterial
/** This function asserts on the result before returning.  */
inline  void  GLColorMaterial( GLenum face, GLenum mode )
{
  glColorMaterial( face, mode );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glColorPointer
/** This function caches the color pointer settings and asserts on the result before returning.  */
inline  void  GLColorPointer( GLint size, GLenum type, GLsizei stride, const GLvoid *pointer )
{
  glColorPointer( size, type, stride, pointer );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glCompressedTexImage1D
/** This function hides the extension character of glCompressedTexImage1D and asserts on the result before returning.  */
inline  void  GLCompressedTexImage1D( GLenum target, int level, GLenum internalformat, GLsizei width, int border, GLsizei imageSize, const void *data )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glCompressedTexImage1D );
  glContext->m_glFunctions->glCompressedTexImage1D( target, level, internalformat, width, border, imageSize, data );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glCompressedTexImage2D
/** This function hides the extension character of glCompressedTexImage2D and asserts on the result before returning.  */
inline  void  GLCompressedTexImage2D( GLenum target, int level, GLenum internalformat, GLsizei width, GLsizei height, int border, GLsizei imageSize, const void *data )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glCompressedTexImage2D );
  glContext->m_glFunctions->glCompressedTexImage2D( target, level, internalformat, width, height, border, imageSize, data );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glCompressedTexImage3D
/** This function hides the extension character of glCompressedTexImage3D and asserts on the result before returning.  */
inline  void  GLCompressedTexImage3D( GLenum target, int level, GLenum internalformat, GLsizei width, GLsizei height, GLsizei depth, int border, GLsizei imageSize, const void *data )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glCompressedTexImage3D );
  glContext->m_glFunctions->glCompressedTexImage3D( target, level, internalformat, width, height, depth, border, imageSize, data );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glCopyPixels
/** This function asserts on the result before returning.  */
inline  void  GLCopyPixels( GLint x, GLint y, GLsizei width, GLsizei height, GLenum type )
{
  glCopyPixels( x, y, width, height, type );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glCopyTexImage1D
/** This function asserts on the result before returning.  */
inline  void  GLCopyTexImage1D( GLenum target, GLint level, GLenum internalFormat, GLint x, GLint y, GLsizei width, GLint border )
{
  glCopyTexImage1D( target, level, internalFormat, x, y, width, border );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glCopyTexImage2D
/** This function asserts on the result before returning.  */
inline  void  GLCopyTexImage2D( GLenum target, GLint level, GLenum internalFormat, GLint x, GLint y, GLsizei width, GLsizei height, GLint border )
{
  glCopyTexImage2D( target, level, internalFormat, x, y, width, height, border );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glCopyTexSubImage1D
/** This function asserts on the result before returning.  */
inline  void  GLCopyTexSubImage1D( GLenum target, GLint level, GLint xoffset, GLint x, GLint y, GLsizei width )
{
  glCopyTexSubImage1D( target,level, xoffset, x, y, width );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glCopyTexSubImage2D
/** This function asserts on the result before returning.  */
inline  void  GLCopyTexSubImage2D( GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint x, GLint y, GLsizei width, GLsizei height )
{
  glCopyTexSubImage2D( target,level, xoffset, yoffset, x, y, width, height );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glCullFace
/** This function caches the cull face setting and asserts on the result before returning.  */
inline  void  GLCullFace( GLenum mode )
{
  glCullFace( mode );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glDeleteLists
/** This function asserts on the result before returning.  */
inline  void  GLDeleteLists( GLuint list, GLsizei range )
{
  glDeleteLists( list, range );
  glContext->m_glObjects->removeLists(range, list);
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glDeleteOcclusionQueriesNV
/** This function hides the extension character of glDeleteOcclusionQueriesNV and asserts on the result before returning.  */
inline  void  GLDeleteOcclusionQueries( GLsizei n, const GLuint *ids )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glDeleteOcclusionQueriesNV );
  glContext->m_glFunctions->glDeleteOcclusionQueriesNV( n, ids );
  glContext->m_glObjects->removeOcclusionQueries(n, ids);
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glDeleteProgramsNV
/** This function hides the extension character of glDeleteProgramsNV and asserts on the result before returning.  */
inline  void  GLDeletePrograms( GLsizei n, const GLuint *ids )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glDeleteProgramsNV );
  glContext->m_glFunctions->glDeleteProgramsNV( n, ids );
  glContext->m_glObjects->removePrograms(n, ids);
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glDeleteQueriesARB
/** This function hides the extension character of glDeleteQueriesARB and asserts on the result before returning. */
inline  void  GLDeleteQueries( GLsizei n, const GLuint *ids )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glDeleteQueriesARB );
  glContext->m_glFunctions->glDeleteQueriesARB( n, ids );
  glContext->m_glObjects->removeQueries( n, ids );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glDeleteTextures
/** This function asserts on the result before returning.  */
inline  void  GLDeleteTextures( GLsizei n, const GLuint *textures )
{
  glDeleteTextures( n, textures );
  glContext->m_glObjects->removeTextures(n, textures);
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glDepthFunc
/** This function caches the depth function setting and asserts on the result before returning.  */
inline  void  GLDepthFunc( GLenum func )
{
  glDepthFunc( func );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glDepthMask
/** This function caches the depth mask setting and asserts on the result before returning.  */
inline  void  GLDepthMask( GLboolean flag )
{
  glDepthMask( flag );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glDepthRange
/** This function caches the depth range settings and asserts on the result before returning.  */
inline  void  GLDepthRange( GLclampd znear, GLclampd zfar )
{
  glDepthRange( znear, zfar );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glDisable
/** This function caches the settings and asserts on the result before returning.  */
inline  void  GLDisable( GLenum cap ) 
{
  glDisable( cap );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glDisableClientState
/** This function caches the client state settings and asserts on the result before returning.  */
inline  void  GLDisableClientState( GLenum array ) 
{
  glDisableClientState( array );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glDrawArrays
/** This function asserts on the result before returning.  */
inline  void  GLDrawArrays( GLenum mode, GLint first, GLsizei count )
{
  glDrawArrays( mode, first, count );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glDrawBuffer
/** This function asserts on the result before returning.  */
inline  void  GLDrawBuffer( GLenum mode ) 
{ 
  glDrawBuffer( mode );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glDrawElements
/** This function asserts on the result before returning.  */
inline  void  GLDrawElements( GLenum mode, GLsizei count, GLenum type, const GLvoid *indices )
{
  glDrawElements( mode, count, type, indices );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glMultiDrawElements
/** This function asserts on the result before returning.  */
inline  void  GLMultiDrawElements( GLenum mode, const GLsizei *count, GLenum type, const GLvoid* *indices, GLsizei primcount )
{
  glContext->m_glFunctions->glMultiDrawElements( mode, count, type, indices, primcount );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glDrawPixels
/** This function asserts on the result before returning.  */
inline  void  GLDrawPixels( GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels )
{
  glDrawPixels( width, height, format, type, pixels );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glDrawRangeElements
/** This function asserts on the result before returning.  */
inline  void  GLDrawRangeElements( GLenum mode, GLuint start, GLuint end, GLsizei count, GLenum type, const GLvoid *indices )
{
  glContext->m_glFunctions->glDrawRangeElements( mode, start, end, count, type, indices );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glEdgeFlagPointer
/** This function caches the edge flag pointer setting and asserts on the result before returning.  */
inline  void  GLEdgeFlagPointer( GLsizei stride, const GLboolean *pointer )
{
  glEdgeFlagPointer( stride, pointer );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glEdgeFlag
/** This function asserts on the result before returning.  */
inline  void  GLEdgeFlag( GLboolean flag )
{
  glEdgeFlag( flag );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glEdgeFlagv
/** This function asserts on the result before returning.  */
inline  void  GLEdgeFlagv( const GLboolean *flag )
{
  glEdgeFlagv( flag );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glEnable
/** This function caches the settings and asserts on the result before returning.  */
inline  void  GLEnable( GLenum cap ) 
{ 
  glEnable( cap );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glEnableClientState
/** This function caches the client state settings and asserts on the result before returning.  */
inline  void  GLEnableClientState( GLenum array ) 
{
  glEnableClientState( array );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glEnd
/** This function asserts on the result before returning.  */
inline  void  GLEnd( void )
{
  glEnd();
#if defined( _DEBUG )
  glContext->m_inBeginEndLoop = false;
#endif
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glEndList
/** This function asserts on the result before returning.  */
inline  void  GLEndList( void )
{
  glEndList();
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glEndOcclusionQueryNV
/** This function hides the extension character of glEndOcclusionQueryNV and asserts on the result before returning.  */
inline  void  GLEndOcclusionQuery( void )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glEndOcclusionQueryNV );
  glContext->m_glFunctions->glEndOcclusionQueryNV();
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glEndQueryARB
/** This function hides the extension character of glEndQueryARB and asserts on the result before returning. */
inline  void  GLEndQuery( GLenum target )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glEndQueryARB );
  glContext->m_glFunctions->glEndQueryARB( target );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glEvalCoord1d
/** This function asserts on the result before returning.  */
inline  void  GLEvalCoord1d( GLdouble u )
{
  glEvalCoord1d( u );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glEvalCoord1f
/** This function asserts on the result before returning.  */
inline  void  GLEvalCoord1f( GLfloat u )
{
  glEvalCoord1f( u );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glEvalCoord2d
/** This function asserts on the result before returning.  */
inline  void  GLEvalCoord2d( GLdouble u, GLdouble v )
{
  glEvalCoord2d( u, v );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glEvalCoord2f
/** This function asserts on the result before returning.  */
inline  void  GLEvalCoord2f( GLfloat u, GLfloat v )
{
  glEvalCoord2f( u, v );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glEvalMesh1
/** This function asserts on the result before returning.  */
inline  void  GLEvalMesh1( GLenum mode, GLint i1, GLint i2 )
{
  glEvalMesh1( mode, i1, i2 );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glEvalMesh2
/** This function asserts on the result before returning.  */
inline  void  GLEvalMesh2( GLenum mode, GLint i1, GLint i2, GLint j1, GLint j2 )
{
  glEvalMesh2( mode, i1, i2, j1, j2 );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glEvalPoint1
/** This function asserts on the result before returning.  */
inline  void  GLEvalPoint1( GLint i )
{
  glEvalPoint1( i );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glEvalPoint2
/** This function asserts on the result before returning.  */
inline  void  GLEvalPoint2( GLint i, GLint j )
{
  glEvalPoint2( i, j );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glFeedbackBuffer
/** This function asserts on the result before returning.  */
inline  void  GLFeedbackBuffer( GLsizei size, GLenum type, GLfloat *buffer )
{
  glFeedbackBuffer( size, type, buffer );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glFinish
/** This function asserts on the result before returning.  */
inline  void  GLFinish( void )
{
  glFinish();
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glFlush
/** This function asserts on the result before returning.  */
inline  void  GLFlush( void )
{
  glFlush();
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glFogCoordPointer
/** This function caches the fog coordinate pointer setting and asserts on the result before returning.  */
inline  void  GLFogCoordPointer( GLenum type, GLsizei stride, const GLvoid *pointer )
{
  glContext->m_glFunctions->glFogCoordPointer( type, stride, pointer );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glFogf
/** This function caches the fog settings and asserts on the result before returning.  */
inline  void  GLFogf( GLenum pname, GLfloat param ) 
{
  glFogf( pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glFogfv
/** This function caches the fog settings and asserts on the result before returning.  */
inline  void  GLFogfv( GLenum pname, const GLfloat *params )  
{ 
  glFogfv( pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glFogi
/** This function caches the fog settings and asserts on the result before returning.  */
inline  void  GLFogi( GLenum pname, GLint param ) 
{ 
  glFogi(pname, param);
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glFogiv
/** This function asserts on the result before returning.  */
inline  void  GLFogiv( GLenum pname, const GLint *params ) 
{ 
  glFogiv(pname, params);
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glFrontFace
/** This function caches the front face setting and asserts on the result before returning.  */
inline  void  GLFrontFace( GLenum mode )
{
  if (mode != glContext->m_fontFaceMode)
  {
    glFrontFace( mode );
    ASSERT_GL_NO_ERROR;
    
    glContext->m_fontFaceMode = mode;
  }
}

//! Replacement of glFrustum
/** This function asserts on the result before returning.  */
inline  void  GLFrustum( GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble zNear, GLdouble zFar )
{
  glFrustum( left, right, bottom, top, zNear, zFar );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGenLists
/** This function asserts on the result before returning.  */
inline  GLuint  GLGenLists( GLsizei range )
{
  GLuint  firstList = glGenLists( range );
  glContext->m_glObjects->addLists(range, firstList);
  ASSERT_GL_NO_ERROR;
  return( firstList );
}

//! Replacement of glGenOcclusionQueriesNV
/** This function hides the extension character of glGenOcclusionQueriesNV and asserts on the result before returning.  */
inline  void  GLGenOcclusionQueries( GLsizei n, GLuint *ids )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glGenOcclusionQueriesNV );
  glContext->m_glFunctions->glGenOcclusionQueriesNV( n, ids );
  glContext->m_glObjects->addOcclusionQueries(n, ids);
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGenProgramsNV
/** This function hides the extension character of glGenProgramsNV and asserts on the result before returning.  */
inline  void  GLGenPrograms( GLsizei n, GLuint *ids )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glGenProgramsNV );
  glContext->m_glFunctions->glGenProgramsNV( n, ids );
  glContext->m_glObjects->addPrograms(n, ids);
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGenQueriesARB
/** This function hides the extension character of glGenQueriesARB and asserts on the result before returning. */
inline  void  GLGenQueries( GLsizei n, GLuint *ids )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glGenQueriesARB );
  glContext->m_glFunctions->glGenQueriesARB( n, ids );
  glContext->m_glObjects->addQueries( n, ids );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGenTextures
/** This function asserts on the result before returning.  */
inline  void  GLGenTextures( GLsizei n, GLuint *textures )
{
  glGenTextures( n, textures );
  glContext->m_glObjects->addTextures(n, textures);
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetBooleanv
/** This function asserts on the result before returning.  */
inline  void  GLGetBooleanv( GLenum pname, GLboolean *params )
{
  glGetBooleanv( pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetDoublev
/** This function asserts on the result before returning.  */
inline  void  GLGetDoublev( GLenum pname, GLdouble *params )
{
  glGetDoublev( pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetFloatv
/** This function asserts on the result before returning.  */
inline  void  GLGetFloatv( GLenum pname, GLfloat *params )
{
  glGetFloatv( pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetIntegerv
/** This function asserts on the result before returning.  */
inline  void  GLGetIntegerv( GLenum pname, GLint *params )
{
  glGetIntegerv( pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetClipPlane
/** This function asserts on the result before returning.  */
inline  void  GLGetClipPlane( GLenum plane, GLdouble *equation )
{
  glGetClipPlane( plane, equation );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetLightfv
/** This function asserts on the result before returning.  */
inline  void  GLGetLightfv( GLenum light, GLenum pname, GLfloat *params )
{
  glGetLightfv( light, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetLightiv
/** This function asserts on the result before returning.  */
inline  void  GLGetLightiv( GLenum light, GLenum pname, GLint *params )
{
  glGetLightiv( light, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetMapdv
/** This function asserts on the result before returning.  */
inline  void  GLGetMapdv( GLenum target, GLenum query, GLdouble *v )
{
  glGetMapdv( target, query, v );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetMapfv
/** This function asserts on the result before returning.  */
inline  void  GLGetMapfv( GLenum target, GLenum query, GLfloat *v )
{
  glGetMapfv( target, query, v );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetMapiv
/** This function asserts on the result before returning.  */
inline  void  GLGetMapiv( GLenum target, GLenum query, GLint *v )
{
  glGetMapiv( target, query, v );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetMaterialfv
/** This function asserts on the result before returning.  */
inline  void  GLGetMaterialfv( GLenum face, GLenum pname, GLfloat *params )
{
  glGetMaterialfv( face, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetMaterialiv
/** This function asserts on the result before returning.  */
inline  void  GLGetMaterialiv( GLenum face, GLenum pname, GLint *params )
{
  glGetMaterialiv( face, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetOcclusionQueryuivNV
/** This function hides the extension character of glGetOcclusionQueryuivNV and asserts on the result before returning.  */
inline  void  GLGetOcclusionQueryuiv( GLuint id, GLenum pname, GLuint *params )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glGetOcclusionQueryuivNV );
  glContext->m_glFunctions->glGetOcclusionQueryuivNV( id, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetPixelMapfv
/** This function asserts on the result before returning.  */
inline  void  GLGetPixelMapfv( GLenum map, GLfloat *values )
{
  glGetPixelMapfv( map, values );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetPixelMapuiv
/** This function asserts on the result before returning.  */
inline  void  GLGetPixelMapuiv( GLenum map, GLuint *values )
{
  glGetPixelMapuiv( map, values );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetPixelMapusv
/** This function asserts on the result before returning.  */
inline  void  GLGetPixelMapusv( GLenum map, GLushort *values )
{
  glGetPixelMapusv( map, values );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetPointerv
/** This function asserts on the result before returning.  */
inline  void  GLGetPointerv( GLenum pname, GLvoid* *params )
{
  glGetPointerv( pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetPolygonStipple
/** This function asserts on the result before returning.  */
inline  void  GLGetPolygonStipple( GLubyte *mask )
{
  glGetPolygonStipple( mask );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetQueryivARB
/** This function hides the extension character of glGetQueryivARB and asserts on the result before returning. */
inline  void  GLGetQueryiv( GLenum target, GLenum pname, GLint *params )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glGetQueryivARB );
  glContext->m_glFunctions->glGetQueryivARB( target, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetQueryObjectivARB
/** This function hides the extension character of glGetQueryObjectivARB and asserts on the result before returning. */
inline  void  GLGetQueryObjectiv( GLuint id, GLenum pname, GLint *params )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glGetQueryObjectivARB );
  glContext->m_glFunctions->glGetQueryObjectivARB( id, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetQueryObjectuivARB
/** This function hides the extension character of glGetQueryObjectuivARB and asserts on the result before returning. */
inline  void  GLGetQueryObjectuiv( GLuint id, GLenum pname, GLuint *params )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glGetQueryObjectuivARB );
  glContext->m_glFunctions->glGetQueryObjectuivARB( id, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetString
/** This function asserts on the result before returning.  */
inline  const GLubyte * GLGetString( GLenum name )
{
  const GLubyte * ret = glGetString( name );
  ASSERT_GL_NO_ERROR;
  return( ret );
}

//! Replacement of glGetTexEnvfv
/** This function asserts on the result before returning.  */
inline  void  GLGetTexEnvfv( GLenum target, GLenum pname, GLfloat *params )
{
  glGetTexEnvfv( target,pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetTexEnviv
/** This function asserts on the result before returning.  */
inline  void  GLGetTexEnviv( GLenum target, GLenum pname, GLint *params )
{
  glGetTexEnviv( target,pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetTexGendv
/** This function asserts on the result before returning.  */
inline  void  GLGetTexGendv( GLenum coord, GLenum pname, GLdouble *params )
{
  glGetTexGendv( coord, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetTexGenfv
/** This function asserts on the result before returning.  */
inline  void  GLGetTexGenfv( GLenum coord, GLenum pname, GLfloat *params )
{
  glGetTexGenfv( coord, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetTexGeniv
/** This function asserts on the result before returning.  */
inline  void  GLGetTexGeniv( GLenum coord, GLenum pname, GLint *params )
{
  glGetTexGeniv( coord, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetTexImage
/** This function asserts on the result before returning.  */
inline  void  GLGetTexImage( GLenum target, GLint level, GLenum format, GLenum type, GLvoid *pixels )
{
  glGetTexImage( target, level, format, type, pixels );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetTexLevelParameterfv
/** This function asserts on the result before returning.  */
inline  void  GLGetTexLevelParameterfv( GLenum target, GLint level, GLenum pname, GLfloat *params )
{
  glGetTexLevelParameterfv( target, level, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetTexLevelParameteriv
/** This function asserts on the result before returning.  */
inline  void  GLGetTexLevelParameteriv( GLenum target, GLint level, GLenum pname, GLint *params )
{
  glGetTexLevelParameteriv( target, level, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetTexParameterfv
/** This function asserts on the result before returning.  */
inline  void  GLGetTexParameterfv( GLenum target, GLenum pname, GLfloat *params )
{
  glGetTexParameterfv( target, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glGetTexParameteriv
/** This function asserts on the result before returning.  */
inline  void  GLGetTexParameteriv( GLenum target, GLenum pname, GLint *params )
{
  glGetTexParameteriv( target, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glHint
/** This function asserts on the result before returning.  */
inline  void  GLHint( GLenum target, GLenum mode )
{
  glHint( target, mode );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glIndexd
/** This function asserts on the result before returning.  */
inline  void  GLIndexd( GLdouble c )
{
  glIndexd( c );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glIndexf
/** This function asserts on the result before returning.  */
inline  void  GLIndexf( GLfloat c )
{
  glIndexf( c );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glIndexi
/** This function asserts on the result before returning.  */
inline  void  GLIndexi( GLint c )
{
  glIndexi( c );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glIndexMask
/** This function asserts on the result before returning.  */
inline  void  GLIndexMask( GLuint mask )
{
  glIndexMask( mask );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glIndexPointer
/** This function asserts on the result before returning.  */
inline  void  GLIndexPointer( GLenum type, GLsizei stride, const GLvoid *pointer )
{
  glIndexPointer( type, stride, pointer );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glIndexs
/** This function asserts on the result before returning.  */
inline  void  GLIndexs( GLshort c )
{
  glIndexs( c );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glIndexub
/** This function asserts on the result before returning.  */
inline  void  GLIndexub( GLubyte c )
{
  glIndexub( c );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glInitNames
/** This function asserts on the result before returning.  */
inline  void  GLInitNames( void )
{
  glInitNames();
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glInterleavedArrays
/** This function asserts on the result before returning.  */
inline  void  GLInterleavedArrays( GLenum format, GLsizei stride, const GLvoid *pointer )
{
  glInterleavedArrays( format, stride, pointer );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glIsEnabled
/** This function returns the cached state of the queried capability. */
inline  GLboolean  GLIsEnabled( GLenum cap ) 
{
  GLboolean isEnabled = glIsEnabled( cap );
  ASSERT_GL_NO_ERROR;
  return isEnabled;
}

//! Replacement of glIsList
/** This function tests if for the current context, list denotes a valid display list. */
inline  GLboolean GLIsList( GLuint list )
{
  __ASSERT(glContext && glContext->m_glFunctions);
  __ASSERT(!!glContext->isDisplayList(list)==!!glIsList(list));
  return glContext->isDisplayList(list);
}

//! Replacement of glIsOcclusionQueryNV
/** This function tests if for the current context, id denotes a valid query. */
inline  GLboolean GLIsOcclusionQuery( GLuint id )
{
  __ASSERT(glContext->m_glFunctions && glContext->m_glFunctions->glIsOcclusionQueryNV);
  __ASSERT(!!glContext->isOcclusionQuery(id)==!!glContext->m_glFunctions->glIsOcclusionQueryNV(id));
  return glContext->m_glFunctions->glIsOcclusionQueryNV(id);
}

//! Replacement of glIsProgramNV
/** This function tests if for the current context, id denotes a valid program. */
inline  GLboolean GLIsProgram(GLuint id)
{
  __ASSERT(glContext->m_glFunctions && glContext->m_glFunctions->glIsProgramNV);
  __ASSERT(!!glContext->isProgram(id)==!!glContext->m_glFunctions->glIsProgramNV(id));
  return glContext->isProgram(id);
}

//! Replacement of glIsQueryARB
/** This function tests if for the current context, id denotes a valid query. */
inline  GLboolean GLIsQuery( GLuint id )
{
  __ASSERT(glContext->m_glFunctions && glContext->m_glFunctions->glIsQueryARB);
  __ASSERT(!!glContext->isQuery(id)==!!glContext->m_glFunctions->glIsQueryARB(id));
  return glContext->isQuery(id);
}

//! Replacement of glIsTexture
/** This function asserts on the result before returning.  */
inline  GLboolean GLIsTexture( GLuint texture )
{
  __ASSERT(glContext && glContext->m_glFunctions);
  __ASSERT(!!glContext->isTexture(texture)==!!glIsTexture(texture));
  return(glContext->isTexture(texture));
}

//! Replacement of glLightf
/** This function caches the light model settings and asserts on the result before returning.  */
inline  void  GLLightf( GLenum light, GLenum pname, GLfloat param ) 
{ 
  glLightf( light, pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glLightfv
/** This function caches the light settings and asserts on the result before returning.
  * \note GL_POSITION and GL_SPOT_DIRECTION are not cached, because they depend on the GL_MODELVIEW.  */
inline  void  GLLightfv( GLenum light, GLenum pname, const GLfloat *params ) 
{
  glLightfv( light, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glLighti
/** This function asserts on the result before returning.  */
inline  void  GLLighti( GLenum light, GLenum pname, GLint param ) 
{ 
  glLighti( light, pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glLightiv
/** This function asserts on the result before returning.  */
inline  void  GLLightiv( GLenum light, GLenum pname, GLint *params ) 
{ 
  glLightiv( light, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glLightModelf
/** This function caches the light model settings and asserts on the result before returning.  */
inline  void  GLLightModelf( GLenum pname, GLfloat param )  
{
  glLightModelf( pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glLightModelfv
/** This function caches the light model settings and asserts on the result before returning.  */
inline  void  GLLightModelfv( GLenum pname, const GLfloat *params ) 
{ 
  glLightModelfv( pname, params );
}

//! Replacement of glLightModeli
/** This function caches the light model settings and asserts on the result before returning.  */
inline  void  GLLightModeli( GLenum pname, GLint param )  
{ 
  glLightModeli( pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glLightModeliv
/** This function caches the light model settings and asserts on the result before returning.  */
inline  void  GLLightModeliv( GLenum pname, const GLint *params ) 
{ 
  glLightModeliv( pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glLineStipple
/** This function caches the line stipple settings and asserts on the result before returning.  */
inline  void  GLLineStipple( GLint factor, GLushort pattern )
{
  glLineStipple( factor, pattern );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glLineWidth
/** This function caches the line width setting and asserts on the result before returning.  */
inline  void  GLLineWidth( GLfloat width )
{
  glLineWidth( width );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glListBase
/** This function asserts on the result before returning.  */
inline  void  GLListBase( GLuint base )
{
  glListBase( base );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glLoadItentity
/** This function asserts on the result before returning.  */
inline  void  GLLoadIdentity( void )
{
  glLoadIdentity();
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glLoadMatrixd
/** This function asserts on the result before returning.  */
inline  void  GLLoadMatrixd( const GLdouble *m )
{
  glLoadMatrixd( m );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glLoadMatrixf
/** This function asserts on the result before returning.  */
inline  void  GLLoadMatrixf( const GLfloat *m )
{
  glLoadMatrixf( m );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glLoadName
/** This function asserts on the result before returning.  */
inline  void  GLLoadName( GLuint name )
{
  glLoadName( name );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glLoadProgramNV
/** This function hides the extension character of glLoadProgramNV and asserts on the result before returning.  */
inline  void  GLLoadProgram( GLenum target, GLuint id, GLsizei len, const GLubyte *program )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glLoadProgramNV );
  glContext->m_glFunctions->glLoadProgramNV( target, id, len, program );
#if defined( _DEBUG )
  ASSERT_GL_NO_ERROR;
  GLint errorPos;
  glGetIntegerv( GL_PROGRAM_ERROR_POSITION_NV, &errorPos );
  __ASSERT( errorPos == -1 );
#endif
}

//! Replacement of glLogicOp
/** This function asserts on the result before returning.  */
inline  void  GLLogicOp( GLenum opcode )
{
  glLogicOp( opcode );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glMap1d
/** This function asserts on the result before returning.  */
inline  void  GLMap1d( GLenum target, GLdouble u1, GLdouble u2, GLint stride, GLint order, const GLdouble *points )
{
  glMap1d( target, u1, u2, stride, order, points );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glMap1f
/** This function asserts on the result before returning.  */
inline  void  GLMap1f( GLenum target, GLfloat u1, GLfloat u2, GLint stride, GLint order, const GLfloat *points )
{
  glMap1f( target, u1, u2, stride, order, points );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glMap2d
/** This function asserts on the result before returning.  */
inline  void  GLMap2d( GLenum target, GLdouble u1, GLdouble u2, GLint ustride, GLint uorder, GLdouble v1, GLdouble v2, GLint vstride, GLint vorder, const GLdouble *points )
{
  glMap2d( target, u1, u2, ustride, uorder, v1, v2, vstride, vorder, points );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glMap2f
/** This function asserts on the result before returning.  */
inline  void  GLMap2f( GLenum target, GLfloat u1, GLfloat u2, GLint ustride, GLint uorder, GLfloat v1, GLfloat v2, GLint vstride, GLint vorder, const GLfloat *points )
{
  glMap2f( target, u1, u2, ustride, uorder, v1, v2, vstride, vorder, points );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glMapGrid1d
/** This function asserts on the result before returning.  */
inline  void  GLMapGrid1d( GLint un, GLdouble u1, GLdouble u2 )
{
  glMapGrid1d( un, u1, u2 );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glMapGrid1f
/** This function asserts on the result before returning.  */
inline  void  GLMapGrid1f( GLint un, GLfloat u1, GLfloat u2 )
{
  glMapGrid1f( un, u1, u2 );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glMapGrid2d
/** This function asserts on the result before returning.  */
inline  void  GLMapGrid2d( GLint un, GLdouble u1, GLdouble u2, GLint vn, GLdouble v1, GLdouble v2 )
{
  glMapGrid2d( un, u1, u2, vn, v1, v2 );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glMapGrid2f
/** This function asserts on the result before returning.  */
inline  void  GLMapGrid2f( GLint un, GLfloat u1, GLfloat u2, GLint vn, GLfloat v1, GLfloat v2 )
{
  glMapGrid2f( un, u1, u2, vn, v1, v2 );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glMaterialf
/** This function caches the material settings and asserts on the result before returning.  */
inline  void  GLMaterialf( GLenum face, GLenum pname, const GLfloat param ) 
{
  glMaterialf( face, pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glMaterialfv
/** This function caches the material settings and asserts on the result before returning.  */
inline  void  GLMaterialfv( GLenum face, GLenum pname, const GLfloat *params )  
{ 
  glMaterialfv( face, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glMateriali
/** This function caches the material settings and asserts on the result before returning.  */
inline  void  GLMateriali( GLenum face, GLenum pname, const GLint param ) 
{ 
  glMateriali( face, pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glMaterialiv
/** This function caches the material settings and asserts on the result before returning.  */
inline  void  GLMaterialiv( GLenum face, GLenum pname, const GLint *params )
{
  glMaterialiv( face, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glMatrixMode
/** This function caches the matrix mode setting and asserts on the result before returning.  */
inline  void  GLMatrixMode( GLenum mode )
{
  glMatrixMode( mode );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glMultiTexCoord2f
/** This function hides the extension character of glMultiTexCoord2f and asserts on the result before returning.  */
inline  void  GLMultiTexCoord2f( GLenum texture, GLfloat tx, GLfloat ty )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glMultiTexCoord2f );
  glContext->m_glFunctions->glMultiTexCoord2f( texture, tx, ty );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glMultMatrixd
/** This function asserts on the result before returning.  */
inline  void  GLMultMatrixd( const GLdouble *m )
{
  glMultMatrixd( m );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glMultMatrixf
/** This function asserts on the result before returning.  */
inline  void  GLMultMatrixf( const GLfloat *m )
{
  glMultMatrixf( m );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glNewList
/** This function asserts on the result before returning.  */
inline  void  GLNewList( GLuint list, GLenum mode )
{
  glNewList( list, mode );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glNormal3b
/** This function asserts on the result before returning.  */
inline  void  GLNormal3b( GLbyte nx, GLbyte ny, GLbyte nz )
{
  glNormal3b( nx, ny, nz );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glNormal3bv
/** This function asserts on the result before returning.  */
inline  void  GLNormal3bv( const GLbyte *v )
{
  glNormal3bv( v );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glNormal3d
/** This function asserts on the result before returning.  */
inline  void  GLNormal3d( GLdouble nx, GLdouble ny, GLdouble nz )
{
  glNormal3d( nx, ny, nz );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glNormal3dv
/** This function asserts on the result before returning.  */
inline  void  GLNormal3dv( const GLdouble *v )
{
  glNormal3dv( v );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glNormal3f
/** This function asserts on the result before returning.  */
inline  void  GLNormal3f( GLfloat nx, GLfloat ny, GLfloat nz )
{
  glNormal3f( nx, ny, nz );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glNormal3fv
/** This function asserts on the result before returning.  */
inline  void  GLNormal3fv( const GLfloat *v )
{
  glNormal3fv( v );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glNormal3i
/** This function asserts on the result before returning.  */
inline  void  GLNormal3i( GLint nx, GLint ny, GLint nz )
{
  glNormal3i( nx, ny, nz );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glNormal3iv
/** This function asserts on the result before returning.  */
inline  void  GLNormal3iv( const GLint *v )
{
  glNormal3iv( v );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glNormal3s
/** This function asserts on the result before returning.  */
inline  void  GLNormal3s( GLshort nx, GLshort ny, GLshort nz )
{
  glNormal3s( nx, ny, nz );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glNormal3sv
/** This function asserts on the result before returning.  */
inline  void  GLNormal3sv( const GLshort *v )
{
  glNormal3sv( v );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glNormalPointer
/** This function caches the normal pointer setings and asserts on the result before returning.  */
inline  void  GLNormalPointer( GLenum type, GLsizei stride, const GLvoid *pointer )
{
  glNormalPointer( type, stride, pointer );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glOrtho
/** This function asserts on the result before returning.  */
inline  void  GLOrtho( GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble zNear, GLdouble zFar )
{
  glOrtho( left, right, bottom, top, zNear, zFar );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPassThrough
/** This function asserts on the result before returning.  */
inline  void  GLPassThrough( GLfloat token )
{
  glPassThrough( token );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPixelMapufv
/** This function asserts on the result before returning.  */
inline  void  GLPixelMapfv( GLenum map, GLsizei mapsize, const GLfloat *values )
{
  glPixelMapfv( map, mapsize, values );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPixelMapuiv
/** This function asserts on the result before returning.  */
inline  void  GLPixelMapuiv( GLenum map, GLsizei mapsize, const GLuint *values )
{
  glPixelMapuiv( map, mapsize, values );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPixelMapusv
/** This function asserts on the result before returning.  */
inline  void  GLPixelMapusv( GLenum map, GLsizei mapsize, const GLushort *values )
{
  glPixelMapusv( map, mapsize, values );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPixelStoref
/** This function asserts on the result before returning.  */
inline  void  GLPixelStoref( GLenum pname, GLfloat param )
{
  glPixelStoref( pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPixelStorei
/** This function asserts on the result before returning.  */
inline  void  GLPixelStorei( GLenum pname, GLint param )
{
  glPixelStorei( pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPixelTransferf
/** This function asserts on the result before returning.  */
inline  void  GLPixelTransferf( GLenum pname, GLfloat param )
{
  glPixelTransferf( pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPixelTransferi
/** This function asserts on the result before returning.  */
inline  void  GLPixelTransferi( GLenum pname, GLint param )
{
  glPixelTransferi( pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPixelZoom
/** This function asserts on the result before returning.  */
inline  void  GLPixelZoom( GLfloat xfactor, GLfloat yfactor )
{
  glPixelZoom( xfactor, yfactor );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPointSize
/** This function caches the point size setting and asserts on the result before returning.  */
inline  void  GLPointSize( GLfloat size )
{
  glPointSize( size );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPolygonMode
/** This function caches the polygon mode setting and asserts on the result before returning.  */
inline  void  GLPolygonMode( GLenum face, GLenum mode ) 
{
  glPolygonMode( face, mode );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPolygonOffset
/** This function caches the polygon offset settings and asserts on the result before returning.  */
inline  void  GLPolygonOffset( GLfloat factor, GLfloat units )
{
  glPolygonOffset( factor, units );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPolygonStipple
/** This function asserts on the result before returning.  */
inline  void  GLPolygonStipple( const GLubyte *mask )
{
  glPolygonStipple( mask );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPopAttrib
/** This function asserts on the result before returning.  */
inline  void  GLPopAttrib( void )
{
  glPopAttrib();
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPopClientAttrib
/** This function asserts on the result before returning.  */
inline  void  GLPopClientAttrib( void )
{
  glPopClientAttrib();
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPopMatrix
/** This function asserts on the result before returning.  */
inline  void  GLPopMatrix( void )
{
  glPopMatrix();
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPopName
/** This function asserts on the result before returning.  */
inline  void  GLPopName( void )
{
  glPopName();
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPrioritizeTextures
/** This function asserts on the result before returning.  */
inline  void  GLPrioritizeTextures( GLsizei n, const GLuint *textures, const GLclampf *priorities )
{
  glPrioritizeTextures( n, textures, priorities );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glProgramNamedParameter4fNV
/** This function hides the extension character of glProgramNamedParameter4fNV and asserts on the result before returning.  */
inline  void  GLProgramNamedParameter4f( GLuint id, GLsizei len, const GLubyte *name, GLfloat x, GLfloat y, GLfloat z, GLfloat w )
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glProgramNamedParameter4fNV );
  glContext->m_glFunctions->glProgramNamedParameter4fNV( id, len, name, x, y, z, w );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPushAttrib
/** This function asserts on the result before returning.  */
inline  void  GLPushAttrib( GLbitfield mask )
{
  glPushAttrib( mask );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPushClientAttrib
/** This function asserts on the result before returning.  */
inline  void  GLPushClientAttrib( GLbitfield mask )
{
  glPushClientAttrib( mask );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPushMatrix
/** This function asserts on the result before returning.  */
inline  void  GLPushMatrix( void )
{
  glPushMatrix();
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glPushName
/** This function asserts on the result before returning.  */
inline  void  GLPushName( GLuint name )
{
  glPushName( name );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRasterPos2d
/** This function asserts on the result before returning.  */
inline  void  GLRasterPos2d( GLdouble x, GLdouble y )
{
  glRasterPos2d( x, y );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRasterPos2f
/** This function asserts on the result before returning.  */
inline  void  GLRasterPos2f( GLfloat x, GLfloat y )
{
  glRasterPos2f( x, y );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRasterPos2i
/** This function asserts on the result before returning.  */
inline  void  GLRasterPos2i( GLint x, GLint y )
{
  glRasterPos2i( x, y );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRasterPos2s
/** This function asserts on the result before returning.  */
inline  void  GLRasterPos2s( GLshort x, GLshort y )
{
  glRasterPos2s( x, y );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRasterPos3d
/** This function asserts on the result before returning.  */
inline  void  GLRasterPos3d( GLdouble x, GLdouble y, GLdouble z )
{
  glRasterPos3d( x, y, z );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRasterPos3f
/** This function asserts on the result before returning.  */
inline  void  GLRasterPos3f( GLfloat x, GLfloat y, GLfloat z )
{
  glRasterPos3f( x, y, z );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRasterPos3i
/** This function asserts on the result before returning.  */
inline  void  GLRasterPos3i( GLint x, GLint y, GLint z )
{
  glRasterPos3i( x, y, z );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRasterPos3s
/** This function asserts on the result before returning.  */
inline  void  GLRasterPos3s( GLshort x, GLshort y, GLshort z )
{
  glRasterPos3s( x, y, z );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRasterPos4d
/** This function asserts on the result before returning.  */
inline  void  GLRasterPos4d( GLdouble x, GLdouble y, GLdouble z, GLdouble w )
{
  glRasterPos4d( x, y, z, w );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRasterPos4f
/** This function asserts on the result before returning.  */
inline  void  GLRasterPos4f( GLfloat x, GLfloat y, GLfloat z, GLfloat w )
{
  glRasterPos4f( x, y, z, w );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRasterPos4i
/** This function asserts on the result before returning.  */
inline  void  GLRasterPos4i( GLint x, GLint y, GLint z, GLint w )
{
  glRasterPos4i( x, y, z, w );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRasterPos4s
/** This function asserts on the result before returning.  */
inline  void  GLRasterPos4s( GLshort x, GLshort y, GLshort z, GLshort w )
{
  glRasterPos4s( x, y, z, w );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glReadBuffer
/** This function asserts on the result before returning.  */
inline  void  GLReadBuffer( GLenum mode ) 
{ 
  glReadBuffer( mode );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glReadPixels
/** This function asserts on the result before returning.  */
inline  void  GLReadPixels( GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid *pixels )
{
  glReadPixels( x, y, width, height, format, type, pixels );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRectd
/** This function asserts on the result before returning.  */
inline  void  GLRectd( GLdouble x1, GLdouble y1, GLdouble x2, GLdouble y2 )
{
  glRectd( x1, y1, x2, y2 );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRectf
/** This function asserts on the result before returning.  */
inline  void  GLRectf( GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2 )
{
  glRectf( x1, y1, x2, y2 );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRecti
/** This function asserts on the result before returning.  */
inline  void  GLRecti( GLint x1, GLint y1, GLint x2, GLint y2 )
{
  glRecti( x1, y1, x2, y2 );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRects
/** This function asserts on the result before returning.  */
inline  void  GLRects( GLshort x1, GLshort y1, GLshort x2, GLshort y2 )
{
  glRects( x1, y1, x2, y2 );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRenderMode
/** This function asserts on the result before returning.  */
inline  void  GLRenderMode( GLenum mode )
{
  glRenderMode( mode );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRotated
/** This function asserts on the result before returning.  */
inline  void  GLRotated( GLdouble angle, GLdouble x, GLdouble y, GLdouble z )
{
  glRotated( angle, x, y, z );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glRotatef
/** This function asserts on the result before returning.  */
inline  void  GLRotatef( GLfloat angle, GLfloat x, GLfloat y, GLfloat z )
{
  glRotatef( angle, x, y, z );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glScaled
/** This function asserts on the result before returning.  */
inline  void  GLScaled( GLdouble x, GLdouble y, GLdouble z )
{
  glScaled( x, y, z );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glScalef
/** This function asserts on the result before returning.  */
inline  void  GLScalef( GLfloat x, GLfloat y, GLfloat z )
{
  glScalef( x, y, z );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glScissor
/** This function caches the scissor settings and asserts on the result before returning.  */
inline  void  GLScissor( GLint x, GLint y, GLsizei width, GLsizei height )
{
  glScissor( x, y, width, height );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glSecondaryColorPointer
/** This function caches the secondary color pointer setting and asserts on the result before returning.  */
inline  void  GLSecondaryColorPointer( GLint size, GLenum type, GLsizei stride, const GLvoid *pointer )
{
  glContext->m_glFunctions->glSecondaryColorPointer( size, type, stride, pointer );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glSelectBuffer
/** This function asserts on the result before returning.  */
inline  void  GLSelectBuffer( GLsizei size, GLuint *buffer )
{
  glSelectBuffer( size, buffer );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glSetWindowStereoModeNV
/** This function hides the extension character of glSetWindowStereoModeNV and asserts on the result before returning.  */
inline  void  GLSetWindowStereoMode( GLboolean displayMode )
{
  __ASSERT( glContext->m_glFunctions );
  if ( glContext->m_glFunctions->glSetWindowStereoModeNV )
  {
    glContext->m_glFunctions->glSetWindowStereoModeNV( displayMode );
    ASSERT_GL_NO_ERROR;
  }
}

//! Replacement of glShadeModel
/** This function caches the shade model setting and asserts on the result before returning.  */
inline  void  GLShadeModel( GLenum mode )
{
  glShadeModel( mode );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glStencilFunc
/** This function caches the stencil fuctnion setting and asserts on the result before returning.  */
inline  void  GLStencilFunc( GLenum func, GLint ref, GLuint mask )
{
  glStencilFunc( func, ref, mask );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glStencilMask
/** This function caches the stencil mask setting and asserts on the result before returning.  */
inline  void  GLStencilMask( GLuint mask )
{
  glStencilMask( mask );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glStencilOp
/** This function caches the stencil operation setting and asserts on the result before returning.  */
inline  void  GLStencilOp( GLenum fail, GLenum zfail, GLenum zpass )
{
  glStencilOp( fail, zfail, zpass );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexCoord1d
/** This function asserts on the result before returning.  */
inline  void  GLTexCoord1d( GLdouble s )
{
  glTexCoord1d( s );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glTexCoord1f
/** This function asserts on the result before returning.  */
inline  void  GLTexCoord1f( GLfloat s )
{
  glTexCoord1f( s );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glTexCoord1i
/** This function asserts on the result before returning.  */
inline  void  GLTexCoord1i( GLint s )
{
  glTexCoord1i( s );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glTexCoord1s
/** This function asserts on the result before returning.  */
inline  void  GLTexCoord1s( GLshort s )
{
  glTexCoord1s( s );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glTexCoord2d
/** This function asserts on the result before returning.  */
inline  void  GLTexCoord2d( GLdouble s, GLdouble t )
{
  glTexCoord2d( s, t );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glTexCoord2f
/** This function asserts on the result before returning.  */
inline  void  GLTexCoord2f( GLfloat s, GLfloat t )
{
  glTexCoord2f( s, t );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glTexCoord2fv
/** This function asserts on the result before returning.  */
inline  void  GLTexCoord2fv( const GLfloat *v )
{
  glTexCoord2fv( v );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glTexCoord2i
/** This function asserts on the result before returning.  */
inline  void  GLTexCoord2i( GLint s, GLint t )
{
  glTexCoord2i( s, t );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glTexCoord2s
/** This function asserts on the result before returning.  */
inline  void  GLTexCoord2s( GLshort s, GLshort t )
{
  glTexCoord2s( s, t );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glTexCoord3d
/** This function asserts on the result before returning.  */
inline  void  GLTexCoord3d( GLdouble s, GLdouble t, GLdouble r )
{
  glTexCoord3d( s, t, r );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glTexCoord3f
/** This function asserts on the result before returning.  */
inline  void  GLTexCoord3f( GLfloat s, GLfloat t, GLfloat r )
{
  glTexCoord3f( s, t, r );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glTexCoord3i
/** This function asserts on the result before returning.  */
inline  void  GLTexCoord3i( GLint s, GLint t, GLint r )
{
  glTexCoord3i( s, t, r );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glTexCoord3s
/** This function asserts on the result before returning.  */
inline  void  GLTexCoord3s( GLshort s, GLshort t, GLshort r )
{
  glTexCoord3s( s, t, r );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glTexCoord4d
/** This function asserts on the result before returning.  */
inline  void  GLTexCoord4d( GLdouble s, GLdouble t, GLdouble r, GLdouble q )
{
  glTexCoord4d( s, t, r, q );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glTexCoord4f
/** This function asserts on the result before returning.  */
inline  void  GLTexCoord4f( GLfloat s, GLfloat t, GLfloat r, GLfloat q )
{
  glTexCoord4f( s, t, r, q );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glTexCoord4i
/** This function asserts on the result before returning.  */
inline  void  GLTexCoord4i( GLint s, GLint t, GLint r, GLint q )
{
  glTexCoord4i( s, t, r, q );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glTexCoord4s
/** This function asserts on the result before returning.  */
inline  void  GLTexCoord4s( GLshort s, GLshort t, GLshort r, GLshort q )
{
  glTexCoord4s( s, t, r, q );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glTexCoordPointer
/** This function caches the tex coord pointer setting and asserts on the result before returning.  */
inline  void  GLTexCoordPointer( GLint size, GLenum type, GLsizei stride, const GLvoid *pointer )
{
  glTexCoordPointer( size, type, stride, pointer );    
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexEnvf
/** This function asserts on the result before returning.  */
inline  void  GLTexEnvf( GLenum target, GLenum pname, GLfloat param )
{
  glTexEnvf( target, pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexEnvfv
/** This function asserts on the result before returning.  */
inline  void  GLTexEnvfv( GLenum target, GLenum pname, const GLfloat *params )
{
  glTexEnvfv( target, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexEnvi
/** This function asserts on the result before returning.  */
inline  void  GLTexEnvi( GLenum target, GLenum pname, GLint param )
{
  glTexEnvi( target, pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexEnviv
/** This function asserts on the result before returning.  */
inline  void  GLTexEnviv( GLenum target, GLenum pname, const GLint *params )
{
  glTexEnviv( target, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexGend
/** This function asserts on the result before returning.  */
inline  void  GLTexGend( GLenum coord, GLenum pname, GLdouble param )
{
  glTexGend( coord, pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexGenf
/** This function asserts on the result before returning.  */
inline  void  GLTexGenf( GLenum coord, GLenum pname, GLfloat param )
{
  glTexGenf( coord, pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexGenfv
/** This function asserts on the result before returning.  */
inline  void  GLTexGenfv( GLenum coord, GLenum pname, const GLfloat *param )
{
  glTexGenfv( coord, pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexGeni
/** This function asserts on the result before returning.  */
inline  void  GLTexGeni( GLenum coord, GLenum pname, GLint param )
{
  glTexGeni( coord, pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexImage1D
/** This function asserts on the result before returning.  */
inline  void  GLTexImage1D( GLenum target, GLint level, GLint internalformat, GLsizei width, GLint border, GLenum format, GLenum type, const GLvoid *pixels )
{
  glTexImage1D( target, level, internalformat, width, border, format, type, pixels );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexImage2D
/** This function asserts on the result before returning.  */
inline  void  GLTexImage2D( GLenum target, GLint level, GLint internalformat, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const GLvoid *pixels )
{
  glTexImage2D( target, level, internalformat, width, height, border, format, type, pixels );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexImage3D
/** This function hides the extension character of glTexImage3D and asserts on the result before returning.  */
inline  void  GLTexImage3D( GLenum target, int level, GLenum internalformat, GLsizei width, GLsizei height, GLsizei depth, int border, GLenum format, GLenum type, const void* pixels)
{
  __ASSERT( glContext->m_glFunctions && glContext->m_glFunctions->glTexImage3D );
  glContext->m_glFunctions->glTexImage3D( target, level, internalformat, width, height, depth, border, format, type, pixels );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexParameterf
/** This function asserts on the result before returning.  */
inline  void  GLTexParameterf( GLenum target, GLenum pname, GLfloat param )
{
  glTexParameterf( target, pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexParameteri
/** This function asserts on the result before returning.  */
inline  void  GLTexParameteri( GLenum target, GLenum pname, GLint param )
{
  glTexParameteri( target, pname, param );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexParameterfv
/** This function asserts on the result before returning.  */
inline  void  GLTexParameterfv( GLenum target, GLenum pname, const GLfloat *params )
{
  glTexParameterfv( target, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexParameteriv
/** This function asserts on the result before returning.  */
inline  void  GLTexParameteriv( GLenum target, GLenum pname, const GLint *params )
{
  glTexParameteriv( target, pname, params );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexSubImate1D
/** This function asserts on the result before returning.  */
inline  void  GLTexSubImage1D( GLenum target, GLint level, GLint xoffset, GLsizei width, GLenum format, GLenum type, const GLvoid *pixels )
{
  glTexSubImage1D( target, level, xoffset, width, format, type, pixels );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTexSubImate2D
/** This function asserts on the result before returning.  */
inline  void  GLTexSubImage2D( GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels )
{
  glTexSubImage2D( target, level, xoffset, yoffset, width, height, format, type, pixels );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTranslated
/** This function asserts on the result before returning.  */
inline  void  GLTranslated( GLdouble x, GLdouble y, GLdouble z )
{
  glTranslated( x, y, z );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glTranslatef
/** This function asserts on the result before returning.  */
inline  void  GLTranslatef( GLfloat x, GLfloat y, GLfloat z )
{
  glTranslatef( x, y, z );
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glVertex2d
/** This function asserts on the result before returning.  */
inline  void  GLVertex2d( GLdouble x, GLdouble y )
{
  glVertex2d( x, y );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glVertex2f
/** This function asserts on the result before returning.  */
inline  void  GLVertex2f( GLfloat x, GLfloat y )
{
  glVertex2f( x, y );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glVertex2i
/** This function asserts on the result before returning.  */
inline  void  GLVertex2i( GLint x, GLint y )
{
  glVertex2i( x, y );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glVertex2s
/** This function asserts on the result before returning.  */
inline  void  GLVertex2s( GLshort x, GLshort y )
{
  glVertex2s( x, y );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glVertex3d
/** This function asserts on the result before returning.  */
inline  void  GLVertex3d( GLdouble x, GLdouble y, GLdouble z )
{
  glVertex3d( x, y, z );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glVertex3f
/** This function asserts on the result before returning.  */
inline  void  GLVertex3f( GLfloat x, GLfloat y, GLfloat z )
{
  glVertex3f( x, y, z );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glVertex3fv
/** This function asserts on the result before returning.  */
inline  void  GLVertex3fv( const GLfloat *v )
{
  glVertex3fv( v );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glVertex3i
/** This function asserts on the result before returning.  */
inline  void  GLVertex3i( GLint x, GLint y, GLint z )
{
  glVertex3i( x, y, z );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glVertex3s
/** This function asserts on the result before returning.  */
inline  void  GLVertex3s( GLshort x, GLshort y, GLshort z )
{
  glVertex3s( x, y, z );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glVertex4d
/** This function asserts on the result before returning.  */
inline  void  GLVertex4d( GLdouble x, GLdouble y, GLdouble z, GLdouble w )
{
  glVertex4d( x, y, z, w );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glVertex4f
/** This function asserts on the result before returning.  */
inline  void  GLVertex4f( GLfloat x, GLfloat y, GLfloat z, GLfloat w )
{
  glVertex4f( x, y, z, w );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glVertex4i
/** This function asserts on the result before returning.  */
inline  void  GLVertex4i( GLint x, GLint y, GLint z, GLint w )
{
  glVertex4i( x, y, z, w );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glVertex4s
/** This function asserts on the result before returning.  */
inline  void  GLVertex4s( GLshort x, GLshort y, GLshort z, GLshort w )
{
  glVertex4s( x, y, z, w );
  ASSERT_GL_NO_ERROR_IN_BEGIN_END;
}

//! Replacement of glVertexPointer
/** This function caches the vertex pointer setting and asserts on the result before returning.  */
inline  void  GLVertexPointer( GLint size, GLenum type, GLsizei stride, const GLvoid *pointer )
{
  glVertexPointer( size, type, stride, pointer );    
  ASSERT_GL_NO_ERROR;
}

//! Replacement of glViewport
/** This function caches the viewport settings and asserts on the result before returning.  */
inline  void  GLViewport( GLint x, GLint y, GLsizei width, GLsizei height )
{
  glViewport( x, y, width, height );
  ASSERT_GL_NO_ERROR;
}

#if defined(_WIN32)
//! Replacement of wglBindTexImageARB
/** This function hides the extension character of wglBindTexImageARB and asserts on its result before returning. */
inline  BOOL  WGLBindTexImage( HPBUFFERARB hPbuffer, int iBuffer )
{
  __ASSERT( glContext->m_glFunctions->wglBindTexImageARB );
  return glContext->m_glFunctions->wglBindTexImageARB( hPbuffer, iBuffer );
}

//! Replacement of wglChoosePixelFormatARB
/** This function hides the extension character of wglChoosePixelFormatARB and asserts on its result before returning. */
inline  BOOL  WGLChoosePixelFormat( HDC hdc, const int *piAttribIList, const FLOAT *pfAttribFList, UINT nMaxFormats, int *piFormats, UINT *nNumFormats )
{
  __ASSERT( glContext->m_glFunctions->wglChoosePixelFormatARB );
  return glContext->m_glFunctions->wglChoosePixelFormatARB( hdc, piAttribIList, pfAttribFList, nMaxFormats, piFormats, nNumFormats );
}

//! Replacement of wglCreateContext.
/** If this function is used in conjunction with WGLCreateContext, WGLMakeCurrent, and the gl functions with
  * uppercase GL prefix, then the states and attributes are cached per context. Moreover, the extension functions of
  * OpenGL are used per context (as they really should!).
  */
inline  HGLRC WGLCreateContext( HDC hdc )
{
  HGLRC hglrc = wglCreateContext( hdc );
  WGLAttachContext(hdc, hglrc);
  return hglrc;
}

inline void WGLAttachContext(HDC hdc, HGLRC hglrc)
{
  __ASSERT(hdc);
  __ASSERT(hglrc);
  // there should be no entry for the new hglrc right now
  __ASSERT(hglrcContextMap.end()==hglrcContextMap.find(hglrc));
  hglrcContextMap.insert(std::make_pair(hglrc, new GLContext(hdc, hglrc))); 
}

inline void WGLDetachContext(HGLRC hglrc)
{
  std::map<HGLRC, GLContext*>::iterator it = hglrcContextMap.find( hglrc );
  if ( it != hglrcContextMap.end() )
  {
    __ASSERT( it->second );
    if ( glContext == it->second )
    {
      glContext = NULL;
    }
    it->second->exit();
    delete it->second;
    hglrcContextMap.erase( it );
  }
}

//! Replacement of wglCreatePbufferARB
/** This function hides the extension character of wglCreatePbufferARB. 
  * \note The creation of a PBuffer might fail dependent on available hardware resources. 
  * In this case a NULL handle will be returned. */
inline  HPBUFFERARB WGLCreatePbuffer( HDC hdc, int pixelFormat, int width, int height, const int *attribList )
{
  __ASSERT( wglGetCurrentContext() != NULL );
  __ASSERT( glContext->m_glFunctions->wglCreatePbufferARB );
  HPBUFFERARB hPbuffer = glContext->m_glFunctions->wglCreatePbufferARB( hdc, pixelFormat, width, height, attribList );
  return( hPbuffer );
}

//! Replacement of wglDeleteContext.
/** If this function is used in conjunction with WGLDeleteContext, WGLMakeCurrent, and functions with
  * uppercase 'GL' prefix, the states and attributes are cached per context. 
  * Moreover, OpenGL extension functions are used per context (as they really should!).
  */
inline  BOOL  WGLDeleteContext( HGLRC hglrc )
{
  WGLDetachContext(hglrc);
  return wglDeleteContext(hglrc);
}

//! Replacement of wglDestroyPbufferARB
/** This function hides the extension character of wglDestroyPbufferARB and asserts on its result before returning. */
inline  BOOL  WGLDestroyPbuffer( HPBUFFERARB hPbuffer )
{
  __ASSERT(glContext->m_glFunctions->wglDestroyPbufferARB);
  return glContext->m_glFunctions->wglDestroyPbufferARB( hPbuffer );
}

//! Replacement for wglGetCurrenContext
/** The function obtains a handle to the current OpenGL rendering context of the calling thread. 
  * \returns A handle to the current OpenGL context of the calling thread, or NULL if the calling thread 
  * has no current OpenGL context.
  */
inline  HGLRC  WGLGetCurrentContext(void)
{
  return wglGetCurrentContext();
}

//! Replacement for wglGetCurrentDC
/** The function obtains a handle to the device context that is associated with the current OpenGL rendering context 
  * of the calling thread. 
  * \returns A handle to the device context which is currently associated with the current OpenGL context of the 
  * calling thread by means of WGLMakeCurrent, or NULL if the calling thread has no current OpenGL context.
  */
inline  HDC  WGLGetCurrentDC(void)
{
  return wglGetCurrentDC();
}

//! Replacement of wglMakeCurrent.
/** If this function is used in conjunction with WGLCreateContext, WGLDeleteContext, and functions with
  * uppercase 'GL' prefix, the states and attributes are cached per context. 
  * Moreover, OpenGL extension functions are used per context (as they really should!).
  */
inline BOOL WGLMakeCurrent( HDC hdc, HGLRC hglrc )
{
  BOOL success = FALSE;

  if ( hglrc )
  {
    // lookup the context
    std::map<HGLRC, GLContext*>::iterator it = hglrcContextMap.find(hglrc);
    if ( it!=hglrcContextMap.end() )
    {
      if ( glContext == (*it).second )
      { // already current
        // should have been properly initialized if we get here!
        __ASSERT(glContext->m_glFunctions); 
        success = TRUE; // suceeded
      }
      else
      {
        success =  (hglrc==wglGetCurrentContext()) // bypassed?
                   // only necessary if not yet current
                   || wglMakeCurrent(hdc, hglrc);

        if ( success )
        { 
          glContext = (*it).second;
          if ( !glContext->m_glFunctions )
          { // was not initialized before
            glContext->init();
          }
        }
      }
    }
    else
    { // ERROR: hglrc was not attached yet 
      __ASSERT(!"ERROR: hglrc was not attached yet!");
      
      // Note: It is an error to call WGLMakeCurrent on a context hglrc!=NULL that has not been attached yet
      // either by a call to WGLCreateContext or WGLAttachContext. Hence, make the current context no longer
      // current (this is the same behaviour as with the original wglMakeCurrent).
      wglMakeCurrent(hdc, NULL);
      glContext = NULL;
    }
  }
  else 
  { // make the calling thread's current rendering context no longer current
    success = wglMakeCurrent(hdc, NULL);
    if ( success )
    { // hglrc was NULL, so the pointer to our context object should
      glContext = NULL;
    }
  }
  
  return success;
}

//! Replacement of wglGetPbufferDCARB
/** This function hides the extension character of wglGetPbufferDCARB and asserts on its result before returning. */
inline  HDC WGLGetPbufferDC( HPBUFFERARB hPbuffer )
{
  __ASSERT( glContext->m_glFunctions->wglGetPbufferDCARB );
  return glContext->m_glFunctions->wglGetPbufferDCARB( hPbuffer );
}

//! Replacement of wglQueryPbufferARB
/** This function hides the extension character of wglQueryPbufferARB and asserts on its result before returning. */
inline  BOOL  WGLQueryPbuffer( HPBUFFERARB hPbuffer, int attribute, int *value )
{
  __ASSERT( glContext->m_glFunctions->wglQueryPbufferARB );
  return glContext->m_glFunctions->wglQueryPbufferARB( hPbuffer, attribute, value );
}

//! Replacement of wglReleasePbufferDCARB
/** This function hides the extension character of wglReleasePbufferDCARB and asserts on its result before returning. */
inline  int WGLReleasePbufferDC( HPBUFFERARB hPbuffer, HDC hdc )
{
  __ASSERT( glContext->m_glFunctions->wglReleasePbufferDCARB );
  int i = glContext->m_glFunctions->wglReleasePbufferDCARB( hPbuffer, hdc );
  __ASSERT( i == 1 );
  return( i );
}

//! Replacement of wglReleaseTexImageARB
/** This function hides the extension character of wglReleaseTexImageARB and asserts on its result before returning. */
inline  BOOL  WGLReleaseTexImage( HPBUFFERARB hPbuffer, int iBuffer )
{
  __ASSERT( glContext->m_glFunctions->wglReleaseTexImageARB );
  return glContext->m_glFunctions->wglReleaseTexImageARB( hPbuffer, iBuffer );
}

//! Replacement of wglShareLists
/** This function asserts on the result of wglShareLists before returning. */
inline  BOOL  WGLShareLists( HGLRC hglrc1, HGLRC hglrc2 )
{
  if ( hglrc1 == hglrc2 )
  { // do nothing in case of identical contexts
    return TRUE;
  }

  std::map<HGLRC, GLContext*>::iterator it1 = hglrcContextMap.find(hglrc1);
  if ( it1 != hglrcContextMap.end() )
  {
    std::map<HGLRC, GLContext*>::iterator it2 = hglrcContextMap.find(hglrc2);
    if ( it2 != hglrcContextMap.end() )
    {
      (*it2).second->share(*(*it1).second);
    }
#ifdef _DEBUG
    else
    { // hglrc probably was created by bypassing the WGLCreateContext routine
      __ASSERT(!"hglrc2 probably was created by bypassing the WGLCreateContext routine!");
    }
#endif
  }
#ifdef _DEBUG
  else
  { // hglrc probably was created by bypassing the WGLCreateContext routine
    __ASSERT(!"hglrc1 probably was created by bypassing the WGLCreateContext routine!");
  }
#endif

  return wglShareLists( hglrc1, hglrc2 );
}

//! Replacement of wglJointSwapGroupNV
inline BOOL WGLJoinSwapGroupNV(HDC hDC, GLuint group)
{
  return glContext->m_glFunctions->wglJoinSwapGroupNV(hDC, group);
}

//! Replacement of wglBindSwapBarrierNV
inline BOOL WGLBindSwapBarrierNV(GLuint group, GLuint barrier)
{
  return glContext->m_glFunctions->wglBindSwapBarrierNV(group, barrier);
}

//! Replacement of wglQuerySwapGroupNV
inline BOOL WGLQuerySwapGroupNV(HDC hDC, GLuint * group, GLuint * barrier)
{
  return glContext->m_glFunctions->wglQuerySwapGroupNV(hDC, group, barrier);
}

//! Replacement of wglQueryMaxSwapGroupsNV
inline BOOL WGLQueryMaxSwapGroupsNV(HDC hDC, GLuint * maxGroups, GLuint * maxBarriers)
{
  return glContext->m_glFunctions->wglQueryMaxSwapGroupsNV(hDC, maxGroups, maxBarriers);
}

//! Replacement of wglQueryFrameCountNV
inline BOOL WGLQueryFrameCountNV(HDC hDC, GLuint * count)
{
  return glContext->m_glFunctions->wglQueryFrameCountNV(hDC, count);
}

//! Replacement of wglResetFrameCountNV
inline BOOL WGLResetFrameCountNV(HDC hDC)
{
  return glContext->m_glFunctions->wglResetFrameCountNV(hDC);
}

#elif defined(LINUX)
//! Replacement of glXCreateContext
inline GLXContext GLXCreateContext(Display *dpy, XVisualInfo *vis, GLXContext shareList, Bool direct)
{
  GLXContext ctx = glXCreateContext(dpy, vis, shareList, direct);
  GLXAttachContext(dpy, ctx);
  return ctx;
}

//! Replacement of glXDestroyContext
inline void GLXDestroyContext(Display *dpy, GLXContext ctx)
{
  GLXDetachContext(ctx);
  return glXDestroyContext(dpy, ctx);
}

inline void GLXAttachContext(Display * dpy, GLXContext ctx)
{
  __ASSERT(dpy);
  __ASSERT(ctx);
  // there should be no entry for the new hglrc right now
  __ASSERT(hglrcContextMap.end()==hglrcContextMap.find(ctx));
  hglrcContextMap.insert(std::make_pair(ctx, new GLContext(dpy, ctx))); 
}

inline void GLXDetachContext(GLXContext ctx)
{
  std::map<GLXContext, GLContext*>::iterator it = hglrcContextMap.find(ctx);
  if ( it != hglrcContextMap.end() )
  {
    __ASSERT((*it).second);
    (*it).second->exit();
    delete (*it).second;
    hglrcContextMap.erase(it);
  }
}

//! Replacement of glXGetCurrentContext
inline GLXContext GLXGetCurrentContext()
{
  return glXGetCurrentContext();
}

//! Replacement of glXGetCurrentDrawable
inline GLXDrawable GLXGetCurrentDrawable()
{
  return glXGetCurrentDrawable();
}

//! Replacement of glXMakeCurrent
inline Bool GLXMakeCurrent(Display *dpy, GLXDrawable drawable, GLXContext ctx)
{
  Bool success = False;

  if ( ctx )
  {
    // lookup the context
    std::map<GLXContext, GLContext*>::iterator it = hglrcContextMap.find(ctx);
    if ( it!=hglrcContextMap.end() )
    {
      if ( glContext == (*it).second )
      { // already current
        // should have been properly initialized if we get here!
        __ASSERT(glContext->m_glFunctions); 
        success = True; // suceeded
      }
      else
      {
        success =  (ctx==GLXGetCurrentContext()) // bypassed?
                   // only necessary if not yet current
                   || glXMakeCurrent(dpy, drawable, ctx);

        if ( success )
        { 
          glContext = (*it).second;
          if ( !glContext->m_glFunctions )
          { // was not initialized before
            glContext->init(drawable);
          }
        }
      }
    }
    else
    { // ERROR: ctx was not attached yet 
      __ASSERT(!"ERROR: ctx was not attached yet!");
      
      // Note: It is an error to call GLXMakeCurrent on a context ctx!=NULL that has not been attached yet
      // either by a call to GLXCreateContext or GLXAttachContext. Hence, make the current context no longer
      // current (this is the same behaviour as with the original glxMakeCurrent).
      glXMakeCurrent(dpy, None, NULL);
      glContext = NULL;
    }
  }
  else 
  { // make the calling thread's current rendering context no longer current
    success = glXMakeCurrent(dpy, None, NULL);
    if ( success )
    { // hglrc was NULL, so the pointer to our context object should
      glContext = NULL;
    }
  }
 
  return success;
}

#endif // LINUX

} // namespace nvgl
