// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include "nvmath/Mat44f.h"
#include "nvmath/Sphere3f.h"

namespace nvmath
{
  //! Sphere class that represents the bounding sphere of a set of points.
  /** Because determining the optimal bounding sphere of a set of points is rather cumbersome, this BoundingSphere is an
    * approximation that uses the baricenter of a set of points. 
    * On extending a BoundingSphere with another, the sphere that includes both and is tangent to them is used. */
  class BoundingSphere : public Sphere3f
  {
    public:
      //! Constructor
      BoundingSphere();

    public:
      //! Extend this BoundingSphere by a sphere.
      /** This BoundingSphere is extended to include the given sphere. If they don't include each other
        * the new sphere is tangent to both of them.
        * \returns \c true if the extended BoundingSphere is valid, otherwise \c false */
      NVSG_API bool extend( const Sphere3f &sphere    //!<  sphere to extend with
                          );

      //! Get the sphere.
      /** It is assumed that this BoundingSphere is valid.
        * \returns This BoundingSphere as a Sphere3f. */
      NVSG_API const Sphere3f & getSphere( void ) const;

      //! Set the sphere.
      /** Makes the BoundingSphere valid. */
      NVSG_API void setSphere( const Sphere3f &sphere );

      //! Get the center.
      /** It is assumed that this BoundingSphere is valid.
        * \returns The center of this BoundingSphere. */
      NVSG_API const Vec3f & getCenter( void ) const;

      //! Get the radius.
      /** It is assumed that this BoundingSphere is valid.
        * \returns The radius of this BoundingSphere. */
      NVSG_API float getRadius( void ) const;

      //! Invalidate the BoundingSphere.
      /** Marks this BoundingSphere as invalid. After this, neither center nor radius can be assumed to be valid.  */
      NVSG_API void invalidate( void );

      //! Determine if the BoundingSphere is valid.
      /** A BoundingSphere is valid if it has been set with at least two points.
        * \returns \c true if the BoundingSphere is valid, otherwise \c false */
      NVSG_API bool isValid( void ) const;

      //! Set the BoundingSphere to include a vector of points.
      /** A BoundingSphere of the given points is calculated by using the baricenter of the points as the center of the
        * BoundingSphere. */
      NVSG_API bool set( const Vec3f * vertices       //!<  points to be bounded by a sphere
                       , size_t        numVertices    //!<  number of points 
                       );

      //! Transform the BoundingSphere by a matrix.
      /** The center of the BoundingSphere is pre-multiplied by the given matrix. */
      NVSG_API void transform( const Mat44f &m    //!<  matrix to use in the transformation
                             );

      //! Enlarge the bounding sphere by a factor.
      NVSG_API void enlarge( float factor   //!<  factor to use for enlarging the bounding sphere
                           );

    private:
      bool    m_valid;
  };

  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // inlines
  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  inline  BoundingSphere::BoundingSphere()
  : m_valid(false)
  {
  }

  inline const Sphere3f & BoundingSphere::getSphere( void ) const
  {
    __ASSERT( m_valid );
    return( *this );
  }

  inline void BoundingSphere::setSphere( const Sphere3f &sphere )
  {
    m_valid = true;
    setCenter( sphere.getCenter() );
    setRadius( sphere.getRadius() );
  }

  inline const Vec3f & BoundingSphere::getCenter( void ) const
  {
    __ASSERT( m_valid );
    return( Sphere3f::getCenter() );
  }

  inline float BoundingSphere::getRadius( void ) const
  {
    __ASSERT( m_valid );
    return( Sphere3f::getRadius() );
  }

  inline  void  BoundingSphere::invalidate( void )
  {
    m_valid = false;
  }

  inline  bool  BoundingSphere::isValid( void ) const
  {
    return( m_valid );
  }

  inline  void  BoundingSphere::transform( const Mat44f &m )
  {
    __ASSERT( m_valid );
    setCenter( Vec3f( m * Vec4f( Sphere3f::getCenter(), 1.0f ) ) );
  }

  inline  void  BoundingSphere::enlarge( float factor )
  {
    __ASSERT( m_valid );
    setRadius( Sphere3f::getRadius() * factor );
  }

  inline bool  BoundingSphere::set( const Vec3f * vertices, size_t numVertices )
  {
    __ASSERT( numVertices > 1 );
    bool m_valid = false;
    for ( size_t i=1 ; !m_valid && i<numVertices ; i++ )
    {
      m_valid = ( FLT_EPSILON < distance( vertices[0], vertices[i] ) );
    }
    if ( m_valid )
    {
      setSphere( boundingSphere( vertices, numVertices ) );
    }
    return( m_valid );
  }

  inline bool  BoundingSphere::extend( const Sphere3f &sphere )
  {
    setSphere( m_valid ? boundingSphere( getSphere(), sphere ) : sphere );
    return( m_valid );
  }

} // end namspace nvmath
