// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include "nvmath/Matnnf.h"
#include "nvmath/Vec3f.h"

namespace nvmath
{
  class Quatf;

  //! 3x3 matrix class.
  /** This class is used to handle pure rotational and scaling transformations (no translations). */
  class Mat33f : public Matnnf<3>
  {
    public:
      //! Default constructor.
      /** For performance reasons no initialization is performed. */
      NVSG_API Mat33f( void );

      //! Constructor using 9 floats.
      /** The 9 floats are used as the matrix elements. No normalization is performed.  */
      NVSG_API Mat33f( float a00, float a01, float a02
                     , float a10, float a11, float a12
                     , float a20, float a21, float a22 );

      //! Constructor using three row vectors.
      /** The three vectors are used as the three rows of the matrix. No normalization is performed.  */
      NVSG_API Mat33f( const Vec3f &row0    //!<  row zero
                     , const Vec3f &row1    //!<  row one
                     , const Vec3f &row2    //!<  row two
                     );

      //! Constructor using a normalized rotation axis and angle (in radians).
      /** This results in a pure rotation matrix. */
      NVSG_API Mat33f( const Vec3f &axis    //!<  rotation axis
                     , float rad            //!<  angle to rotate (in radians)
                     );

      //! Constructor using a normalized quaternion.
      /** This results in a pure rotation matrix. */
      NVSG_API Mat33f( const Quatf &quat  //!<  quaternion for rotation
                     );

      //! Copy constructor
      NVSG_API Mat33f( const Matnnf<3> & m );

      //! Set the matrix with three row vectors.
      NVSG_API void set( const Vec3f &row0    //!<  row zero
                       , const Vec3f &row1    //!<  row one
                       , const Vec3f &row2    //!<  row two
                       );
  };

  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // non-member functions
  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  /*! \relates nvmath::Mat33f
   *  Decompose the matrix \a mat into scaling, shearing, and orientation */
  NVSG_API void decompose( const Mat33f &mat  //!<  matrix to decompose
                         , Vec3f &scaling     //!<  scaling factors
                         , Vec3f &shearing    //!<  shearing factors (xy, xz, yz)
                         , Quatf &orientation //!<  rotational part
                         );

  /*! \relates nvmath::Mat33f
   *  Determine the determinant of a matrix.
   *  \returns determinant of the matrix */
  inline float determinant( const Mat33f &mat  //!<  matrix to determine determinante of
                          )
  {
    return(   mat[0][0] * ( mat[1][1] * mat[2][2] - mat[1][2] * mat[2][1] )
            - mat[0][1] * ( mat[1][0] * mat[2][2] - mat[1][2] * mat[2][0] )
            + mat[0][2] * ( mat[1][0] * mat[2][1] - mat[1][1] * mat[2][0] ) );
  }

  /*! \relates nvmath::Mat33f
   *  Test if a matrix is a rotation.
   *  A matrix is a rotation if it is normalized, orthogonal, and the determinant equals (within some tolerance) 1.0.
   *  \returns true if the matrix is a rotation, otherwise false */
  inline bool isRotation( const Mat33f &mat     //!<  matrix to test
                        )
  {
    return(   isNormalized( mat )
          &&  isOrthogonal( mat )
          && ( fabs( determinant( mat ) - 1.0f ) <= FLT_EPSILON ) );
  }

  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // inlines
  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  inline Mat33f::Mat33f( void )
  {
  }

  inline Mat33f::Mat33f( float a00, float a01, float a02
                       , float a10, float a11, float a12
                       , float a20, float a21, float a22 )
  {
    (*this)[0][0] = a00;
    (*this)[0][1] = a01;
    (*this)[0][2] = a02;
    (*this)[1][0] = a10;
    (*this)[1][1] = a11;
    (*this)[1][2] = a12;
    (*this)[2][0] = a20;
    (*this)[2][1] = a21;
    (*this)[2][2] = a22;
  }

  inline Mat33f::Mat33f( const Vec3f &row0, const Vec3f &row1, const Vec3f &row2 )
  {
    set( row0, row1, row2 );
  }

  inline Mat33f::Mat33f( const Vec3f &axis, float rad )
  {
    __ASSERT( isNormalized( axis ) );
    float c = cosf( rad );
    float s = sinf( rad );
    __ASSERT( fabs( s * s + c * c - 1.0f ) <= FLT_EPSILON );
    float t = 1 - c;
    float x = axis[0];
    float y = axis[1];
    float z = axis[2];

    (*this)[0] = Vec3f( t * x * x + c,     t * x * y - s * z, t * x * z + s * y );
    (*this)[1] = Vec3f( t * x * y + s * z, t * y * y + c,     t * y * z - s * x );
    (*this)[2] = Vec3f( t * x * z - s * y, t * y * z + s * x, t * z * z + c     );

//    __ASSERT( isRotation( *this ) );
  }

  inline Mat33f::Mat33f( const Matnnf<3> &m )
    : Matnnf<3>( m )
  {
  }

  inline void Mat33f::set( const Vec3f &row0, const Vec3f &row1, const Vec3f &row2 )
  {
    (*this)[0]=row0;
    (*this)[1]=row1;
    (*this)[2]=row2;
  }

} // end namspace nvmath
