// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include "nvmath/Mat33f.h"
#include "nvmath/Vec4f.h"

namespace nvmath
{
  //! 4x4 matrix class.
  /** This class is used to handle full transformations (including rotations, scalings, translations, and projections).
    * \note Have a look into the nvmath namespace reference to find additional convenient 
    *       functions (e.g. decompose,...) */
  class Mat44f : public Matnnf<4>
  {
    public:
      //! Default constructor.
      /** For performance reasons no initialization is performed. */
      NVSG_API Mat44f( void );

      //! Constructor using 16 floats.
      /** The 16 floats are used as the matrix elements. No normalization is performed.  */
      NVSG_API Mat44f( float a00, float a01, float a02, float a03
                     , float a10, float a11, float a12, float a13
                     , float a20, float a21, float a22, float a23
                     , float a30, float a31, float a32, float a33 );

      //! Constructor using four row vectors.
      /** The four vectors are used as the four rows of the matrix. No normalization is performed.  */
      NVSG_API Mat44f( const Vec4f &row0    //!<  row zero
                     , const Vec4f &row1    //!<  row one
                     , const Vec4f &row2    //!<  row two
                     , const Vec4f &row3    //!<  row three
                     ); 

      //! Constructor by Translation, Rotation, and optional Scaling.
      /** This constructs a matrix that represents the concatenation T * R. */
      NVSG_API Mat44f( const Vec3f &trans                               //!<  translation
                     , const Quatf &rot                                 //!<  quaternion for rotation
                     );


      //! Copy constructor
      NVSG_API Mat44f( const Matnnf<4> & m );

      //! Set the matrix with four row vectors.
      NVSG_API void set( const Vec4f &row0    //!<  row zero
                       , const Vec4f &row1    //!<  row one
                       , const Vec4f &row2    //!<  row two
                       , const Vec4f &row3    //!<  row three
                       );
  };

  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // non-member functions
  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  /*! \relates nvmath::Mat44f
   *  Decompose the matrix \a mat into scaling, shearing, orientation, and translation. */
  inline  void  decompose( const Mat44f &mat    //!<  matrix to decompose
                         , Vec3f &scaling       //!<  scaling factors
                         , Vec3f &shearing      //!<  shearing factors (xy, xz, yz)
                         , Quatf &orientation   //!<  rotational part
                         , Vec3f &translation   //!<  translation vector
                         )
  {
    Mat33f  m33( mat[0][0], mat[0][1], mat[0][2],
                 mat[1][0], mat[1][1], mat[1][2],
                 mat[2][0], mat[2][1], mat[2][2] );
    decompose( m33, scaling, shearing, orientation );
    translation = Vec3f( mat[0][3], mat[1][3], mat[2][3] );
  }

  /*! \relates nvmath::Mat44f
   *  Matrix multiplication with a matrix (i.e. concatenation) operator.
   *  \returns multiplied (concatenated) matrix  */
  NVSG_API Mat44f operator*( const Mat44f &m0, const Mat44f &m1 );

  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // inlines
  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  inline Mat44f::Mat44f( void )
  {
  };

  inline Mat44f::Mat44f( float a00, float a01, float a02, float a03
                       , float a10, float a11, float a12, float a13
                       , float a20, float a21, float a22, float a23
                       , float a30, float a31, float a32, float a33 )
  {
    (*this)[0][0] = a00;
    (*this)[0][1] = a01;
    (*this)[0][2] = a02;
    (*this)[0][3] = a03;
    (*this)[1][0] = a10;
    (*this)[1][1] = a11;
    (*this)[1][2] = a12;
    (*this)[1][3] = a13;
    (*this)[2][0] = a20;
    (*this)[2][1] = a21;
    (*this)[2][2] = a22;
    (*this)[2][3] = a23;
    (*this)[3][0] = a30;
    (*this)[3][1] = a31;
    (*this)[3][2] = a32;
    (*this)[3][3] = a33;
  }

  inline Mat44f::Mat44f( const Vec4f &row0, const Vec4f &row1, const Vec4f &row2, const Vec4f &row3 )
  {
    set( row0, row1, row2, row3 );
  }

  inline Mat44f::Mat44f( const Vec3f &trans, const Quatf &ori )
  {
    Mat33f  m( ori );
    (*this)[0] = Vec4f( m[0], trans[0] );
    (*this)[1] = Vec4f( m[1], trans[1] );
    (*this)[2] = Vec4f( m[2], trans[2] );
    (*this)[3] = Vec4f( 0.0f, 0.0f, 0.0f, 1.0f );
  }

  inline Mat44f::Mat44f( const Matnnf<4> &m )
    : Matnnf<4>( m )
  {
  }

  inline void Mat44f::set( const Vec4f &row0, const Vec4f &row1, const Vec4f &row2, const Vec4f &row3 )
  {
    (*this)[0] = row0;
    (*this)[1] = row1;
    (*this)[2] = row2;
    (*this)[3] = row3;
  }

  //! global identity matrix.
  extern NVSG_API const Mat44f  cIdentity44f;

} // end namspace nvmath
