// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include  <algorithm>
#include  <vector>
#include  "nvmath/Vecnf.h"

// __WIN64__wa__XXX macros to workaround some win64 issues
#include "__WIN64__workarounds.h"

namespace nvmath
{
  class Sphere3f;
  class Vec4f;

  //! 3-component vector class.
  class Vec3f : public Vecnf<3>
  {
    public:
      //! Default constructor.
      /** For performance reasons no initialization is performed. */
      NVSG_API Vec3f( void );

      //! Constructor using three floats.
      /** A vector with the given values is constructed.  */
      NVSG_API Vec3f( float x   //!<  x component
                    , float y   //!<  y component
                    , float z   //!<  z component
                    );

      //! Copy constructor.
      NVSG_API Vec3f( const Vecnf<3> &v //!<  vector to copy
                    );

      //! Constructor by a Vec4f.
      /** Ignores fourth component of constructing vector.  */
      NVSG_API Vec3f( const Vecnf<4> &v //!<  vector to use first three components from
                    );

      //! Assignment operator.
      /** Ignores the fourth component of the assigning vector.
          \returns reference to the assigned vector */
      NVSG_API Vec3f& operator=( const Vec4f &v   //!<  copy \a v to the vector
                               );

      //! Negation operator.
      /** \returns negated vector */
      NVSG_API Vec3f operator-(void) const;
      // don't hide base class' operator-
      using Vecnf<3>::operator-;

      //! Set the vector with three float values.
      NVSG_API void set( float x    //!<  x component
                       , float y    //!<  y component
                       , float z    //!<  z component
                       );
  };

  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // non-member functions
  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  /*! \relates nvmath::Vec3f
   *  Calculate the normalized vector of a given one.
   *  \returns normalized version of \a v */
  inline Vec3f normalize( const Vec3f& v  //!<  vector to determine normalized vector from
                        )
  {
    float norm = __WIN64__wa__std_max( length(v), FLT_EPSILON );
    // return value optimization: safe construction/destruction of a temporary object 
    return( Vec3f( v[0]/norm, v[1]/norm, v[2]/norm ) );
  }

  /*! \relates nvmath::Vec3f
   *  Calculate the normalized vector of a vector of Vec3fs */
  inline  void  normalizeV( const std::vector<Vec3f> &vecIn, std::vector<Vec3f> &vecOut )
  {
    // std::transform (the MS implementation) is not intelligent enough. It does not reserve enough memory
    // to hold the transformed elements from input, but reallocates memory with every single element copied.
    // Reserving memory on our own will spare us the insert iterator here :-)
    vecOut.reserve( vecIn.size() );
    std::transform( vecIn.begin(), vecIn.end(), vecOut.begin(), &normalize );
  }

  /*! \relates nvmath::Vec3f
   *  Multiplication of a scalar with a vector.
   *  \returns product of \a f with \a v */
  inline Vec3f operator*( float f         //!<  scalar to multiply with
                        , const Vec3f &v  //!<  vector to multiply
                        )
  {
    return( v * f );
  }

  /*! \relates nvmath::Vec3f
   *  Calculation of the cross product of \a v0 with \a v1.
   *  The cross product of two vectors is a vector that is orthogonal to both arguments
   *  \returns cross product of \a v0 and \a v1. */
  inline Vec3f  operator^( const Vecnf<3> &v0, const Vecnf<3> &v1 )
  {
    return( Vec3f( v0[1] * v1[2] - v0[2] * v1[1],
                   v0[2] * v1[0] - v0[0] * v1[2],
                   v0[0] * v1[1] - v0[1] * v1[0]
                )
          );
  }

  /*! \relates nvmath::Vec3f
   *  Test two Vec3f for collinearity.
   *  The test is done by testing the cross product (operator^) of the two objects being less than some epsilon.
   *  \returns true if t0 and t1 are collinear, otherwise false */
  inline bool areCollinear( const Vec3f &v0   //!<  first Vec3f
                          , const Vec3f &v1   //!<  second Vec3f
                          )
  {
    return( length( v0 ^ v1 ) <= FLT_EPSILON );
  }

  /*! \relates nvmath::Vec3f
   *  Determine smoothed normals for a set of vertices. */
  NVSG_API void  smoothNormals( const std::vector<Vec3f> &vertices, const Sphere3f &sphere, float creaseAngle,
                                std::vector<Vec3f> &normals );

  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // inlines
  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  inline Vec3f::Vec3f( void )
  {
  }

  inline Vec3f::Vec3f( float x, float y, float z )
  {
    (*this)[0]=x;
    (*this)[1]=y;
    (*this)[2]=z;
  }

  inline Vec3f::Vec3f( const Vecnf<3> &v )
    : Vecnf<3>( v )
  {
  }

  inline Vec3f::Vec3f( const Vecnf<4> &v )
  {
    (*this)[0] = v[0];
    (*this)[1] = v[1];
    (*this)[2] = v[2];
  }

  inline void Vec3f::set( float x, float y, float z )
  {
    (*this)[0]=x;
    (*this)[1]=y;
    (*this)[2]=z;
  }

  inline Vec3f Vec3f::operator-(void) const
  {
    return (Vec3f(-m_vec[0], -m_vec[1], -m_vec[2]));
  }

} // end namspace nvmath
