// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include  "nvutil/Assert.h"
#include  <float.h>
#include  <math.h>

namespace nvmath
{
  //! Template for n-dimensional vector
  template<unsigned char n> class Vecnf
  {
    public:
      //! Default constructor.
      /** For performance reasons no initialization is performed. */
      Vecnf<n>( void );

      //! Copy constructor
      Vecnf<n>( const Vecnf<n> & v    //!< vector to copy
              );

      //! Non-constant subscript operator.
      /** \returns reference to \a i th element of the vector  */
      float& operator[]( size_t i           //!<  index to vector to use (i=0,..,n-1)
                       );

      //! Constant subscript operator.
      /** \returns constant reference to \a i th element of the vector */
      const float& operator[]( size_t i     //!<  index to vector to use (i=0,..,n-1)
                             ) const;

      //! Negation operator.
      /** \returns negated vector */
      Vecnf<n> operator-( void ) const;

      //! Vector multiplication with a scalar operator.
      /** \returns multiplied vector */
      Vecnf<n> operator*( float  f    //!<  scalar to multiply with
                        ) const;

      //! Vector multiplication with a vector (i.e. dot product) operator.
      /** \returns dot product of this with \a v */
      float operator*( const Vecnf<n> &v    //!<  vector to multiply with
                     ) const;

      //! Vector division by a scalar operator.
      /** \returns divided vector */
      Vecnf<n> operator/( float f   //!<  scalar to divide through
                        ) const;

      //! Vector addition operator.
      /** \returns sum of the two vectors  */
      Vecnf<n> operator+( const Vecnf<n> &v //!<  vector to add
                        ) const;

      //! Vector subtraction operator.
      /** \returns difference of the two vectors */
      Vecnf<n> operator-( const Vecnf<n> &v //!<  vector to subtract
                        ) const;

      //! Equality operator.
      /** \returns true if the vector is equal to \a v, otherwise false  */
      bool operator==( const Vecnf<n> &v    //!<  vector to compare with
                     ) const;

      //! Inequality operator.
      /** \returns true if the vector is not equal to \a v, otherwise false  */
      bool operator!=( const Vecnf<n> &v    //!< vector to compare with
                     ) const;

      //! Assignment operator.
      /** \returns reference to the assigned vector */
      Vecnf<n> & operator=( const Vecnf<n> &v   //!<  vector to copy
                          );

      //! Vector multiplication with a scalar and assignment operator.
      /** \returns reference to the assigned vector after multiplication */
      Vecnf<n> & operator*=( float f   //!<  scalar to multiply with
                           );

      //! Vector division by a scalar and assignment operator.
      /** \returns reference to the assigned vector after division */
      Vecnf<n> & operator/=( float f   //!<  scalar to divide by
                           );

      //! Vector addition and assignment operator.
      /** \returns  reference to the assigned vector after addition */
      Vecnf<n> & operator+=( const Vecnf<n> &v    //!<  vector to add
                           );

      //! Vector subtraction and assignment operator.
      /** \returns reference to the assigned vector after subtraction  */
      Vecnf<n> & operator-=( const Vecnf<n> &v    //!<  vector to subtract
                           );

      //! Get a constant pointer to the floats of the vector.
      /** \returns constant pointer to vector elements */
      const float* getPtr( void ) const;

      //! Normalize the vector and return the previous length.
      /** \returns length of the vector before normalization */
      float normalize( void );

      //! Orthogonolize the vector with respect to a vector.
      /** Sets the vector to be orthogonal to \a v. */
      void orthogonalize( const Vecnf<n> &v     //!<  vector to orthogonalize to
                        );

      //! Orthonormalize the vector with respect to an other.
      /** Sets the vector to be orthonormal to \a v.  */
      void orthonormalize( const Vecnf<n>& v    //!<  vector to orthonormalize to
                         );

    protected:
      float m_vec[n];
  };

  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // non-member functions
  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  /*! \relates nvmath::Vecnf
   *  Test if two vectors are orthogonal to each other.
   *  \returns true if the vectors are orthogonal, otherwise false  */
  template<unsigned char n> bool areOrthogonal( const Vecnf<n> &v0   //!<  first vector
                                              , const Vecnf<n> &v1   //!<  second vector
                                              )
  {
    return( fabs( v0 * v1 ) <= FLT_EPSILON );
  }

  /*! \relates nvmath::Vecnf
   *  Test if two vectors are orthonormal to each other.
   *  Two vectors are orthonormal if they are normalized and orthogonal to each other.
   *  \returns true if the vector is orthonormal to \a v, otherwise false */
  template<unsigned char n> bool areOrthonormal( const Vecnf<n> &v0   //!<  first vector
                                               , const Vecnf<n> &v1   //!<  second vector
                                               )
  {
    return( areOrthogonal( v0, v1 ) && isNormalized( v0 ) && isNormalized( v1 ) );
  }

  /*! \relates nvmath::Vecnf
   *  Test if two vectors are similar to each other.
   *  Two vectors are considered to be similar if they differ within a given tolerance in each component.
   *  \returns true if the vector and \a v differ less than \a tol in each component. */
  template<unsigned char n> bool areSimilar( const Vecnf<n> &v0       //!<  first vector
                                           , const Vecnf<n> &v1       //!<  second vector
                                           , float tol = FLT_EPSILON  //!<  maximal allowed tolerance for similarity
                                           )
  {
    __ASSERT( 0.0f <= tol );
    bool similar = true;
    for ( size_t i=0 ; similar && i<n ; i++ )
    {
      similar = ( fabs( v0[i] - v1[i] ) <= tol );
    }
    return( similar );
  }

  /*! \relates nvmath::Vecnf
   *  Determine the distance between two points.
   *  \returns The distance between \a p0 and \a p1 */
  template<unsigned char n> float distance( const Vecnf<n> &p0    //!<  first point
                                          , const Vecnf<n> &p1    //!<  second point
                                          )
  {
    return( length( p0 - p1 ) );
  }

  /*! \relates nvmath::Vecnf
   *  Test if a vector is normalized.
   *  A vector is normalized if its length is one.
   *  \returns true if the vector is normalized, otherwise false */
  template<unsigned char n> bool isNormalized( const Vecnf<n> &v    //!<  vector to test
                                             )
  {
    return( fabs( length( v ) - 1.0f ) < 2 * FLT_EPSILON );
  }

  /*! \relates nvmath::Vecnf
   *  Test if a vector is a null vector.
   *  A vector is a null vector if its length is zero.
   *  \returns true if the vector is a null vector, otherwise false */
  template<unsigned char n> bool isNull( const Vecnf<n> &v        //!<  vector to test
                                       )
  {
    return( length( v ) <= FLT_EPSILON );
  }

  /*! \relates nvmath::Vecnf
   *  Calculate the length of a vector.
   *  \returns length of \a v */
  template<unsigned char n> float length( const Vecnf<n> &v         //!<  vector to get length of
                                        )
  {
    return( sqrtf( lengthSquared( v ) ) );
  }

  /*! \relates nvmath::Vecnf
   *  Calculate the squared length of a vector.
   *  \returns squared length of \a v */
  template<unsigned char n> float lengthSquared( const Vecnf<n> &v  //!<  vector to get squared length of
                                               )
  {
    return( v * v );
  }

  /*! \relates nvmath::Vecnf
   *  Get the maximal absolute value of a vector.
   *  \returns maximal element of \a v */
  template<unsigned char n> float maxElement( const Vecnf<n> &v     //!<  vector to get maximal element of
                                            )
  {
    float f = fabsf( v[0] );
    for ( size_t i=1 ; i<n ; i++ )
    {
      float t = fabsf( v[i] );
      if ( f < t )
      {
        f = t;
      }
    }
    return( f );
  }

  /*! \relates nvmath::Vecnf
   *  Get the minimal element of a vector.
   *  \returns minimal element of \a v */
  template<unsigned char n> float minElement( const Vecnf<n> &v     //!<  vector to get maximal element of
                                            )
  {
    float f = v[0];
    for ( size_t i=1 ; i<n ; i++ )
    {
      if ( v[i] < f )
      {
        f = v[i];
      }
    }
    return( f );
  }

  /*! \relates nvmath::Vecnf
   *  Multiplication of a scalar with a vector.
   *  \returns product of \a f with \a v */
  template<unsigned char n> Vecnf<n>  operator*( float f            //!<  scalar to multiply with
                                               , const Vecnf<n> &v  //!<  vector to multiply
                                               )
  {
    return( v * f );
  }

  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // "inlines"
  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  template<unsigned char n> Vecnf<n>::Vecnf( void )
  {
  }

  template<unsigned char n> Vecnf<n>::Vecnf( const Vecnf<n> &v )
  {
    for ( size_t i=0 ; i<n ; i++ )
    {
      m_vec[i] = v[i];
    }
  }

  template<unsigned char n> float& Vecnf<n>::operator[]( size_t i ) 
  { 
    __ASSERT( 0 <= i && i < n );
    return( m_vec[i] );
  }

  template<unsigned char n> const float& Vecnf<n>::operator[]( size_t i ) const 
  { 
    __ASSERT( 0 <= i && i < n );
    return( m_vec[i] );
  }

  template<unsigned char n> Vecnf<n> Vecnf<n>::operator-( void ) const
  {
    Vecnf<n> v;
    for ( size_t i=0 ; i<n ; i++ )
    {
      v[i] = -m_vec[i];
    }
    return( v );
  }

  template<unsigned char n> Vecnf<n> Vecnf<n>::operator*( float f ) const
  {
    Vecnf<n> ret( *this );
    return( ret *= f );
  }

  template<unsigned char n> float Vecnf<n>::operator*( const Vecnf<n> &v ) const
  {
    float f = 0.0f;
    for ( size_t i=0 ; i<n ; i++ )
    {
      f += m_vec[i] * v[i];
    }
    return( f );
  }

  template<unsigned char n> Vecnf<n> Vecnf<n>::operator/( float f ) const
  {
    __ASSERT( fabs( f ) > FLT_EPSILON );
    Vecnf<n> ret( *this );
    return( ret /= f );
  }

  template<unsigned char n> Vecnf<n> Vecnf<n>::operator+( const Vecnf<n> &v ) const
  {
    Vecnf<n> ret( *this );
    return( ret += v );
  }

  template<unsigned char n> Vecnf<n> Vecnf<n>::operator-( const Vecnf<n> &v ) const
  {
    Vecnf<n> ret( *this );
    return( ret -= v );
  }

  template<unsigned char n> bool Vecnf<n>::operator==( const Vecnf<n> &v ) const
  {
    bool equal = true;
    for ( unsigned char i=0 ; equal && i<n ; i++ )
    {
      equal = ( fabs( m_vec[i] - v[i] ) < FLT_EPSILON );
    }
    return( equal );
  }

  template<unsigned char n> bool Vecnf<n>::operator!=( const Vecnf<n> &v ) const
  {
    return( ! ( *this == v ) );
  }

  template<unsigned char n> Vecnf<n> & Vecnf<n>::operator=( const Vecnf<n> &v )
  {
    for ( size_t i=0 ; i<n ; i++ )
    {
      m_vec[i] = v[i];
    }
    return( *this );
  }

  template<unsigned char n> Vecnf<n> & Vecnf<n>::operator*=( float f )
  {
    for ( size_t i=0 ; i<n ; i++ )
    {
      m_vec[i] *= f;
    }
    return( *this );
  }

  template<unsigned char n> Vecnf<n> & Vecnf<n>::operator/=( float f )
  {
    __ASSERT( fabs( f ) > FLT_EPSILON );
    for ( size_t i=0 ; i<n ; i++ )
    {
      m_vec[i] /= f;
    }
    return( *this );
  }

  template<unsigned char n> Vecnf<n> & Vecnf<n>::operator+=( const Vecnf<n> &v )
  {
    for ( size_t i=0 ; i<n ; i++ )
    {
      m_vec[i] += v[i];
    }
    return( *this );
  }

  template<unsigned char n> Vecnf<n> & Vecnf<n>::operator-=( const Vecnf<n> &v )
  {
    for ( size_t i=0 ; i<n ; i++ )
    {
      m_vec[i] -= v[i];
    }
    return( *this );
  }

  template<unsigned char n> const float* Vecnf<n>::getPtr( void ) const
  {
    return( &m_vec[0] );
  }

  template<unsigned char n> float Vecnf<n>::normalize( void )
  {
    float norm = length( *this );
    if ( FLT_EPSILON < norm )
    {
      *this /= norm;
    }
    return( norm );
  }

  template<unsigned char n> void Vecnf<n>::orthogonalize( const Vecnf<n> &v )
  {
    //  determine the orthogonal projection of this on v : ( v * this ) * v
    //  and subtract it from this resulting in the orthogonalized this
    *this = *this - ( v * *this ) * v;
  }

  template<unsigned char n> void Vecnf<n>::orthonormalize( const Vecnf<n> &v )
  {
    __ASSERT( isNormalized( v ) );
    //  determine the orthogonal projection of this on v : ( v * this ) * v
    //  and subtract it from this resulting in the orthogonalized this
    orthogonalize( v );
    normalize();    //  and normalize it
  }

}
