// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include  <algorithm> // for std::max, std::min (_MAX, _MIN)
#include  <float.h>

// __WIN64__wa__XXX macros to workaround some win64 issues
#include "__WIN64__workarounds.h"

//! Namespace for mathematical helper functions and classes.
namespace nvmath
{
  // define the stand trig values
#ifdef PI
# undef PI
# undef PI_HALF
# undef PI_QUARTER
#endif
  //! constant PI
  const float PI            = 3.14159265358979323846f;
  //! constant PI half
  const float PI_HALF       = 1.57079632679489661923f;
  //! constant PI quarter
  const float PI_QUARTER    = 0.78539816339744830962f;
  //! constant square root two
  const float SQRT_TWO      = 1.41421356237309504880f;
  //! constant square root two half
  const float SQRT_TWO_HALF = 0.70710678118654752440f;

  //! Template to clamp an object of type T to a lower and an upper limit.
  /** \returns clamped value of \a v between \a l and \a u */
  template<class T> T     clamp( T v                  //!<  value to clamp
                               , T l                  //!<  lower limit
                               , T u                  //!<  upper limit
                               )
  {
    return __WIN64__wa__std_min(u, __WIN64__wa__std_max(l,v));
  }

  //! Transform an angle in degrees to radians.
  /** \returns angle in radians */
  inline  float degToRad( float angle                 //!<  angle in degrees
                        )
  {
    return( angle*PI/180.0f );
  }

  //! Determine if an integer is a power of two.
  /** \returns \c true if \a n is a power of two, otherwise \c false  */
  inline  bool  isPowerOfTwo( int n                   //!<  integer to test
                            )
  {
    return( n == ( n & -n ) );
  }

  //!  Linear interpolation between two values \a v0 and \a v1.
  /**  v = v0 + alpha * ( v1 - v0 ) */
  template<typename T> T lerp( float alpha  //!<  interpolation parameter
                             , const T &v0  //!<  starting value
                             , const T &v1  //!<  ending value
                             )
  {
    return( v0 + alpha * ( v1 - v0 ) );
  }

  //! Determine the largest integer power of two below an integer.
  /** \returns largest power of two below \a n */
  inline  int   powerOfTwoBelow( int n                //!<  integer to use as the upper limit
                               )
  {
    int h2 = 1;
    for ( size_t h = n ; h ; h>>=1, h2<<=1 )
      ;
    return( h2 >> 1 );
  }

  //! Transform an angle in radian to degree.
  /** \returns angle in degrees */
  inline float radToDeg( float angle                //!<  angle in radians
                       )
  {
    return( angle*180.0f/PI );
  }

  //! Determine the sign of a scalar.
  /** \returns sign of \a t */
  template<typename T> int sign( const T &t         //!<  scalar value
                               )
  {
    return( ( t < 0 ) ? -1 : ( ( t > 0 ) ? 1 : 0 ) );
  }

  //! Template to square an object of Type T.
  /** \returns product of \a t with itself */
  template<typename T> float square( const T& t          //!<  object to square
                                   )
  {
    return( t * t );
  }

}
