// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include "nvmath/Vec2f.h"
#include "nvmath/Vec3f.h"
#include "nvmath/Quatf.h"
#include "nvsg/Object.h"

//  forward declarations
namespace nvmath
{
  class Mat44f;
  class Vec2f;
}
namespace nvsg
{
  class LightSource;
  class Node;
}

namespace nvsg
{
  //! Base class for all camera types.
  class Camera : public Object
  {
    public:
      //! Interface for creating a clone of this camera.
      /** \returns A constant pointer to the new Camera object. */
      NVSG_API virtual const Camera * clone( void ) const = 0;

      //
      // head light handling interfaces
      //

      //! Interface for adding a headlight to this camera.
      NVSG_API virtual void addHeadLight( 
        const LightSource * pLight                                          //!<  Pointer to the LightSource to add.
      , const nvmath::Vec3f &trans = nvmath::Vec3f( 0.0f, 0.0f, 0.0f )      //!<  Specifies the translation relative to camera.
      , const nvmath::Quatf &quat = nvmath::Quatf( 0.0f, 0.0f, 0.0f, 1.0f ) //!<  Specifies the rotation relative to camera.
      ) = 0;

      //! Interface for removing a headlight.
      /** \returns \c true if the headlight was removed, otherwise \c false */
      NVSG_API virtual bool removeHeadLight( size_t index = 0   //!<  index of head light to remove
                                           ) = 0;

      //! Interface for getting the number of headlights attached to the camera.
      /** \returns  Number of headlights attached to this camera. */
      NVSG_API virtual size_t getNumberOfHeadLights() const = 0;

      //! Interface for getting a headlight of this camera.
      /** \returns A constant pointer to the headlight at index \a index, 
        * or a null pointer if there is no headlight available at index \a index. */
      NVSG_API virtual const LightSource* getHeadLight( size_t index  //!<  Index of headlight to get
                                                      ) const = 0;

      //! Interface for getting the position/orientation of a camera's headlight.
      NVSG_API virtual void getHeadLightTrafo( 
        nvmath::Vec3f &trans  //!<  The translation relative to the camera will be received here.
      , nvmath::Quatf &quat   //!<  The rotation relative to the camera will be received here.
      , size_t index = 0      //!<  Index of the headlight to get transform information from.
      ) const = 0;

      //! Interface for setting the headlight position and orientation relative to the camera.
      /** \returns \c true if a headlight with this index exists, \c false otherwise. */
      NVSG_API virtual bool setHeadLightTrafo( 
        const nvmath::Vec3f &trans  //!< position of headlight - default is (0,0,0)
      , const nvmath::Quatf &quat   //!< orientation of headlight - default is (0,0,0,1) 
      , size_t index = 0            //!< index of headlight
      ) = 0;

      //
      // basic camera interface
      //

      //! Interface for getting the distance to the target point.
      /** \returns distance to the target point */
      NVSG_API virtual float getTargetDistance() const = 0;

      //! Interface for setting the distance to the target point.
      NVSG_API virtual void setTargetDistance( float td //!<  distance to target point (or viewing plane)
                                             ) = 0;

      //! Interface for getting the camera position.
      /** \returns position of camera */
      NVSG_API virtual const nvmath::Vec3f& getPosition() const = 0;

      //! Interface for setting the camera position.
      NVSG_API virtual void setPosition( const nvmath::Vec3f& pos //!<  new camera position
                                       ) = 0;

      //! Interface for setting the up direction.
      NVSG_API virtual void setUpVector( const nvmath::Vec3f &up //!<  up direction (might change viewing direction)
                                       ) = 0;

      //! Interface for getting the up direction.
      /** \returns up direction of camera */
      NVSG_API virtual nvmath::Vec3f getUpVector() const = 0;

      //! Interface for getting the viewing direction.
      /** \returns viewing direction  */
      NVSG_API virtual nvmath::Vec3f getDirection() const = 0;

      //! Interface for setting the viewing direction.
      NVSG_API virtual void setDirection( const nvmath::Vec3f& dir //!<  new viewing direction (might change upVector)
                                        ) = 0;

      //! Interface for setting the orientation of the camera.
      NVSG_API virtual void setOrientation( const nvmath::Quatf &quat //!<  quaternion that determines direction and up vector
                                          ) = 0;

      //! Interface for setting the orientation of the camera.
      NVSG_API virtual void setOrientation( const nvmath::Vec3f &dir //!<  direction to look at
                                          , const nvmath::Vec3f &up  //!<  up vector
                                          ) = 0;

      //! Interface for setting the orientation of the camera.
      NVSG_API virtual void setOrientation( const nvmath::Vec3f &axis //!<  axis to rotate about
                                          , float angle               //!<  angle to rotate
                                          ) = 0;

      //! Get the current orientation.
      /** \returns  quaternion that describes the orientation */
      NVSG_API virtual nvmath::Quatf getOrientation( void ) const = 0;

      //! Interface for getting the world-to-view transformation.
      /** \returns world-to-camera transformation */
      NVSG_API virtual nvmath::Mat44f getWorldToViewMatrix() const = 0;

      //! Interface for getting the inverse world to view transformation.
      /** \returns inverse of the world to camera transformation (i.e. camera to world) */
      NVSG_API virtual nvmath::Mat44f getViewToWorldMatrix() const = 0;

      //! Interface for getting the window offset.
      /** \returns window offset  */
      NVSG_API virtual const nvmath::Vec2f& getWindowOffset( void ) const = 0;

      //! Interface for setting the window offset.
      NVSG_API virtual void setWindowOffset( const nvmath::Vec2f &offset //!<  window ofset to viewing reference point
                                           ) = 0;

      //! Interface for getting the relative window region to view
      NVSG_API virtual void getWindowRegion( nvmath::Vec2f &lowerLeft  //!< relative lower left position
                                           , nvmath::Vec2f &upperRight //!< relative upper right position
                                           ) const = 0;

      //! Interface for setting the relative window region to view
      NVSG_API virtual void setWindowRegion( const nvmath::Vec2f &lowerLeft  //!< relative lower left position
                                           , const nvmath::Vec2f &upperRight //!< relative upper right position
                                           ) = 0;

      //! Interface for getting the window size.
      /** \returns window size  */
      NVSG_API virtual const nvmath::Vec2f& getWindowSize() const = 0;

      //! Interface for setting the window size.
      NVSG_API virtual void setWindowSize( const nvmath::Vec2f &size    //!< window size (world coordinates)
                                         ) = 0;

      //! Interface for getting the current aspect ratio.
      /** \returns width versus height of the camera window */
      NVSG_API virtual float getAspectRatio() const = 0;

      //! Interface for setting the aspect ratio.
      NVSG_API virtual void setAspectRatio( float ar                //!< new aspect ratio
                                          , bool keepWidth = false  //!< keep width or height constant
                                          ) = 0;

      //! Interface for determining if auto clipping is on.
      /** If auto clipping is on, the camera determines the optimal near and front 
        * clipping distances from the scene. If the user has additional information 
        * about the scene that might help clipping, it should be turned off.
        * \returns \c true if auto clip is on, otherwise \c false */
      NVSG_API virtual bool isAutoClipPlanes() const = 0;

      //! Interface for enabling/disabling automatic clip plane determination.
      NVSG_API virtual void setAutoClipPlanes( bool on  //!<  \c true to enable, \c false to disable
                                             ) = 0;

      //! Interface for calculating the near and far clipping distances.
      NVSG_API virtual void calcNearFarDistances( const nvmath::Sphere3f &sphere  //!<  bounding sphere of the world to look at
                                                ) = 0;

      //! Interface for getting the far clipping distance.
      /** \returns distance to the far clipping plane */
      NVSG_API virtual float getFarDistance() const = 0;

      //! Interface for setting the far clipping distance.
      NVSG_API virtual void setFarDistance( float fd //!< far clipping distance
                                          ) = 0;

      //! Interface for getting the near clipping distance.
      /** \returns distance to the near clipping plane  */
      NVSG_API virtual float getNearDistance() const = 0;

      //! Interface for setting the near clipping distance.
      NVSG_API virtual void setNearDistance( float nd //!<  near clipping distance
                                           ) = 0;

      //
      // Operations to perform on a camera
      //

      //! Interface for moving the camera.
      NVSG_API virtual void move( const nvmath::Vec3f& delta //!<  delta to move the camera (camera coordinates)
                                ) = 0;

      //! Interface for orbiting the camera.
      NVSG_API virtual void orbit( const nvmath::Vec3f& axis  //!<  axis to orbit around (camera coordinates)
                                 , float rad                  //!<  angle to orbit (radians)
                                 ) = 0;

      //! Interface for rotating the camera.
      NVSG_API virtual void rotate( const nvmath::Vec3f& axis     //!<  axis to rotate around 
                                  , float rad                     //!<  angle to rotate (radians)
                                  , bool  cameraRelative = true   //!<  flag to indicate if the axis is camera or world relative

                                  ) = 0;

      //! Interface for zooming by a certain factor.
      NVSG_API virtual void zoom( float factor    //!< zoom in/out by this factor; changes windowSize
                                ) = 0;

      //! Interface for zooming to fit view to a sphere.
      NVSG_API virtual void zoom( const nvmath::Sphere3f &sphere //!< sphere to zoom on
                                , float fovy = -1.0f             //!< field of view; if negative, keep the current field of view
                                ) = 0;

      //! Move the camera to the right/left (camera coordinates).
      void moveX( float val   //!<  coordinate delta to move
                );

      //! Move the camera up/down (camera coordinates).
      void moveY( float val   //!< coordinate delta to move
                );

      //! Move the camera back/forth (camera coordinates)
      void moveZ( float val   //!< coordinate delta to move
                );

      //! Orbit the camera around the camera's x-axis.
      void orbitX( float rad    //!< angle to orbit (radians)
                 );

      //! Orbit the camera around the up direction.
      void orbitY( float rad    //!< angle to orbit (radians)
                 );

      //! Orbit the camera around the negative view direction.
      void orbitZ( float rad    //!< angle to orbit (radians)
                 );

      //! Rotate the camera around the camera's x-axis
      void rotateX( float rad                     //!<  angle to rotate (radians)
                  , bool  cameraRelative = true   //!<  flag to indicate if the axis is camera or world relative
                  );

      //! Rotate the camera around the up direction.
      void rotateY( float rad                   //!<  angle to rotate (radians)
                  , bool  cameraRelative = true //!<  flag to indicate if the axis is camera or world relative
                  );

      //! Rotate the camera around the negative view direction.
      void rotateZ( float rad                     //!<  angle to rotate (radians)
                  , bool  cameraRelative = true   //!<  flag to indicate if the axis is camera or world relative
                  );

    protected:
      //!  Protected constructor 
      NVSG_API Camera();

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~Camera( void );
  };

  // - - - - - - - - - - - - - - - - - - - - -
  // inlines
  // - - - - - - - - - - - - - - - - - - - - -

  inline Camera::Camera( void )
  {
    __TRACE();
  }

  inline Camera::~Camera(void)
  {
    __TRACE();
  }

  inline void  Camera::moveX( float val )
  { 
    __TRACE();
    move( nvmath::Vec3f( val, 0.f, 0.f ) );
  }

  inline void  Camera::moveY( float val )
  {
    __TRACE();
    move( nvmath::Vec3f( val, 1.f, 0.f ) );
  }

  inline void  Camera::moveZ( float val )
  { 
    __TRACE();
    move( nvmath::Vec3f( val, 0.f, 1.f ) ); 
  }

  inline void  Camera::orbitX( float rad )
  { 
    __TRACE();
    orbit( nvmath::Vec3f( 1.f, 0.f, 0.f ), rad );
  }

  inline void  Camera::orbitY( float rad )
  {
    __TRACE();
    orbit( nvmath::Vec3f( 0.f, 1.f, 0.f ), rad );
  }

  inline void  Camera::orbitZ( float rad )
  { 
    __TRACE();
    orbit( nvmath::Vec3f( 0.f, 0.f, 1.f ), rad ); 
  }

  inline void  Camera::rotateX( float rad, bool  cameraRelative /*= true*/ )
  {
    __TRACE();
    rotate( nvmath::Vec3f(1.f, 0.f, 0.f), rad, cameraRelative );  
  }

  inline void  Camera::rotateY( float rad, bool  cameraRelative /*= true*/ )
  { 
    __TRACE();
    rotate( nvmath::Vec3f(0.f, 1.f, 0.f), rad, cameraRelative );  
  }

  inline void  Camera::rotateZ( float rad, bool  cameraRelative /*= true*/ )
  {
    __TRACE();
    rotate( nvmath::Vec3f(0.f, 0.f, 1.f), rad, cameraRelative ); 
  }

} //  namespace nvsg
