// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvmath/Vec3f.h"
#include "nvsg/StateAttribute.h"
#include "nvutil/BitMask.h"


namespace nvsg
{
  class FaceAttribute;

  //! Data of a \c FaceAttribute that is worth sharing
  class FaceAttributeData : public nvutil::RCObject
  {
    friend class FaceAttribute;

    public:
      //! Copy constructor
      NVSG_API FaceAttributeData(const FaceAttributeData& rhs);

    private:
      // instantiation on stack is not permitted
      FaceAttributeData();
      ~FaceAttributeData();
      // assignment also not permitted
      FaceAttributeData& operator=(const FaceAttributeData& rhs) {}

    private:
      nvmath::Vec3f m_ambientColor;
      bool          m_backFaceCulling; 
      struct
      {
        bool  enabled;
        float factor;
        float units;
      }             m_polygonOffset;
      bool          m_twoSidedLighting;
  };

  //! StateAttribute that determines the render modes to use for the following geometries.
  class FaceAttribute : public StateAttribute
  {
    public:
      //! Create a FaceAttribute.
      /** \returns a constant pointer to a FaceAttribute. */
      NVSG_API static const FaceAttribute * create( void );

      //! Create a FaceAttribute by copying from a StateAttribute.
      /** \returns a constant pointer to a FaceAttribute. */
      NVSG_API static const FaceAttribute * createFromBase( const StateAttribute &rhs );

      //! Get a clone of this render mode.
      /** \returns a constant pointer to a FaceAttribute. */
      NVSG_API virtual const FaceAttribute * clone( void ) const;

      //! Determine whether the data of this object is shared.
      /** \returns  \c true if the object's data is shared, otherwise \c false.  */
      NVSG_API virtual bool isDataShared( void ) const;

      //! Get the DataID of the data of this object.
      /** \returns  DataID of the object's data.  */
      NVSG_API virtual DataID getDataID( void ) const;

      //! Get the ambient color.
      /** This is the constant light for the following geometries. It lights all object vertices the same because it is
        * not dependent on any other lighting factors such as vertex normals, light direction, light position, range, or
        * attenuation. The default is (0.2,0.2,0.2).
        * \returns  The current ambient color */
      NVSG_API const nvmath::Vec3f& getAmbientColor( void ) const;

      //! Get the polygon offset scale factor.
      /** The PolygonOffsetFactor is a scale factor that is used to create a variable depth offset for each polygon. The
        * default is 0.0.
        * \returns  The current polygon offset scale factor */
      NVSG_API float getPolygonOffsetFactor( void ) const;

      //! Get the polygon offset units.
      /** The PolygonOffsetUnits specifies a value that is multiplied by an implementation-specific value to create a
        * constant depth offset. The default is 0.0. 
        * \returns  The current polygon offset units */
      NVSG_API float getPolygonOffsetUnits( void ) const;

      //! Get the back face culling state.
      /** The back face culling state specifies how back-facing facets are handled. If it is disabled, back-facing facets
        * are drawn regulary. Otherwise, back-facing facets are not drawn. Lines and points are not affected by this
        * state. The default is true.
        * \returns  The current back face culling state */
      NVSG_API bool isBackFaceCulling( void ) const;

      //! Get the polygon offset state.
      /** The polygon offset state specifies how polygon offsets are handled. If it is disabled, no polygon offsets are
        * used. Otherwise, each fragment's depth value will be offset after it is interpolated from the depth values of
        * the appropriate vertices. The offset is added before the depth test is performed and before the value is written
        * into the depth buffer. The default is false.
        * \returns  The current polygon offset state */
      NVSG_API bool isPolygonOffset( void ) const;

      //! Get the two sided lighting state.
      /** The two sided lighting state specifies whether one or two-sided lighting calculations are performed for polygons.
        * It has no effect on the lighting calculations for points, lines, or bitmaps. If it is disabled, one-sided lighting
        * is specified. Otherwise, two-sided lighting is specified. In this case, vertices of back-facing polygons are
        * lighted, and have their normals reversed before the lighting equation is evaluated. Vertices of front-facing
        * polygons are always lighted, with no change to their normals. The default is false. 
        * \returns  The current two sided lighting state */
      NVSG_API bool isTwoSidedLighting( void ) const;

      //! Set the ambient color.
      /** This is the constant light for the following geometries. It lights all object vertices the same because it is
        * not dependent on any other lighting factors such as vertex normals, light direction, light position, range, or
        * attenuation. The default is (0.2,0.2,0.2). */
      NVSG_API void setAmbientColor( const nvmath::Vec3f &color   //!<  ambient color to set
                                   );

      //! Set the back face culling state.
      /** The back face culling state specifies how back-facing facets are handled. If it is disabled, back-facing facets
        * are drawn regulary. Otherwise, back-facing facets are not drawn. Lines and points are not affected by this
        * state. The default is true. */
      NVSG_API void setBackFaceCulling( bool onOff   //!< \c true to switch back face culling on, \c false to switch it off
                                      );

      //! Set the polygon offset state.
      /** The polygon offset state specifies how polygon offsets are handled. If it is disabled, no polygon offsets are
        * used. Otherwise, each fragment's depth value will be offset after it is interpolated from the depth values of
        * the appropriate vertices. The offset is added before the depth test is performed and before the value is written
        * into the depth buffer. The default is false. */
      NVSG_API void setPolygonOffset( bool onOff    //!<  \c true to switch polygon offset on, \c false to switch it off
                                    );

      //! Set the polygon offset scale factor and units.
      /** The PolygonOffset factor is a scale factor that is used to create a variable depth offset for each polygon. The
        * default is 0.0. The PolygonOffset units specifies a value that is multiplied by an implementation-specific value
        * to create a constant depth offset. The default is 0.0. */
      NVSG_API void setPolygonOffset( float factor    //!<  polygon offset scale factor to use
                                    , float units     //!<  polygon offset units to use
                                    );

      //! Set the two sided lighting state.
      /** The two sided lighting state specifies whether one or two-sided lighting calculations are performed for polygons. 
        * It has no effect on the lighting calculations for points, lines, or bitmaps. If it is disabled, one-sided lighting
        * is specified. Otherwise, two-sided lighting is specified. In this case, vertices of back-facing polygons are
        * lighted, and have their normals reversed before the lighting equation is evaluated. Vertices of front-facing
        * polygons are always lighted, with no change to their normals. The default is false. */
      NVSG_API void setTwoSidedLighting( bool onOff  //!< \c true to switch two sided lighting on, \c false to switch it off
                                       );

    protected:
      //! Default constructor
      NVSG_API FaceAttribute();

      //! Partial Constructor
      NVSG_API FaceAttribute( const StateAttribute &rhs );

      //! Copy Constructor
      NVSG_API FaceAttribute( const FaceAttribute &rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~FaceAttribute( void );

    private:
      // shareable render mode data
      nvutil::RCPtr<FaceAttributeData> m_faceAttribute;
  };

  inline const nvmath::Vec3f& FaceAttribute::getAmbientColor( void ) const
  {
    __TRACE();
    return( m_faceAttribute->m_ambientColor );
  }

  inline void FaceAttribute::setAmbientColor( const nvmath::Vec3f &color )
  {
    __TRACE();
    nvutil::writeAccess( m_faceAttribute );   //  nvutil::writeAccess performs a deep copy if necessary
    m_faceAttribute->m_ambientColor = color;
  }

  inline void FaceAttribute::setBackFaceCulling( bool onOff )
  {
    __TRACE();
    nvutil::writeAccess( m_faceAttribute );   //  nvutil::writeAccess performs a deep copy if necessary
    m_faceAttribute->m_backFaceCulling = onOff;
  }

  inline bool FaceAttribute::isBackFaceCulling( void ) const
  {
    __TRACE();
    return m_faceAttribute->m_backFaceCulling;
  }

  inline void FaceAttribute::setPolygonOffset( bool onOff )
  {
    __TRACE();
    nvutil::writeAccess( m_faceAttribute );   //  nvutil::writeAccess performs a deep copy if necessary
    m_faceAttribute->m_polygonOffset.enabled = onOff;
  }

  inline void FaceAttribute::setPolygonOffset( float factor, float units )
  {
    __TRACE();
    nvutil::writeAccess( m_faceAttribute );   //  nvutil::writeAccess performs a deep copy if necessary
    m_faceAttribute->m_polygonOffset.factor = factor;
    m_faceAttribute->m_polygonOffset.units = units;
  }

  inline bool FaceAttribute::isPolygonOffset( void ) const
  {
    __TRACE();
    return( m_faceAttribute->m_polygonOffset.enabled );
  }

  inline float FaceAttribute::getPolygonOffsetFactor( void ) const
  {
    __TRACE();
    return( m_faceAttribute->m_polygonOffset.factor );
  } 

  inline float FaceAttribute::getPolygonOffsetUnits( void ) const
  {
    __TRACE();
    return( m_faceAttribute->m_polygonOffset.units );
  } 

  inline void FaceAttribute::setTwoSidedLighting( bool onOff )
  {
    __TRACE();
    nvutil::writeAccess( m_faceAttribute );   //  nvutil::writeAccess performs a deep copy if necessary
    m_faceAttribute->m_twoSidedLighting = onOff;
  }

  inline bool FaceAttribute::isTwoSidedLighting( void ) const
  {
    __TRACE();
    return( m_faceAttribute->m_twoSidedLighting );
  }

} // namespace nvsg
