// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include  "nvsg/Animation.h"

#ifdef _DEBUG
# ifndef new
#  define new new(__FILE__, __LINE__)
#  define _DEFINED_DBGNEW // restrict the 'new' macro to this header file only
# endif
#endif

namespace nvsg
{
  class VNVector;
  
  //! An \c Animation that holds a complete data set for each frame.
  /** \note Currently the only supported types for FramedAnimation are \c Trafo and \c VNVector.  */
  template<typename T> class FramedAnimation : public Animation<T>
  {
    public:
      //! Create a FramedAnimation.
      /** \returns a constant pointer to an FramedAnimation. */
      static const FramedAnimation<T> * create( void );

      //! Create a FramedAnimation by copying from an Animation.
      /** \returns a constant pointer to a FramedAnimation. */
      static const FramedAnimation<T> * createFromBase( const Animation<T> &rhs   //!< Animation to copy from
                                                      );

      //! Create a new FramedAnimation by copying from this.
      /** \returns a constant pointer to an identical FramedAnimation. */
      virtual const FramedAnimation<T> * clone( void ) const;

      //! Set the number of frames in the animation.
      /** Resizes the array of animation values. On growing this array, the new elements are not initialized. */
      virtual void setNumberOfFrames( size_t count );

      //! Random access operator for more convenient usage.
      /** \returns reference to value of frame \a i  */
      virtual T & operator[]( size_t i );

      //--  Functions reimplemented from Animation  --
      virtual size_t getNumberOfFrames( void ) const;
      virtual const T & operator[]( size_t i ) const;

    protected:
      //! Default Constructor
      NVSG_API FramedAnimation( void );

      //! Partial Constructor
      NVSG_API FramedAnimation( const Animation<T> &rhs );

      //! Copy Constructor
      NVSG_API FramedAnimation( const FramedAnimation<T> &rhs );

      //! Protected destructor to prevent explicit creation on stack.
      virtual ~FramedAnimation( void );

    private:
      std::vector<T> m_values;
  };

  template<typename T> FramedAnimation<T>::~FramedAnimation( void )
  {
    __TRACE();
  }

  template<typename T> const FramedAnimation<T> * FramedAnimation<T>::create( void )
  {
    __TRACE();
    return( new FramedAnimation<T> );
  }

  template<typename T> const FramedAnimation<T> * FramedAnimation<T>::createFromBase( const Animation<T> &rhs )
  {
    __TRACE();
    return( new FramedAnimation<T>( rhs ) );
  }

  template<typename T> const FramedAnimation<T> * FramedAnimation<T>::clone( void ) const
  {
    __TRACE();
    return( new FramedAnimation<T>( *this ) );
  }

  template<typename T> void FramedAnimation<T>::setNumberOfFrames( size_t count )
  {
    __TRACE();
    m_values.resize( count );
  }

  template<typename T> T & FramedAnimation<T>::operator[]( size_t i )
  {
    __TRACE();
    __ASSERT( i < m_values.size() );
    return( m_values[i] );
  }

  template<typename T> size_t FramedAnimation<T>::getNumberOfFrames( void ) const
  {
    __TRACE();
    return( m_values.size() );
  }

  template<typename T> const T & FramedAnimation<T>::operator[]( size_t i ) const
  {
    __TRACE();
    __ASSERT( i < m_values.size() );
    return( m_values[i] );
  }
}

#ifdef _DEFINED_DBGNEW
# undef new
#endif
