// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvsg/HighlightObject.h"
#include "nvmath/Sphere3f.h"
#include "nvmath/Vec3f.h"

namespace nvsg
{
  //! Base class for Highlighting objects that can be attached to Geosets
  class HighlightEdges : public HighlightObject
  {
    public:
      //! Create a HighlightEdges.
      /** \returns a constant pointer to a HighlightEdges. */
      NVSG_API static const HighlightEdges * create( float width = 2.0f                                    //!< default line width
                                                   , nvmath::Vec3f color = nvmath::Vec3f(1.0f, 0.0f, 0.0f) //!< default line color
                                                   );

      //! Create a HighlightEdges by copying from a HighlightObject.
      /** \returns a constant pointer to a HighlightBox. */
      NVSG_API static const HighlightEdges * createFromBase( const HighlightObject &rhs                           //!< HighlightObject to copy from
                                                           , float width = 2.0f                                   //!< default line width
                                                           , nvmath::Vec3f color = nvmath::Vec3f(1.0f,0.0f,0.0f)  //!< default line color
                                                           );

      //! Create a HighlightEdges by cloning this.
      /** \returns a constant pointer to a HighlightEdges. */
      NVSG_API virtual const HighlightEdges * clone( void ) const;

      //! This object does not hold any geometry, so it's bounding sphere is not valid.
      /** \returns \c false */
      NVSG_API virtual bool isBoundingSphereValid( void ) const;

      //! Get the bounding sphere of this HighlightObject.
      /** A HighlightEdges object does not have a bounding sphere, therefor this function shouldn't be used. */
      NVSG_API virtual const nvmath::Sphere3f & getBoundingSphere( void ) const;

      //! return the line width used for drawing the box
      NVSG_API float getLineWidth( void ) const;      

      //! return the line width used for drawing the box
      NVSG_API const nvmath::Vec3f & getColor( void ) const;
      
    protected:
      //! Constructor with line width and color.
      NVSG_API HighlightEdges( float width, nvmath::Vec3f color );

      //! Partial Constructor
      NVSG_API HighlightEdges( const HighlightObject &rhs, float width, nvmath::Vec3f color );

      //! Copy Constructor
      NVSG_API HighlightEdges( const HighlightEdges &rhs );

    protected:
      float m_lineWidth;    //!< line width used for drawing the bounding box
      nvmath::Vec3f m_color;           //!< color used for drawing

    private:
      HighlightEdges( void );   //  private default constructor to prevent usage
  };

  inline bool HighlightEdges::isBoundingSphereValid( void ) const
  {
    return( false );
  }

  inline const nvmath::Sphere3f & HighlightEdges::getBoundingSphere( void ) const 
  { 
    __TRACE();
    __ASSERT( false );
    return( nvmath::cUnitSphere );
  }

  inline float HighlightEdges::getLineWidth( void ) const 
  { 
    __TRACE();
    return m_lineWidth; 
  }       

  inline const nvmath::Vec3f & HighlightEdges::getColor( void ) const 
  { 
    __TRACE();
    return m_color; 
  }
}
