// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvsg/SimpleCamera.h"

namespace nvsg
{
  //! A perspective camera
  /** This is a SimpleCamera that represents a perspective camera.  */
  class PerspectiveCamera : public SimpleCamera
  {
    public:
      //! Create a PerspectiveCamera.
      /** \returns a constant pointer to a PerspectiveCamera. */
      NVSG_API static const PerspectiveCamera * create( void );

      //! Create a PerspectiveCamera by copying from a SimpleCamera.
      /** \returns a constant pointer to a PerspectiveCamera. */
      NVSG_API static const PerspectiveCamera * createFromBase( const SimpleCamera &rhs  //!< SimpleCamera to copy from
                                                              );

      //! Create a clone of this PerspectiveCamera.
      /** \returns a constant pointer to a PerspectiveCamera. */
      NVSG_API virtual const PerspectiveCamera * clone( void ) const;

      //! Get the field of view (fovy).
      /** The field of view is the view angle (in radians) in the y-direction.
        * \returns The field of view (in radians)
        */
      NVSG_API float getFieldOfView()  const;

      //! Set the field of view (fovy).
      /** The field of view is the view angle (in radians) in the y-direction.  */
      NVSG_API void setFieldOfView( float fov   //!<  field of view to set (in radians)
                                  );

      //! Set the distance to the target point.
      /** On perspective cameras, setting the target distance keeps the field of view constant. This means that
        * the window size is adjusted accordingly.  */
      NVSG_API virtual void setTargetDistance( float td   //!<  distance to target point (or viewing plane)
                                             );

      //! Get the projection matrix
      /** \returns  The projection matrix */
      NVSG_API virtual nvmath::Mat44f getProjection() const;

      //! Get the inverse projection matrix
      /** \returns  The inverse projection matrix */
      NVSG_API virtual nvmath::Mat44f getInverseProjection() const;

    protected:
      //! Constructor
      NVSG_API PerspectiveCamera();

      //! Partial Constructor
      NVSG_API PerspectiveCamera( const SimpleCamera &rhs );

      //! Copy Constructor
      NVSG_API PerspectiveCamera( const PerspectiveCamera &rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~PerspectiveCamera(void);
  };

  inline float PerspectiveCamera::getFieldOfView() const
  {
    __TRACE();
    return( 2.0f * atanf( 0.5f * getWindowSize()[1] / getTargetDistance() ) );
  }
} // namespace
