// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include  "nvsg/LightSource.h"

namespace nvsg
{
  /*  NOTE: A PointLight does NOT contain a position !
      It is assumed to be in the origin.
      Any changes to this must be performed by a Transform node.
  */

  //! Class representing a point light
  /** Point lights have color and position within a scene, but no single direction. They give off light equally in all
    * directions. A light bulb is a good example of a point light. Point lights are affected by attenuation and range,
    * and illuminate a mesh on a vertex-by-vertex basis. During lighting, the point light's position in world space and
    * the coordinates of the vertex being lit are used to derive a vector for the direction of the light and the
    * distance that the light has traveled. Both are used, along with the vertex normal, to calculate the contribution of
    * the light to the illumination of the surface. */
  class PointLight : public LightSource
  {
    public:
      //! Create a PointLight.
      /** \returns a constant pointer to a PointLight. */
      NVSG_API static const PointLight * create( void );

      //! Create a PointLight by copying from a LightSource.
      /** \returns a constant pointer to a PointLight. */
      NVSG_API static const PointLight * createFromBase( const LightSource &rhs  //!< LightSource to copy from
                                                       );

      //! Get a clone of this PointLight.
      /** \returns A constant pointer to the new created clone. */
      NVSG_API virtual const PointLight * clone( void ) const;

      //! Determine whether the data of this object is shared.
      /** \returns \c true if the object's data is shared, otherwise \c false.  */
      NVSG_API virtual bool isDataShared( void ) const;

      //! Get the DataID of the data of this object.
      /** \returns  DataID of the object's data.  */
      NVSG_API virtual DataID getDataID( void )  const;

      //! Get the attenuation factors for this light.
      /** The attenuation factors are used to calculate the attenuation of this light at a lit point by the equation
        *     atten = 1 / ( c + l * d + q * d * d )
        * where d is the distance between the light and the vertex. */
      NVSG_API void getAttenuation( float &c    //!<  constant attenuation factor
                                  , float &l    //!<  linear attenuation factor
                                  , float &q    //!<  quadratic attenuation factor
                                  ) const;

      //! Get the attenuation factors for this light as a \c Vec3f.
      /** The attenuation factors are used to calculate the attenuation of this light at a lit point by the equation
        *     atten = 1 / ( v[0] + v[1] * d + v[2] * d * d )
        * v is the returned \c Vec3f, and d is the distance between the light and the vertex. */
      NVSG_API const nvmath::Vec3f& getAttenuation( void ) const;

      //! Ask if this light source is attenuated
      /** A light source is attenuated if the constant attenuation factor is different from one, or if the linear or
        * quadratic attenuation factor is different from zero.
        * \returns \c true if this light source is attenuated, otherwise \c false */
      NVSG_API bool isAttenuated( void ) const;

      //! Set the attenuation factors for this light.
      /** The attenuation factors are used to calculate the attenuation of this light at a lit point by the equation
        *     atten = 1 / ( c + l * d + q * d * d )
        * where d is the distance between the light and the vertex. */
      NVSG_API void  setAttenuation( float c    //!<  constant attenuation factor
                                   , float l    //!<  linear attenuation factor
                                   , float q    //!<  quadratic attenuation factor
                                   );

    protected:
      //! Default Constructor
      NVSG_API PointLight( void );

      //! Partial Constructor
      NVSG_API PointLight( const LightSource &rhs );

      //! Copy Constructor
      NVSG_API PointLight( const PointLight &rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~PointLight( void );

    private:
      nvmath::Vec3f m_attenuation;
  };

  inline void PointLight::getAttenuation( float &c, float &l, float &q ) const
  {
    __TRACE();
    c = m_attenuation[0];
    l = m_attenuation[1];
    q = m_attenuation[2];
  }

  inline const nvmath::Vec3f& PointLight::getAttenuation( void ) const
  {
    __TRACE();
    return m_attenuation;
  }

  inline bool PointLight::isAttenuated( void ) const
  {
    __TRACE();
    return(   ( m_attenuation[0] != 1.0f )
          ||  ( m_attenuation[1] != 0.0f )
          ||  ( m_attenuation[2] != 0.0f )
          );
  }

  inline void PointLight::setAttenuation( float c, float l, float q )
  {
    __TRACE();
    m_attenuation[0] = c;
    m_attenuation[1] = l;
    m_attenuation[2] = q;
  }

} //  namespace nvsg
