// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include  "nvsg/LightSource.h"
#include  "nvmath/Vec3f.h"

namespace nvsg
{
  /*  NOTE: A SpotLight does NOT contain a position or direction !
      As a special PointLight it is assumed to be positioned on the origin.
      Furthermore, it is assumed to be directed in the negative z-direction (just as a DirectedLight).
      Any changes to these must be performed by a Transform node.
  */

  //! Class representing a spot light
  /** A SpotLight restricts the shape of the emitted light to a cone. To specify the angle between the axis of the cone
    * and a ray along the edge of the cone, use \c setCutoffAngle() with the angle in degrees. The angle of the cone at
    * the apex is then twice this value. No light is emitted beyond the edges of the cone.
    * \par
    * In addition, there are two ways to control the intensity distribution of the light within the cone. First, you can
    * set the attenuation factors with \c setAttenuation(). You can also set a falloff exponent using \c
    * setFalloffExponent(), which by default is zero, to control how concentrated the light is. The light's intensity is
    * highest in the center of the cone. It is attenuated toward the edges of the cone by the cosine of the angle between
    * the direction of the light and the direction from the light to the vertex being lit, raised to the power of the
    * falloff exponent. Thus, higher falloff exponents result in a more focused light source. 
    * \note The default light direction is (0,0,-1)*/
  class SpotLight : public LightSource
  {
    public:
      //! Create a SpotLight.
      /** \returns a constant pointer to a SpotLight. */
      NVSG_API static const SpotLight * create( void );

      //! Create a SpotLight by copying from a LightSource.
      /** \returns a constant pointer to a SpotLight. */
      NVSG_API static const SpotLight * createFromBase( const LightSource &rhs  //!< LightSource to copy from
                                                       );

      //! Get a clone of this SpotLight.
      /** \returns A constant pointer to the new created clone. */
      NVSG_API virtual const SpotLight * clone( void ) const;

      //! Determine whether the data of this object is shared.
      /** \returns  \c true if the object's data is shared, otherwise \c false.  */
      NVSG_API virtual bool isDataShared( void ) const;

      //! Get the DataID of the data of this object.
      /** \returns  DataID of the object's data.  */
      NVSG_API virtual DataID getDataID( void ) const;

      //! Get attenuation. 
      /** \returns Attenuation factors in a Vec3f. */
      NVSG_API const nvmath::Vec3f& getAttenuation( void ) const;

      //! Ask if this light source is attenuated
      /** A light source is attenuated if the constant attenuation factor is different from one, or if the linear or
        * quadratic attenuation factor is different from zero.
        * \returns \c true if this light source is attenuated, otherwise \c false */
      NVSG_API bool isAttenuated( void ) const;

      //! Get the attenuation factors for this light.
      /** The attenuation factors are used to calculate the attenuation of this light at a lit point by the equation
        *     atten = 1 / ( c + l * d + q * d * d )
        * where d is the distance between the light and the vertex. */
      NVSG_API void  getAttenuation( float &c   //!<  constant attenuation factor
                                   , float &l   //!<  linear attenuation factor
                                   , float &q   //!<  quadratic attenuation factor
                                   ) const;
 
      //! Set the attenuation factors for this light.
      /** The attenuation factors are used to calculate the attenuation of this light at a lit point by the equation
        *     atten = 1 / ( c + l * d + q * d * d )
        * where d is the distance between the light and the vertex. */
      NVSG_API void setAttenuation( float c   //!<  constant attenuation factor
                                  , float l   //!<  linear attenuation factor
                                  , float q   //!<  quadratic attenuation factor
                                  );
 
      //! Get the cutoff angle.
      /** The light's cone is cut off at this angle (in degree). Points beyond this cone are not lit. The default is 45.0.
        * \returns  The cutoff angle */
      NVSG_API float getCutoffAngle( void ) const;

      //! Set the cutoff angle.
      /** The light's cone is cut off at this angle (in degrees). Points beyond this cone are not lit. The default is 45.0. */
      NVSG_API void setCutoffAngle( float ca    //!<  cutoff angle (in degrees)
                                 );

      //! Get the falloff exponent.
      /** The falloff exponent controls the sharpness of the spot light. The higher the exponent, the more focused the
        * light source. The default is 0.0
        * \returns  The falloff exponent */
      NVSG_API float getFalloffExponent( void ) const;

      //! Set the falloff exponent.
      /** The falloff exponent controls the sharpness of the spot light. The higher the exponent, the more focused the
        * light source. The default is 0.0. */
      NVSG_API void setFalloffExponent( float fe    //!<  falloff exponent
                                      );

    protected:
      //! Constructor
      NVSG_API SpotLight( void );

      //! Partial Constructor
      NVSG_API SpotLight( const LightSource &rhs );

      //! Copy Constructor
      NVSG_API SpotLight( const SpotLight &rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~SpotLight( void );

    private:
      nvmath::Vec3f m_attenuation;
      float m_cutoffAngle;
      float m_falloffExponent;
  };

  inline  const nvmath::Vec3f&  SpotLight::getAttenuation()  const
  {
    __TRACE();
    return m_attenuation;
  }

  inline  void  SpotLight::getAttenuation( float &c, float &l, float &q )  const
  {
    __TRACE();
    c = m_attenuation[0];
    l = m_attenuation[1];
    q = m_attenuation[2];
  }

  inline  float SpotLight::getCutoffAngle( void )  const
  {
    __TRACE();
    return( m_cutoffAngle );
  }

  inline  float SpotLight::getFalloffExponent( void )  const
  {
    __TRACE();
    return( m_falloffExponent );
  }

  inline  bool  SpotLight::isAttenuated( void )  const
  {
    __TRACE();
    return(   ( m_attenuation[0] != 1.0f )
          ||  ( m_attenuation[1] != 0.0f )
          ||  ( m_attenuation[2] != 0.0f )
          );
  }

  inline  void  SpotLight::setAttenuation( float c, float l, float q )
  {
    __TRACE();
    m_attenuation[0] = c;
    m_attenuation[1] = l;
    m_attenuation[2] = q;
  }

  inline  void  SpotLight::setCutoffAngle( float ca )
  {
    __TRACE();
    m_cutoffAngle = ca;
  }

  inline  void  SpotLight::setFalloffExponent( float fe )
  {
    __TRACE();
    m_falloffExponent = fe;
  }
} //  namespace nvsg
