// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvmath/Trafo.h"
#include "nvsg/Group.h"

namespace nvsg
{

  //! Transformation node.
  /** A Transform contains a transformation that is appended to the current transform and a number of children that are
    * relative to this transformation.
    */
  class Transform : public Group
  {
    public:
      //! Create a Transform.
      /** \returns a constant pointer to a Transform. */
      NVSG_API static const Transform * create( void );

      //! Create a Transform by copying from a Group.
      /** \returns a constant pointer to a Transform. */
      NVSG_API static const Transform * createFromBase( const Group &rhs );

      //! Get a clone of this \c Transform object.
      /** \returns A constant pointer to a Transform. */
      NVSG_API virtual const Transform * clone( void ) const;

      //! Determine whether the data of this object is shared.
      /** \returns  \c true if the object's data is shared, otherwise \c false.  */
      NVSG_API virtual bool    isDataShared( void )  const;

      //! Get the DataID of the data of this object.
      /** \returns  DataID of the object's data.  */
      NVSG_API virtual DataID  getDataID( void )     const;

      //! Get the inverse of the current transformation.
      /** \returns  inverse transformation as a Mat44f  */
      nvmath::Mat44f  getInverse( void )            const;

      //! Get the center of rotation of this transformation.
      /** \returns  vector that describes the center of rotation. */
      const nvmath::Vec3f & getCenter( void )       const;

      //! Get the rotational part of this transformation.
      /** \returns  The quaternion that describes the rotational part */
      const nvmath::Quatf & getOrientation( void )  const;

      //! Get the scaling part of this transformation.
      /** \returns  The vector that describes the scaling part  */
      const nvmath::Vec3f & getScaling( void )      const;

      //! Get the current transformation.
      /** \returns  The transformation as a Mat44f  */
      nvmath::Mat44f  getMatrix( void  )             const;

      //! Get the translational part of this transformation.
      /** \returns  The vector that describes the translational part  */
      const nvmath::Vec3f & getTranslation( void )  const;

      //! Determine if this Transform is an identity.
      /** \returns \c true, if this Transform is an identity, otherwise \c false. */
      NVSG_API bool isIdentity( void ) const;

      //! Set the center of rotation of this transformation.
      NVSG_API virtual void  setCenter( const nvmath::Vec3f &center   //!<  center of rotation
                                      );

      //! Set the Transform to be the identity.
      NVSG_API virtual void setIdentity( void );

      //! Set the rotational part of this transformation, using a quaternion.
      NVSG_API virtual void  setOrientation( const nvmath::Quatf &orientation  //!<  rotational part of transformation
                                           );

      //! Set the scaling part of this transformation.
      NVSG_API virtual void  setScaling( const nvmath::Vec3f &scaling
                                       );

      //! Set the translational part of this transformation.
      NVSG_API virtual void  setTranslation( const nvmath::Vec3f &translation  //!<  translational part of transformation

                                           );

    protected:
      //! Constructor
      NVSG_API Transform(void);

      //! Partial Constructor
      NVSG_API Transform( const Group &rhs );

      //! Copy Constructor
      NVSG_API Transform( const Transform &rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~Transform(void);

      //! Calculate the bounding sphere of this transform
      /** To calculate the bounding sphere of a transform, it transforms the bounding sphere of all its children.
        * \returns \c true if the calculated bounding sphere is valid, otherwise \c false. */
      NVSG_API virtual bool calcBoundingSphere( void ) const;

    private:
      nvmath::Trafo m_trafo;
  };

  inline  nvmath::Mat44f  Transform::getInverse( void ) const
  {
    return( m_trafo.getInverse() );
  }

  inline  const nvmath::Vec3f & Transform::getCenter( void ) const
  {
    return( m_trafo.getCenter() );
  }

  inline  const nvmath::Quatf & Transform::getOrientation( void ) const
  {
    return( m_trafo.getOrientation() );
  }

  inline  const nvmath::Vec3f & Transform::getScaling( void ) const
  {
    return( m_trafo.getScaling() );
  }

  inline  nvmath::Mat44f  Transform::getMatrix( void  ) const
  {
    return( m_trafo.getMatrix() );
  }

  inline  const nvmath::Vec3f & Transform::getTranslation( void ) const
  {
    return( m_trafo.getTranslation() );
  }

  inline  bool Transform::isIdentity( void ) const
  {
    return( nvmath::isIdentity( m_trafo.getMatrix() ) );
  }
}
