// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include  "nvsg/Face.h"
#include  "nvsg/GeoSet.h"

namespace nvsg
{  
  //! GeoSet for Triangles.
  /** A Triangles is a GeoSet that holds an array of index-triplets. Each of these triplets describe a triangle with
    * the indices pointing in the array of vertices. All vertex attributes (like normal, texture,...) are indexed with
    * the same index.
    */
  class Triangles : public GeoSet
  {
    public:
      //! Create a Triangles.
      /** \returns a constant pointer to a Triangles. */
      NVSG_API static const Triangles * create( void );

      //! Create a Triangles by copying from a GeoSet.
      /** \returns a constant pointer to a Triangles. */
      NVSG_API static const Triangles * createFromBase( const GeoSet &rhs  //!<  GeoSet to copy from
                                                      );
      
      //! Get a clone of this Triangles object
      /** \returns A constant pointer to a Triangles. */
      NVSG_API virtual const Triangles * clone( void ) const;
   
      //! Determine whether the data of this object is shared.
      /** \returns  \c true if the object's data is shared, otherwise \c false.  */
      NVSG_API virtual bool isDataShared( void )  const;

      //! Get the DataID of the data of this object.
      /** \returns  DataID of the object's data.  */
      NVSG_API virtual DataID getDataID( void ) const;

      //! Ask if this Triangles has faces.
      /** \returns  \c true, if this Triangles has faces, otherwise \c false */
      NVSG_API bool hasFaces( void ) const;

      //! Get number of faces.
      /** \returns  The number of faces */
      NVSG_API size_t getNumberOfFaces( void ) const;

      //! Get constant pointer to faces.
      /** \returns A constant pointer to faces. */
      NVSG_API const Face3 * getFaces( void ) const;

      //! Set the faces for this Triangles.
      /** Copies \a numFaces faces from \a pFaces into this Triangles. 
        * Pre-existing faces will be entirely erased. */
      NVSG_API void setFaces( const Face3* pFaces    //!<  faces to set
                            , size_t numFaces        //!<  number of faces
                            );

      //! Set the faces for this Triangles.
      /** Copies \a numFaces faces from \a pFaces into this Triangles, startin at position \a pos
        * inside the range of pre-existing faces.
        *
        * Pre-existing faces in the range [\a pos, \a pos + \a numFaces) will be replaced.
        * Pre-existing faces outside this range remain untouched.
        *
        * If you specify -1 for \a pos or if \a pos specifies the number of faces currently stored,
        * the faces pointed to by \a pFaces will be appended to the pre-existing faces. 
        *
        * If you not specify -1 for \a pos or \a pos neither specifies the number of faces currently
        * stored nor does it specify a valid position inside the range of pre-existing faces, 
        * the behaviour is undefined! */
      NVSG_API void setFaces( size_t pos //!< Start position inside the range of pre-existing faces
                            , const Face3* pFaces    //!<  faces to set
                            , size_t numFaces        //!<  number of faces
                            );

      //! Calculate tangents and binormals.
      /** The two dimensional texture coordinates out of texture unit \a tu are used as the parameterization for the
        * calculation. The tangents are written to texture unit \a tg and the binormals to texture unit \a bn.
        */
      NVSG_API virtual void  calculateTangentSpace( size_t tu = 0  //!<  texture unit to get parameterizing coords
                                                  , size_t tg = 1  //!<  texture unit to write tangents to
                                                  , size_t bn = 2  //!<  texture unit to write binormals to
                                                  );

    protected:
      //! Constructor
      NVSG_API Triangles(void);

      //! Partial Constructor from a GeoSet
      NVSG_API Triangles( const GeoSet &rhs );

      //! Copy Constructor
      NVSG_API Triangles( const Triangles &rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~Triangles(void);

    private:
      nvutil::RCPtr<nvutil::RCVector<Face3> >  m_faces; // triangle faces
  };

  // - - - - - - - - - - - - - - - - - - - - - - - - - 
  // inlines
  // - - - - - - - - - - - - - - - - - - - - - - - - - 

  inline bool Triangles::hasFaces()  const 
  { 
    return( !m_faces->empty() ); 
  }

  inline size_t Triangles::getNumberOfFaces()  const
  { 
    return( m_faces->size() ); 
  }

  inline const Face3 * Triangles::getFaces( void ) const 
  { 
    return( &(*m_faces)[0] );
  }
}
