// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include  "nvmath/nvmath.h"
#include  "nvmath/Vec3f.h"

namespace nvsg
{  

  //! Helper class for use in \c Animation
  /** A VNVector holds a vector of normals and vertices. It is used to produce animations on vertices/normals (instead
    * of transformations only). */
  class VNVector
  {
    public:
      std::vector<nvmath::Vec3f> normals;   //!<  vector of normals
      std::vector<nvmath::Vec3f> vertices;  //!<  vector of vertices
  };

  //! Linear interpolation between two \c VNVector
  /** Performs a linear interpolation v = v0 + alpha * ( v1 - v0 ) of each corresponding vertex and normal of the two
    * \c VNVector. The interpolated normals are normalized then.
    * \note The two \c VNVector are assumed to be the same size.
    * \return a \c VNVector with the interpolated vertices and normals. */
  inline VNVector lerp( float alpha           //!<  interpolation parameter
                      , const VNVector &v0    //!<  start value of interpolation
                      , const VNVector &v1    //!<  end value of interpolation
                      )
  {
    VNVector  vnv;

    __ASSERT( v0.vertices.size() == v1.vertices.size() );
    vnv.vertices.resize( v0.vertices.size() );
    for ( size_t i=0 ; i<v0.vertices.size() ; i++ )
    {
      vnv.vertices[i] = nvmath::lerp( alpha, v0.vertices[i], v1.vertices[i] );
    }

    __ASSERT( v0.normals.size()  == v1.normals.size() );
    vnv.normals.resize( v0.normals.size() );
    for ( size_t i=0 ; i<v0.normals.size() ; i++ )
    {
      vnv.normals[i] = normalize( lerp( alpha, v0.normals[i], v1.normals[i] ) );
    }

    return( vnv );
  }

}
