// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include <string>
#include <vector>

//! Scenegrah classes - core NVSG.
/** This namespace contains all core NVSG classes. 
  * Put all the scenegraph and node related classes in here.*/ 
namespace nvsg
{  
  //! Per-application initialization of NVSG global data
  /** An application must call this routine once before any other NVSG API call. 
    * This routine properly initializes all global data required by the NVSG API.
    *
    * In addition, this routine optionally configures the NVSG library to run safe in a 
    * multithreaded environment. This will be achieved by specifying \c true for the 
    * optional \a runInMultithreadedEnvironment parameter. For a singlethreaded environment,
    * it is strongly recommended to either omit this parameter, or to specify \c false here.
    * Making the NVSG library multithreaded safe requires some overhead, which might affect
    * the runtime behavior in a singlethreaded environment.
    *
    * It is recommended to call this routine inside the application's initial entry point. 
    * \note The nvsg runtime behaviour is undefined if an application issues nvsg API calls
    * without a prior call to this routine. */
  NVSG_API void nvsgInitialize( 
    bool runInMultithreadedEnvironment = false /*!< Specifies whether to run in a multithreaded environment, or not.
                                                    Specify \c false when running in a singlethreaded environment,
                                                    true when running in a multithreaded environment. */
  );

  //! Per-application termination of nvsg global data
  /** It is strongly recommended to call this routine when the application terminates. 
    * This routine properly frees all global data required by the nvsg API.
    * \note The nvsg runtime behaviour is undefined if an application issues nvsg API calls
    * after having called this routine. */
  NVSG_API void nvsgTerminate();

  //! Get the version string of the NVSGSDK
  NVSG_API void getVersionString(
    std::string & string  //!<  string to hold the NVSGSDK version string
  );

  //! Get the NVSGSDK name string
  NVSG_API void getSDKName(
    std::string & string  //!<  string to hold the NVSGSDK name string
  );

  //! Get the NVSGSDK copyright string
  NVSG_API void getCopyrightString(
    std::string & string  //!<  string to hold the NVSGSDK coppyright string
  );

  //! Get the NVSGSDK vendor name
  NVSG_API void getVendorName(
    std::string & string //!<  string to hold the NVSGSDK vendor name string
  );

  //! Shortcut for a vector of indices as used in TriStrips and QuadStrips.
  typedef std::vector<unsigned int> IndexSet;
}
