// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once

#include "nvsgcommon.h"

#include "nvsg/Face.h"
#include "nvtraverser/Traverser.h"

namespace nvtraverser
{
  //! Traverser that combines Drawables under one GeoNode.
  /** All Drawables of the same type that are bound to the same StateSet are combined to one Drawable. */
  class DrawableUnifyTraverser : public Traverser
  {
    public:
      //! Constructor
      NVSG_API DrawableUnifyTraverser( void );

    public:
      //! Get the 'ignore names' flag.
      /** If the 'ignore names' flag is set, the names of the GeoNodes are ignored.
        * \return true if the names will be ignored, otherwise false */
      NVSG_API bool getIgnoreNames( void ) const;

      //! Set the 'ignore names' flags.
      /** If the 'ignore names' flag is set, the names of the GeoNodes are ignored. */
      NVSG_API void setIgnoreNames( bool ignore   //!<  set true to ignore names on comparison
                                  );

    protected:
      //! Protected destructor to prevent instantiation of a DrawableUnifyTraverser on stack.
      NVSG_API virtual ~DrawableUnifyTraverser( void );

      //! Routine to handle a \link nvsg::GeoNode GeoNode \endlink while traversing the scene graph.
      /** After traversing, drawables of the same type bound to the same StateSet are combined. */
      NVSG_API virtual void handleGeoNode( const nvsg::GeoNode * p //!< Points to the currently visited GeoNode object.
                                         );

      //! Routine to handle a \link nvsg::StateSet StateSet \endlink object while traversing the scene graph.
      /** Just holds the StateSet as the current one for subsequent comparisons. */
      NVSG_API virtual void handleStateSet( const nvsg::StateSet * p //!< Points to the currently visited StateSet object.
                                          );

      //! Routine to handle a \link nvsg::Quads Quads \endlink object while traversing the scene graph.
      /** If the Quads \a p is bound to the same StateSet as a previously traversed one, they are marked for combining. */
      NVSG_API virtual void handleQuads( const nvsg::Quads * p //!< Points to the currently visited Quads object.
                                       );

      //! Routine to handle a \link nvsg::AnimatedQuads AnimatedQuads \endlink object while traversing the scene graph.
      /** AnimatedQuads are explicitly ignored here! */
      NVSG_API virtual void handleAnimatedQuads( const nvsg::AnimatedQuads * p //!< Points to the currently visited AnimatedQuads object.
                                               );

      //! Routine to handle a \link nvsg::QuadStrips QuadStrips \endlink object while traversing the scene graph.
      /** If the QuadStrips \a p is bound to the same StateSet as a previously traversed one, they are marked for combining. */
      NVSG_API virtual void handleQuadStrips( const nvsg::QuadStrips * p //!< Points to the currently visited QuadStrips object.
                                            );

      //! Routine to handle a \link nvsg::Triangles Triangles \endlink object while traversing the scene graph.
      /** If the Triangles \a p is bound to the same StateSet as a previously traversed one, they are marked for combining. */
      NVSG_API virtual void handleTriangles( const nvsg::Triangles * p //!< Points to the currently visited Triangles object.
                                           );

      //! Routine to handle a \link nvsg::AnimatedTriangles AnimatedTriangles \endlink object while traversing the scene graph.
      /** AnimatedTriangles are explicitly ignored here! */
      NVSG_API virtual void handleAnimatedTriangles( const nvsg::AnimatedTriangles * p //!< Points to the currently visited AnimatedTriangles object.
                                                   );

      //! Routine to handle a \link nvsg::SkinnedTriangles SkinnedTriangles \endlink object while traversing the scene graph.
      /** SkinnedTriangles are explicitly ignored here! */
      NVSG_API virtual void handleSkinnedTriangles( const nvsg::SkinnedTriangles * p //!< Points to the currently visited SkinnedTriangles object.
                                                  );

      //! Routine to handle a \link nvsg::TriStrips TriStrips \endlink object while traversing the scene graph.
      /** If the TriStrips \a p is bound to the same StateSet as a previously traversed one, they are marked for combining. */
      NVSG_API virtual void handleTriStrips( const nvsg::TriStrips * p //!< Points to the currently visited TriStrips object.
                                           );

    protected:
      void combine( nvsg::GeoSet *p0, const nvsg::GeoSet *p1 );
      void combine( nvsg::Quads *p0, const nvsg::Quads *p1 );
      void combine( nvsg::QuadStrips *p0, const nvsg::QuadStrips *p1 );
      void combine( nvsg::Triangles *p0, const nvsg::Triangles *p1 );
      void combine( nvsg::TriStrips *p0, const nvsg::TriStrips *p1 );

    private:
      bool  m_ignoreNames;
      const nvsg::StateSet *                                                  m_currentStateSet;
      std::vector<std::pair<const nvsg::Drawable*,const nvsg::StateSet*> >    m_drawableToDelete;
      std::vector<std::pair<const nvsg::Quads*,const nvsg::StateSet*> >       m_quadsToKeep;
      std::vector<std::pair<const nvsg::QuadStrips*,const nvsg::StateSet*> >  m_quadStripsToKeep;
      std::vector<std::pair<const nvsg::Triangles*,const nvsg::StateSet*> >   m_trianglesToKeep;
      std::vector<std::pair<const nvsg::TriStrips*,const nvsg::StateSet*> >   m_triStripsToKeep;
      std::vector<nvsg::Face3>    m_shiftedFaces3;
      std::vector<nvsg::Face4>    m_shiftedFaces4;
      std::vector<nvsg::IndexSet> m_shiftedIndexSets;
  };

  inline bool DrawableUnifyTraverser::getIgnoreNames( void ) const
  {
    return( m_ignoreNames );
  }

  inline void DrawableUnifyTraverser::setIgnoreNames( bool ignore )
  {
    m_ignoreNames = ignore;
  }

}
