// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#include "nvsgcommon.h"

#include "nvtraverser/Traverser.h"

namespace nvtraverser
{
  //! Traverser that deletes unneccessary Groups.
  /** A Group is unneccessary if it holds only one child. In that case, that child replaces the group in it's parent. */
  class GroupUnifyTraverser : public Traverser
  {
    public:
      //! Constructor
      NVSG_API GroupUnifyTraverser( void );

    public:
      //! Get the 'ignore names' flag.
      /** If the 'ignore names' flag is set, the names of the GeoNodes are ignored.
        * \return true if the names will be ignored, otherwise false */
      NVSG_API bool getIgnoreNames( void ) const;

      //! Set the 'ignore names' flags.
      /** If the 'ignore names' flag is set, the names of the GeoNodes are ignored. */
      NVSG_API void setIgnoreNames( bool ignore   //!<  set true to ignore names on comparison
                                  );

    protected:
      //! Protected destructor to prevent instantiation of a GroupUnifyTraverser on stack.
      NVSG_API virtual ~GroupUnifyTraverser( void );

      //! Routine to handle a \link nvsg::Group Group \endlink while traversing the scene graph.
      /** This routine records all Groups with one child. */
      NVSG_API virtual void handleGroup( const nvsg::Group * p //!< Points to the currently visited Group object.
                                         );

      //! This function will be called from the framework immediately before a Group is to be traversed.
      /** Before traversing a Group, a list of Groups holding one child is created. */
      NVSG_API virtual void preTraverseGroup( const nvsg::Group *p  //!< Points to the Group that will be traversed immediately after this call.
                                            );

      //! This function will be called from the framework immediately after a Group has been traversed.
      /** After traversing a Group, children with one child are replaced by that child. */
      NVSG_API virtual void postTraverseGroup( const nvsg::Group *p  //!< Points to the Group that was traversed immediately before this call.
                                             );

    private:
      bool  m_ignoreNames;
      std::stack<std::vector<const nvsg::Group *>*> m_oneChildGroups;
  };

  inline bool GroupUnifyTraverser::getIgnoreNames( void ) const
  {
    return( m_ignoreNames );
  }

  inline void GroupUnifyTraverser::setIgnoreNames( bool ignore )
  {
    m_ignoreNames = ignore;
  }
}
