// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include  <list>
#include  <set>

#include "nvsgcommon.h"

#include  "nvsg/nvsg.h"
#include  "nvsg/Face.h"
#include  "nvtraverser/Traverser.h"

namespace nvtraverser
{  

  //! Traverser to convert Triangles to TriStrips, and Quads to QuadStrips
  /** It is strongly recommended to use a CondenseTraverser before using this StrippingTraverser, because it 
    * works only on indices. */
  class StrippingTraverser : public Traverser
  {
    public:
      //! Constructor
      NVSG_API StrippingTraverser(void);

      //! Get the minimal lengths for TriStrips and QuadStrips.
      /** A strip with less than the minimal length is reconverted to triangles or quads. */
      void  getMinimalLengths( size_t &triStripLength, size_t &quadStripLength ) const;

      //! Set the minimal lengths for TriStrips and QuadStrips.
      /** A strip with less than the minimal length is reconverted to triangles or quads. */
      void  setMinimalLengths( size_t triStripLength, size_t quadStripLength );

    protected:
      //! Protected destructor to prevent instantiation of a StrippingTraverser.
      NVSG_API virtual ~StrippingTraverser(void);

      //! Handle an AnimatedQuads object.
      NVSG_API virtual void  handleAnimatedQuads( 
        const nvsg::AnimatedQuads *p      //!<  AnimatedQuads to handle
      );

      //! Handle an AnimatedTriangles object.
      NVSG_API virtual void  handleAnimatedTriangles( 
        const nvsg::AnimatedTriangles *p  //!<  AnimatedTriangles to handle
      );

      //! Handle a GeoNode object.
      NVSG_API virtual void  handleGeoNode( 
        const nvsg::GeoNode *p            //!<  GeoNode to handle
      );

      //! Handle a Quads object.
      NVSG_API virtual void  handleQuads(
        const nvsg::Quads *p              //!<  Quads to handle
      );

      //! Handle an SkinnedTriangles object.
      NVSG_API virtual void  handleSkinnedTriangles(
        const nvsg::SkinnedTriangles *p   //!<  SkinnedTriangles to handle
      );

      //! Handle a Triangles object.
      NVSG_API virtual void  handleTriangles(
        const nvsg::Triangles *p          //!<  Triangles to handle
      );

    private:
      void    addQuadStrip( const std::list<unsigned int> &vertexList );
      void    addTriangle( const nvsg::IndexSet &strip, size_t i0, size_t i1, size_t i2 );
      void    addTriStrip( const std::list<unsigned int> &vertexList );
      void    clearFaceData( void );
      void    connectFaces( const nvsg::Face3 *faces, size_t fi, size_t i0, size_t i1 );
      void    connectFaces( const nvsg::Face4 *faces, size_t fi, size_t i0, size_t i1 );
      void    copyVertexAttributeData( const nvsg::GeoSet *pFrom, nvsg::GeoSet *pTo );
      void    cutConnections( const std::list<size_t> &list, size_t vpf );
      void    determineList( const nvsg::Face3 *faces, size_t fi, size_t le, std::list<size_t> &faceList,
                             std::list<unsigned int> &vertexList );
      void    determineList( const nvsg::Face4 *faces, size_t fi, size_t le, std::list<size_t> &faceList,
                             std::list<unsigned int> &vertexList );

    private:
      std::vector<size_t>                m_faceConnectionCounts;
      std::vector<size_t>                m_faceConnections;
      std::set<size_t>                   m_faceSets[5];
      size_t                             m_minimalQuadStrip;
      size_t                             m_minimalTriStrip;
      std::vector<nvsg::Face4>                 m_quads;
      std::vector<nvsg::IndexSet>              m_strips;
      std::vector<nvsg::Face3>                 m_triangles;
      std::vector<std::list<size_t> >    m_verticesToFaceLists;

      std::vector<std::pair<const nvsg::Quads*,std::pair<const nvsg::QuadStrips*,const nvsg::Quads*> > >         m_quadReplaces;
      std::vector<std::pair<const nvsg::Triangles*,std::pair<const nvsg::TriStrips*,const nvsg::Triangles*> > >  m_triReplaces;
  };

  inline  void StrippingTraverser::getMinimalLengths( size_t &triStripLength, size_t &quadStripLength ) const
  {
    triStripLength  = m_minimalTriStrip;
    quadStripLength = m_minimalQuadStrip;
  }

  inline  void  StrippingTraverser::setMinimalLengths( size_t triStripLength, size_t quadStripLength )
  {
    m_minimalTriStrip   = triStripLength;
    m_minimalQuadStrip  = quadStripLength;
  }
}
