// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include <stack>

#include "nvsgcommon.h"

#include "nvsg/Object.h"
#include "nvutil/MemFunTbl.h"
#include "nvutil/SWMRSync.h"

// forward declarations
namespace nvmath
{
  class Trafo;
}
namespace nvsg
{
  class AnimatedQuads;
  class AnimatedTransform;
  class AnimatedTriangles;
  class CgFx;
  class CullData;
  class DirectedLight;
  class Drawable;
  class FaceAttribute;
  class GeoNode;
  class GeoSet;
  class Group;
  class HighlightBox;
  class HighlightEdges;
  class JitterCamera;
  class LightSource;
  class LOD;
  class Material;
  class Node;
  class ParallelCamera;
  class PerspectiveCamera;
  class PointLight;
  class Quads;
  class QuadStrips;
  class SimpleCamera;
  class SkinnedTriangles;
  class SpotLight;
  class StateSet;
  class StereoCamera;
  class TextureAttribute;
  class TextureAttributeItem;
  class Transform;
  class Triangles;
  class TriStrips;
  class Switch;
  class Scene;
  class ViewState;
  class VNVector;
  template<typename T> class FramedAnimation;
  template<typename T> class LinearInterpolatedAnimation;
}

namespace nvtraverser
{

//! Base class providing an interface for read-write traversing of a given scene graph.
/** A traverser serves as a link between a scene graph and a defined operation to be performed on the scene graph.
  * That is, according to a certain scene graph's hierarchy, a traverser iterates over all scene graph components
  * and performes a defined operation for each by means of an overloadable handler routine.
  *
  * The Traverser class offers base functionality for all traversers that intend to provide read-write operations 
  * on a given scene graph. For all known and concrete \link nvsg::Object Objects \endlink this base class provides 
  * a set of overloadable handler routines, each ensuring that all components following the actual object in the 
  * graph hierarchy will be traversed correctly. Hence, for derived traversers it is recommended to always call the 
  * base implementation of an overloaded handler routine for traversing purposes. 
  *
  * To provide new read-only operations to be applied on known and concrete components or objects arranged in a 
  * scene graph, it is sufficient to derive a new traverser from either Traverser or one of its derived 
  * traversers, and overload the corresponding handler routines as needed.
  */ 
class Traverser : public nvutil::RCObject
{
public:

  //! Start traversing a scene graph.
  /** This serves as the main entry point for clients to traverse a certain scene graph.
    * The function waits to gain read access to the \link nvsg::Scene Scene \endlink object pointed to by \a scene. 
    * If an optional \link nvsg::ViewState ViewState \endlink object is provided, 
    * the function also waits to gain read access to the ViewState object pointed to by \a viewState.
    * After the function gains read access to the Scene object (and the ViewState object, if provided), it calls
    * the protected and overloadable \link nvtraverser::Traverser::doApply doApply \endlink member function 
    * to start traversing the corresponding scene graph.
    * After traversing is completed, the function releases the read access to the Scene object and, 
    * if provided, the optional ViewState object.
    * \note The behavior is undefined if \a scene points to an invalid location. */
  NVSG_API void apply( 
    const nvsg::ViewState * viewState //!< Specifies an optional ViewState object to be applied prior to traversing the scene graph. 
  , const nvsg::Scene * scene //!< Abstracts the scene graph to be traversed.
  );

protected:

  //! Default construction on instantiation.
  /** The constructor is protected, and hence, a Traverser object cannot be instantiated directly but only
    * by deriving from Traverser. */
  NVSG_API Traverser( bool readOnly = true );

  //! Protected destructor to prevent instantiation of a Traverser.
  NVSG_API virtual ~Traverser(void);

  //! Initiate traversing a given scene graph.
  /** This overloadable function is called from the traverser's entry point \link nvtraverser::Traverser::apply
    * apply \endlink. The base implementation first initiates the traversing of the scene's global states, if available.
    * After that the scene itself is traversed. If an optional \link nvsg::ViewState ViewState ViewState \endlink is 
    * passed through \a viewState, it is applied prior to traversing the scene.
    *
    * Custom traversers may override this function in order to implement initial work to be done before traversing
    * the scene. For derived classes to ensure proper traversing of the scene, it is recommended to always call
    * the base implementation after the initial work has been done. */
  NVSG_API virtual void  doApply( 
    const nvsg::ViewState * viewState    //!< Optional ViewState to be applied prior to traversing the scene.
  , const nvsg::Scene * scene            //!< Points to the scene to be traversed.
  );

  //! Add a handler routine for a concrete \link nvsg::Object Object \endlink derived object.
  /** The function registers the handler routine specified by \a handler to handle a concrete Object that is 
    * explicitly identified by \a objectCode.
    *
    * A handler routine must be a member function of a Traverser-derived class. It must return void and expect a
    * pointer to the concrete object as a first parameter, and can have one additional (optional) argument of arbitrary type.
    * A handler routine must not have a default argument! If a handler routine is intended to remain overloadable, it 
    * should be declared virtual.
    *
    * \note A possibly pre-registered handler routine for the specified object will be overridden by the new handler.
    *
    * \remarks For derived classes intended to provide new operations for known objects arranged in a scene graph,
    * it is strongly recommended not to add new handler routines for a certain known objects but to overload the 
    * corresponding handler routines. In this context, we refer to 'known objects' as objects that are known by the
    * Traverser base class and for which a corresponding handler routine is provided.
    *
    * However, \c addObjectHandler is indispensible if newly invented objects, and hence, objects that are unknown by the 
    * Traverser base class, need to be considered for traversing. */
  template<typename HandlerType>
  void addObjectHandler(
    unsigned int objectCode //!< Object code to identify an object type at run-time.
  , HandlerType handler //!< Specifies the address of the handler routine.
  );

  //! Start traversing a general object.
  /** The function first waits to gain read access to the object pointed to by \a object.
    * After that, it calls the corresponding and overloadable handler routine.  
    * After the handler routine returns, the function releases the read access to the object. 
    * \note The behavior is undefined if \a p points to an invalid location. */
  void traverseObject(
    const nvsg::Object* object //!< Points to the object to be traversed.
  );

  //! Start traversing a general object.
  /** The function first waits to gain read access to the object pointed to by \a object.
    * After that, it calls the corresponding and overloadable handler routine with the additional argument \a arg.  
    * After the handler routine returns, the function releases the read access to the object. 
    * \note The behavior is undefined if \a p points to an invalid location.
    * The behavior is also undefined if the additional argument doesn't match that of the corresponding handler routine. */
  template <typename T>
  void traverseObject(
    const nvsg::Object* object //!< Points to the object to be traversed.
  , T arg //!< Specifies an additional argument that will be passed to the handler routine.
  );

  //
  // Routines to handle concrete Node objects while traversing the tree
  //

  //! Routine to handle a \link nvsg::GeoNode GeoNode \endlink while traversing the scene graph.
  /** This routine gets called for each GeoNode object that will be visited while traversing the scene graph.
    * The currently visited GeoNode object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleGeoNode(
    const nvsg::GeoNode * p //!< Points to the currently visited GeoNode object.
  );
                                      
  //! Routine to handle a \link nvsg::Group Group \endlink node while traversing the scene graph.
  /** This routine gets called for each Group object that will be visited while traversing the scene graph.
    * The currently visited Group object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleGroup(
    const nvsg::Group * p //!< Points to the currently visited Group object.
  );
  
  //! Routine to handle a \link nvsg::LOD LOD \endlink (Level Of Detail) node while traversing the scene graph.
  /** This routine gets called for each LOD object that will be visited while traversing the scene graph.
    * The currently visited LOD object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleLOD(
    const nvsg::LOD * p //!< Points to the currently visited LOD object.
  );
                                      
  //! Routine to handle a \link nvsg::Switch Switch \endlink node while traversing the scene graph.
  /** This routine gets called for each Switch object that will be visited while traversing the scene graph.
    * The currently visited Switch object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleSwitch(
    const nvsg::Switch * p //!< Points to the currently visited Switch object.
  );
                                      
  //! Routine to handle a \link nvsg::Transform Transform \endlink node while traversing the scene graph.
  /** This routine gets called for each Transform object that will be visited while traversing the scene graph.
    * The currently visited Transform object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleTransform(
    const nvsg::Transform * p //!< Points to the currently visited Transform object.
  );
  
  //! Routine to handle a \link nvsg::AnimatedTransform AnimatedTransform \endlink node while traversing the scene graph.
  /** This routine gets called for each AnimatedTransform object that will be visited while traversing the scene graph.
    * The currently visited AnimatedTransform object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleAnimatedTransform(
    const nvsg::AnimatedTransform * p //!< Points to the currently visited AnimatedTransform object.
  );
  
  //! Routine to handle a \link nvsg::DirectedLight DirectedLight \endlink node while traversing the scene graph.
  /** This routine gets called for each DirectedLight object that will be visited while traversing the scene graph.
    * The currently visited DirectedLight object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleDirectedLight(
    const nvsg::DirectedLight * p //!< Points to the currently visited DirectedLight object.
  );
  
  //! Routine to handle a \link nvsg::PointLight PointLight \endlink node while traversing the scene graph.
  /** This routine gets called for each PointLight object that will be visited while traversing the scene graph.
    * The currently visited PointLight object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handlePointLight(
    const nvsg::PointLight * p //!< Points to the currently visited PointLight object.
  );

  //! Routine to handle a \link nvsg::SpotLight SpotLight \endlink node while traversing the scene graph.
  /** This routine gets called for each SpotLight object that will be visited while traversing the scene graph.
    * The currently visited SpotLight object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleSpotLight(
    const nvsg::SpotLight * p //!< Points to the currently visited SpotLight object.
  );

  //
  // Routines to handle concrete Drawable objects while traversing the tree
  //

  //! Routine to handle a \link nvsg::Quads Quads \endlink object while traversing the scene graph.
  /** This routine gets called for each Quads object that will be visited while traversing the scene graph.
    * The currently visited Quads object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleQuads(
    const nvsg::Quads * p //!< Points to the currently visited Quads object.
  );

  //! Routine to handle a \link nvsg::AnimatedQuads AnimatedQuads \endlink object while traversing the scene graph.
  /** This routine gets called for each AnimatedQuads object that will be visited while traversing the scene graph.
    * The currently visited AnimatedQuads object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleAnimatedQuads(
    const nvsg::AnimatedQuads * p //!< Points to the currently visited AnimatedQuads object.
  );

  //! Routine to handle a \link nvsg::QuadStrips QuadStrips \endlink object while traversing the scene graph.
  /** This routine gets called for each QuadStrips object that will be visited while traversing the scene graph.
    * The currently visited QuadStrips object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleQuadStrips(
    const nvsg::QuadStrips * p //!< Points to the currently visited QuadStrips object.
  );

  //! Routine to handle a \link nvsg::Triangles Triangles \endlink object while traversing the scene graph.
  /** This routine gets called for each Triangles object that will be visited while traversing the scene graph.
    * The currently visited Triangles object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleTriangles(
    const nvsg::Triangles * p //!< Points to the currently visited Triangles object.
  );

  //! Routine to handle a \link nvsg::AnimatedTriangles AnimatedTriangles \endlink object while traversing the scene graph.
  /** This routine gets called for each AnimatedTriangles object that will be visited while traversing the scene graph.
    * The currently visited AnimatedTriangles object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleAnimatedTriangles(
    const nvsg::AnimatedTriangles * p //!< Points to the currently visited AnimatedTriangles object.
  );

  //! Routine to handle a \link nvsg::SkinnedTriangles SkinnedTriangles \endlink object while traversing the scene graph.
  /** This routine gets called for each SkinnedTriangles object that will be visited while traversing the scene graph.
    * The currently visited SkinnedTriangles object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleSkinnedTriangles(
    const nvsg::SkinnedTriangles * p //!< Points to the currently visited SkinnedTriangles object.
  );

  //! Routine to handle a \link nvsg::TriStrips TriStrips \endlink object while traversing the scene graph.
  /** This routine gets called for each TriStrips object that will be visited while traversing the scene graph.
    * The currently visited TriStrips object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleTriStrips(
    const nvsg::TriStrips * p //!< Points to the currently visited TriStrips object.
  );

  //
  // Routines to handle either a StateSet or concrete StateAttribute objects while traversing the tree
  //

  //! Routine to handle a \link nvsg::StateSet StateSet \endlink object while traversing the scene graph.
  /** This routine gets called for each StateSet object that will be visited while traversing the scene graph.
    * The currently visited StateSet object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleStateSet(
    const nvsg::StateSet * p //!< Points to the currently visited StateSet object.
  );

  //! Routine to handle a \link nvsg::CgFx CgFx \endlink attribute while traversing the scene graph.
  /** This routine gets called for each CgFx object that will be visited while traversing the scene graph.
    * The currently visited CgFx object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleCgFx(
    const nvsg::CgFx * p //!< Points to the currently visited CgFx object.
  );

  //! Routine to handle a \link nvsg::FaceAttribute FaceAttribute \endlink attribute while traversing the scene graph.
  /** This routine gets called for each FaceAttribute object that will be visited while traversing the scene graph.
    * The currently visited FaceAttribute object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleFaceAttribute(
    const nvsg::FaceAttribute * p //!< Points to the currently visited FaceAttribute object.
  );

  //! Routine to handle a \link nvsg::Material Material \endlink attribute while traversing the scene graph.
  /** This routine gets called for each Material object that will be visited while traversing the scene graph.
    * The currently visited Material object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleMaterial(
    const nvsg::Material * p //!< Points to the currently visited Material object.
  );

  //! Routine to handle a \link nvsg::TextureAttribute TextureAttribute \endlink attribute while traversing the scene graph.
  /** This routine gets called for each TextureAttribute object that will be visited while traversing the scene graph.
    * The currently visited TextureAttribute object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleTextureAttribute(
    const nvsg::TextureAttribute * p //!< Points to the currently visited TextureAttribute object.
  );

  //! Routine to handle a \link nvsg::TextureAttributeItem TextureAttributeItem \endlink object while traversing the scene graph.
  /** This routine gets called for each TextureAttributeItem object that will be visited while traversing the scene graph.
    * The currently visited TextureAttributeItem object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleTextureAttributeItem(
    const nvsg::TextureAttributeItem * p //!< Points to the currently visited TextureAttributeItem object.
  , size_t tu //!< Specifies the zero based index of the texture unit to bind the item to.
  );

  //
  // Routines to handle Camera objects while traversing the tree
  //

  //! Routine to handle a \link nvsg::JitterCamera JitterCamera \endlink object while traversing the scene graph.
  /** This routine gets called for each JitterCamera object that will be visited while traversing the scene graph.
    * The currently visited JitterCamera object is pointed to by \a p.
    * If the root node pointed to by \a root is valid (that is, \a root is not a null pointer), 
    * this function traverses the root node after the camera settings have been evaluated.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleJitterCamera(
    const nvsg::JitterCamera * p //!< Points to the currently visited JitterCamera object.
  , const nvsg::Node * root //!< If valid, points to the root node to be traversed with the actual camera settings.
  );

  //! Routine to handle a \link nvsg::ParallelCamera ParallelCamera \endlink object while traversing the scene graph.
  /** This routine gets called for each ParallelCamera object that will be visited while traversing the scene graph.
    * The currently visited ParallelCamera object is pointed to by \a p.
    * If the root node pointed to by \a root is valid (that is, \a root is not a null pointer), 
    * this function traverses the root node after the camera settings have been evaluated.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleParallelCamera(
    const nvsg::ParallelCamera * p //!< Points to the currently visited ParallelCamera object.
  , const nvsg::Node * root //!< If valid, points to the root node to be traversed with the actual camera settings.
  );

  //! Routine to handle a \link nvsg::PerspectiveCamera PerspectiveCamera \endlink object while traversing the scene graph.
  /** This routine gets called for each PerspectiveCamera object that will be visited while traversing the scene graph.
    * The currently visited PerspectiveCamera object is pointed to by \a p.
    * If the root node pointed to by \a root is valid (that is, \a root is not a null pointer), 
    * this function traverses the root node after the camera settings have been evaluated.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handlePerspectiveCamera(
    const nvsg::PerspectiveCamera * p //!< Points to the currently visited PerspectiveCamera object.
  , const nvsg::Node * root //!< If valid, points to the root node to be traversed with the actual camera settings.
  );

  //! Routine to handle a \link nvsg::StereoCamera StereoCamera \endlink object while traversing the scene graph.
  /** This routine gets called for each StereoCamera object that will be visited while traversing the scene graph.
    * The currently visited StereoCamera object is pointed to by \a p.
    * If the root node pointed to by \a root is valid (that is, \a root is not a null pointer), 
    * this function traverses the root node after the camera settings have been evaluated.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleStereoCamera(
    const nvsg::StereoCamera * p //!< Points to the currently visited StereoCamera object.
  , const nvsg::Node * root //!< If valid, points to the root node to be traversed with the actual camera settings.
  );

  //
  // Routines to handle Animation objects while traversing the tree
  //

  //! Routine to handle a \link nvsg::Animation FramedAnimation<Trafo> \endlink object while traversing the scene graph.
  /** This routine gets called for each FramedAnimation<Trafo> object that will be visited while traversing the scene graph.
    * The currently visited FramedAnimation<Trafo> object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleFramedTrafoAnimation(
    const nvsg::FramedAnimation<nvmath::Trafo> * p //!< Points to the currently visited Animation object.
  );

  //! Routine to handle a \link nvsg::Animation FramedAnimation<VNVector> \endlink object while traversing the scene graph.
  /** This routine gets called for each FramedAnimation<VNVector> object that will be visited while traversing the scene graph.
    * The currently visited FramedAnimation<VNVector> object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleFramedVNVectorAnimation(
    const nvsg::FramedAnimation<nvsg::VNVector> * p //!< Points to the currently visited Animation object.
  );

  //! Routine to handle a \link nvsg::Animation LinearInterpolatedAnimation<Trafo> \endlink object while traversing the scene graph.
  /** This routine gets called for each LinearInterpolatedAnimation<Trafo> object that will be visited while traversing the sc
ene graph.
    * The currently visited LinearInterpolatedAnimation<Trafo> object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleLinearInterpolatedTrafoAnimation(
    const nvsg::LinearInterpolatedAnimation<nvmath::Trafo> * p //!< Points to the currently visited Animation object.
  );

  //! Routine to handle a \link nvsg::Animation LinearInterpolatedAnimation<VNVector> \endlink object while traversing the scene graph.
  /** This routine gets called for each LinearInterpolatedAnimation<VNVector> object that will be visited while traversing the
 scene graph.
    * The currently visited LinearInterpolatedAnimation<VNVector> object is pointed to by \a p.
    * The base implementation ensures that all underlying components will be traversed in a correct way.
    * For derived classes, it is recommended to always call the base implementation for traversing purposes. 
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleLinearInterpolatedVNVectorAnimation(
    const nvsg::LinearInterpolatedAnimation<nvsg::VNVector> * p //!< Points to the currently visited Animation object.
  );

  //
  // Routines to handle Highlight objects while traversing the tree
  //

  //! Routine to handle a \link nvsg::HighlightBox HighlightBox \endlink object while traversing the scene graph.
  /** This routine gets called for each HighlightBox object that will be visited while traversing the scene graph.
    * The currently visited Highlight object is pointed to by \a p.
    * \remarks A HighlighBox always is attached to its corresponding \link nvsg::GeoSet GeoSet \endlink object as a 
    * \link nvsg::HighlightObject HighlightObject \endlink.
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleHighlightBox(
    const nvsg::HighlightBox * p //!< Points to the currently visited HighlightBox object.
  , const nvsg::GeoSet * geoSet //!< Points to the corresponding GeoSet object.
  );

  //! Routine to handle a \link nvsg::HighlightEdges HighlightEdges \endlink object while traversing the scene graph.
  /** This routine gets called for each HighlightEdges object that will be visited while traversing the scene graph.
    * The currently visited Highlight object is pointed to by \a p.
    * \remarks A HighlighEdges always is attached to its corresponding \link nvsg::GeoSet GeoSet \endlink object as a 
    * \link nvsg::HighlightObject HighlightObject \endlink.
    * \note: The behavior is undefined if \a p points to an invalid location. */ 
  NVSG_API virtual void handleHighlightEdges(
    const nvsg::HighlightEdges * p //!< Points to the currently visited HighlightEdges object.
  , const nvsg::GeoSet * geoSet //!< Points to the corresponding GeoSet object.
  );


  //! This function will be called from the framework immediately before the scene graph is traversed.
  /** If the function returns \c false, further traversing will be skipped. 
    * \remarks Custom traversers can override this function for any tasks that need to be done immediately 
    * before the scene graph is traversed. */
  NVSG_API virtual bool preTraverseRoot( 
    const nvsg::SimpleCamera *p  //!<  Camera that will be applied while traversing the scene graph.
  , const nvsg::Node *root       //!<  Root node to start traversing from.
  );

  //! This function will be called from the framework immediately after the scene graph has been traversed.
  /** remarks Custom traversers can override this function for any tasks that need to be done immediately 
    * after the scene graph has been traversed. */
  NVSG_API virtual void postTraverseRoot( 
    const nvsg::SimpleCamera *p //!<  Camera that was applied while traversing the scene graph.
  , const nvsg::Node *root      //!<  Root node were traversing has been started from.
  );

  //! This function will be called from the framework immediately before a Group is traversed.
  /** \remarks Custom traversers can override this function for any tasks that need to be done immediately 
    * before the Group pointed to by \a p is traversed. */
  NVSG_API virtual void preTraverseGroup( 
    const nvsg::Group *p  //!< Points to the Group that will be traversed immediately after this call.
  );

  //! This function will be called from the framework immediately after a Group has been traversed.
  /** \remarks Custom traversers can override this function for any tasks that need to be done immediately 
    * after the Group pointed to by \a p has been traversed. */
  NVSG_API virtual void postTraverseGroup( 
    const nvsg::Group *p  //!< Points to the Group that was traversed immediately before this call.
  );

  //! This function will be called from the framework immediately before a Drawable is traversed.
  /** If the function returns \c false, further traversing will be skipped. 
    * \remarks Custom traversers can override this function for any tasks that need to be done immediately 
    * before the Drawable is traversed. */
  NVSG_API virtual bool preTraverseDrawable( const nvsg::Drawable *p    //!<  Drawable that is going to be traversed
                                           );

  //! This function will be called from the framework when a \link nvsg::LightSource LightSource \endlink node is visited.
  /** remarks Custom traversers can override this function for special tasks that need to be performed when
    * visiting a LightSource node. */
  NVSG_API virtual void traverseLight(
    const nvsg::LightSource * p //!< Points to the currently visited LightSource node.
  );

  //! Get a writable pointer to a constant object.
  /** While traversing, all modifications of objects have to be embraced by beginEdit/endEdit.
    * \returns A non-constant pointer to the object. */
  template<typename T>
  T * beginEdit( const T * obj  //!<  object to get a non-constant pointer to
               ) const
  {
    __TRACE();
    __ASSERT( ! m_readOnly );   // only non-readOnly traverser are allowed to edit
    return const_cast<T *>(obj);
  }

  //! End the editing operation on a constant object.
  /** While traversing, all modifications of objects have to be embraced by beginEdit/endEdit.
    * \returns A constant pointer to the object. */
  template<typename T>
  const T * endEdit( const T * obj  //!<  object to end editing operations on
                   ) const
  {
    __TRACE();
    return( obj );
  }

protected:
  std::stack<size_t>            m_lodStack;     //!< holds the index of the current LOD level 
  std::stack<const nvsg::CullData *>  m_cullStack;    //!< holds the cull data corresponding to the traversed tree
  std::stack<const nvsg::Group *>     m_groupStack;   //!< stack of groups just traversing
  
  const nvsg::StateSet * m_defaultStateSet;           //!<  private copy of the scene's default states  
  const nvsg::StateSet * m_overrideStateSet;          //!<  private copy of the scnee's override states


private:
  // geonode handling helper
  void traverse(const nvsg::GeoNode * p);
  void traverse(const nvsg::GeoNode * p, size_t i);
  // group handling helper
  template <typename GrpType>
  void traverseGroup(const GrpType * p);
  // general
  void traverse(const nvsg::Group * p);
  void traverse(const nvsg::Group * p, const nvsg::CullData * cd);
  // LOD specialization
  void traverse(const nvsg::LOD * p);
  void traverse(const nvsg::LOD * p, const nvsg::CullData * cd);
  // Switch specialization
  void traverse(const nvsg::Switch * p);
  void traverse(const nvsg::Switch * p, const nvsg::CullData * cd);
  // traverse a root node with a camera applied
  void traverse(const nvsg::SimpleCamera *p, const nvsg::Node *root);

  bool waitToRead( const nvutil::ISWMRSync *obj ) const;
  void doneReading( const nvutil::ISWMRSync *obj ) const;
  bool waitToWrite( const nvutil::ISWMRSync *obj ) const;
  void doneWriting( const nvutil::ISWMRSync *obj ) const;

private:
  nvutil::MemFunTbl<Traverser> * m_mftbl; // member function table
                                  // accesible for derived classes via addObjectHandler API
  bool m_readOnly;
  bool (Traverser::*m_pfnLock)( const nvutil::ISWMRSync  *obj ) const;
  void (Traverser::*m_pfnUnlock)( const nvutil::ISWMRSync  *obj ) const;
};

template<typename HandlerType>
inline void Traverser::addObjectHandler(unsigned int objectCode, HandlerType handler)
{
  m_mftbl->addEntry(objectCode, handler); 
}

inline bool Traverser::waitToRead( const nvutil::ISWMRSync *obj ) const
{
  return( obj->waitToRead() );
}

inline void Traverser::doneReading( const nvutil::ISWMRSync  *obj ) const
{
  return( obj->doneReading() );
}

inline bool Traverser::waitToWrite( const nvutil::ISWMRSync  *obj ) const
{
  return( obj->waitToWrite() );
}

inline void Traverser::doneWriting( const nvutil::ISWMRSync  *obj ) const
{
  return( obj->doneWriting() );
}

inline void Traverser::traverseObject(const nvsg::Object* object)
{
  __TRACE();
  __ASSERT(object->getObjectCode()!=nvsg::OC_INVALID);
  unsigned int oc = object->getObjectCode();
  
  if ( m_mftbl->testEntry(oc) && (this->*m_pfnLock)( object ) )
  {
#if !defined(_WIN64) && defined(_WIN32) && defined(_DEBUG)
    void * __esp__0;
    void * __esp__1;
    __asm mov __esp__0, esp
#endif
    (this->*(void (Traverser::*)(const nvsg::Object*))(*m_mftbl)[oc])(object);
#if !defined(_WIN64) && defined(_WIN32) && defined(_DEBUG)
    __asm mov __esp__1, esp
    __ASSERT(__esp__1==__esp__0);
#endif
    (this->*m_pfnUnlock)( object );
  }
}

template <typename T>
inline void Traverser::traverseObject( const nvsg::Object* object, T arg)
{
  __TRACE();
  __ASSERT(object->getObjectCode()!=nvsg::OC_INVALID);
  unsigned int oc = object->getObjectCode();
  
  if ( m_mftbl->testEntry(oc) && (this->*m_pfnLock)( object ) )
  {
#if !defined(_WIN64) && defined(_WIN32) && defined(_DEBUG)
    void * __esp__0;
    void * __esp__1;
    __asm mov __esp__0, esp
#endif
    (this->*(void (Traverser::*)(const nvsg::Object*, T))(*m_mftbl)[oc])(object, arg);
#if !defined(_WIN64) && defined(_WIN32) && defined(_DEBUG)
    __asm mov __esp__1, esp
    __ASSERT(__esp__1==__esp__0);
#endif
    (this->*m_pfnUnlock)( object );
  }
}

} // namespace
