// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include  "nvutil/RCObject.h"

namespace nvutil
{
  //! PlugInCallback base class
  /** A PlugInCallback object can be used to report warnings and errors that happen while using a PlugIn. It is applied
    * to a PlugIn via PlugIn::setCallback().
    * A PlugInCallback object consists of a set of virtual functions that can be overloaded by an application specific
    * PlugInCallback object.
    * It has two generic callbacks \a onError and \a onWarning, that can be called with any error/warning code.
    * And it has two families of specific error and warning functions that are called on specific error/warning
    * conditions.
    */
  class PlugInCallback : public RCObject
  {
    public:
      //! Enumeration of PlugInCallback errors.
      typedef enum
      {
        PICE_UNEXPECTED_EOF,
        PICE_UNEXPECTED_TOKEN,
        PICE_UNKNOWN_TOKEN,
        PICE_FILE_ACCESS_FAILED,
        PICE_FILE_MAPPING_FAILED,
        PICE_INCOMPATIBLE_FILE,
        PICE_INVALID_FILE
      } PIC_ERROR;

      //! Enumeration of PlugInCallback warnings.
      typedef enum
      {
        PICW_FILE_EMPTY,
        PICW_FILE_NOT_FOUND,
        PICW_FILES_NOT_FOUND,
        PICW_EMPTY_TOKEN,
        PICW_INCOMPATIBLE_VALUES,
        PICW_INVALID_VALUE,
        PICW_UNDEFINED_TOKEN,
        PICW_UNSUPPORTED_TOKEN
      } PIC_WARNING;

      //! Enumeration of value types used in warnings/errors.
      typedef enum
      {
        PICT_INT,
        PICT_FLOAT
      } PIC_TYPE_ID;

      //! Information structure used for warning PICW_EMPTY_TOKEN
      typedef struct
      {
        size_t  position;           //!<  position in file, where the warning was raised
        std::string  context;            //!<  context of the warning
        std::string  token;              //!<  name of empty token
      } EmptyTokenInfo;

      //! Information structure used for warning PICW_INCOMPATIBLE_VALUES
      typedef struct
      {
        size_t        position;     //!<  position in file, where the warning was raised
        std::string        context;      //!<  context of the warning
        PIC_TYPE_ID   valueType;    //!<  type of the incompatible values
        std::string        value0Name;   //!<  name of the first value
        const void *  value0;       //!<  pointer to the first value
        std::string        value1Name;   //!<  name of the second value
        const void *  value1;       //!<  pointer to the second value
      } IncompatibleValueInfo;

      //! Information structure used for warning PICW_INVALID_VALUE
      typedef struct
      {
        size_t        position;     //!<  position in file, where the warning was raised
        std::string        context;      //!<  context of the warning
        PIC_TYPE_ID   valueType;    //!<  type of the invalid value
        std::string        valueName;    //!<  name of the value
        const void *  value;        //!<  pointer to the value
      } InvalidValueInfo;

      //! Information structure used for warning PICW_UNDEFINED_TOKEN
      typedef struct
      {
        size_t        position;     //!<  position in file, where the warning was raised
        std::string        context;      //!<  context of the undefined token
        std::string        token;        //!<  undefined token
      } UndefinedTokenInfo;

      //! Information structure used for error PICE_UNEXPECTED_TOKEN 
      typedef struct
      {
        size_t        position;     //!<  position in file, where the error was raised
        std::string        expected;     //!<  expected token
        std::string        encountered;  //!<  encountered token
      } UnexpectedTokenInfo;

      //! Information structure used for error PICE_UNKNOWN_TOKEN
      typedef struct
      {
        size_t        position;     //!<  position in file, where the error was raised
        std::string        context;      //!<  context of the unknown token
        std::string        token;        //!<  unknown token
      } UnknownTokenInfo;

      //! Information structure used for warning PICW_UNSUPPORTED_TOKEN
      typedef struct
      {
        size_t        position;     //!<  position in file, where the warning was raised
        std::string        context;      //!<  context of the unsupported token
        std::string        token;        //!<  unsupported token
      } UnsupportedTokenInfo;

      //! Information structure used for error PICE_FILE_ACCESS_ERROR
      typedef struct
      {
        std::string        file;         //!< Specifies the name of the file for which the access error occurred.
        unsigned int  systemSpecificErrorCode; //!< Specifies a system specific error code. 
      } FileAccessFailedInfo;

      //! Information structure used for error PICE_FILE_MAPPING_ERROR
      typedef struct
      {
        unsigned int  systemSpecificErrorCode; //!< Specifies a system specific error code. 
      } FileMappingFailedInfo;

      //! Information structure used for error PICE_INCOMPATIBLE_FILE
      typedef struct
      {
        std::string        file;         //!< Specifies the name of the file that was detected to be incompatible.
        std::string        context;      //!< Specifies the context of detected incompatibility.
        unsigned int  expectedVersion; //!< Specifies the expected version. The high-order 16-bits specify
                                       //!< the major version; the low-order 16-bits specify the compatibility level.   
        unsigned int  detectedVersion; //!< Specifies the detected version. The high-order 16-bits specify
                                       //!< the major version; the low-order 16-bits specify the compatibility level.   
      } IncompatibleFileInfo;      

      //! Information structure used for error PICE_INVALID_FILE
      typedef struct
      {
        std::string        file;         //!< Specifies the name of the file the was detected to be invalid.
        std::string        context;      //!< Specifies the context of detected invalidity.
      } InvalidFileInfo;

    public:
      //! Constructor
      PlugInCallback();

      //! Set whether an exception should be thrown on error.
      void  setThrowExceptionOnError( bool set );

      //! General callback on error.
      /** This general error callback is called with every error that isn't completely handled in a specialized
        * error callback.
        * If throwExceptionOnError is set (default), onError throws an PIC_ERROR exception.
        * The specific error callbacks fall back to this function.
        * This callback can be overloaded to support new error conditions.  */
      virtual void onError( PIC_ERROR eid       //!<  error ID
                          , const void *info    //!<  pointer to information structure corresponding to \a eid
                          ) const;

      //! General callback on warning.
      /** This general warning callback is called with every warning that isn't completely handled in a specialized
        * warning callback.
        * In this base implemetation, onWarning just returns true, meaning the PlugIn operation can be continued
        * without danger.
        * The specific warning callbacks fall back to this function.
        * This callback can be overloaded to support new warning conditions.  */
      virtual bool onWarning( PIC_WARNING wid   //!<  waringing ID
                            , const void *info  //!<  pointer to information structure corresponding to \a wid
                            ) const;

      //! Specialized callback on error: PICE_UNEXPECTED_EOF
      /** This error callback is called, when the end of a file was reached unexpectedly.
        * In this base implementation, the general callback onError is called.
        * Overload this callback to implement special handling on unexpected end of file errors. */
      virtual void onUnexpectedEndOfFile( size_t position //!<  position in file, where error occured
                                        ) const;

      //! Specialized callback on error: PICE_UNEXPECTED_TOKEN
      /** This error callback is called, when an unexpected token is encountered.
        * In this base implementation, a struct UnexpectedTokenInfo is constructed out of the arguments, and the
        * general callback onError is called.
        * Overlaod this callback to implement special handling on unexpected token errors. */
      virtual void onUnexpectedToken( size_t position           //!<  position in file, where error occured
                                    , const std::string &expected    //!<  expected token
                                    , const std::string &encountered //!<  encountered token
                                    ) const;

      //! Specialized callback on error: PICE_UNKNOWN_TOKEN
      /** This error callback is called, when an unknown token is encountered.
        * In this base implementation, a struct UnknownTokenInfo is constructed out of the arguments, and the
        * general callback onError is called.
        * Overload this callback to implement special handling on unknown token errors. */
      virtual void onUnknownToken( size_t position        //!<  position in file, where the error occured
                                 , const std::string &context  //!<  context of the unknown token
                                 , const std::string &token    //!<  unknown token
                                 ) const;

      //! Specialized callback on warning: PICW_EMPTY_TOKEN
      /** This warning callback is called, when a token proved to be empty.
        * In this base implementation, a struct EmptyTokenInfo is constructed out of the arguments, and the
        * general callback onWarning is called.
        * Overload this callback to implement special handling on empty token warnings. */
      virtual bool onEmptyToken( size_t position        //!<  position in file, where the warning occured
                               , const std::string &context  //!<  context of the empty token
                               , const std::string &token    //!<  name of the empty token
                               ) const;

      //! Specialized callback on warning: PICW_FILE_EMPTY
      /** This warning callback is called, when a file proved to be empty.
        * In this base implementation, the general callback onWarning is called.
        * Overload this callback to implement special handling on empty file warnings.  */
      virtual bool onFileEmpty( const std::string &file  //!<  name of the empty file
                              ) const;

      //! Specialized callback on warning: PICW_FILE_NOT_FOUND
      /** This warning callback is called, when a file wasn't found.
        * In this base implementation, the general callback onWarning is called.
        * Overload this callback to implement special handling on file not found warnings.  */
      virtual bool onFileNotFound( const std::string &file //!<  name of the file, that wasn't found
                                 ) const;

      //! Specialized callback on warning: PICW_FILES_NOT_FOUND
      /** This warning callback is called, when none of a number of files was found.
        * In this base implementation, the general callback onWarning is called.
        * Overload this callback to implement special handling on file not found warnings.  */
      virtual bool onFilesNotFound( const std::vector<std::string> &files //!<  vector of files, where none was found
                                  ) const;

      //! Specialized callback on warning: PICW_UNDEFINED_TOKEN
      /** This warning callback is called, a named token is used that wasn't defined before.
        * In this base implementation, a struct UndefinedTokenInfo is constructed out of the arguments, and the
        * general callback onWarning is called.
        * Overload this callback to implement special handling on undefined token warnings. */
      virtual bool onUndefinedToken( size_t position        //!<  position in file, where the warning occured
                                   , const std::string &context  //!<  context of the undefined token
                                   , const std::string &token    //!<  undefined token
                                   ) const;

      //! Specialized callback on warning: PICW_INCOMPATIBLE_VALUES
      /** This warning callback is called, when two incompatible values of type int are found in the file.
        * In this base implementation, a struct IncompatibleValueInfo is constructed out of the arguments, and the
        * general callback onWarning is called.
        * Overload this callback to implement special handling on incompatible values warnings.  */
      virtual bool onIncompatibleValues( size_t position          //!<  position in file, where the warning occured
                                       , const std::string &context    //!<  context of the incompatible values
                                       , const std::string &value0Name //!<  name of the first value
                                       , int value0               //!<  first value
                                       , const std::string &value1Name //!<  name of the second value
                                       , int value1               //!<  second value
                                       ) const;

      //! Specialized callback on warning: PICW_INCOMPATIBLE_VALUES
      /** This warning callback is called, when two incompatible values of type float are found in the file.
        * In this base implementation, a struct IncompatibleValueInfo is constructed out of the arguments, and the
        * general callback onWarning is called.
        * Overload this callback to implement special handling on incompatible values warnings.  */
      virtual bool onIncompatibleValues( size_t position          //!<  position in file, where the warning occured
                                       , const std::string &context    //!<  context of the incompatible values
                                       , const std::string &value0Name //!<  name of the first value
                                       , float value0             //!<  first value
                                       , const std::string &value1Name //!<  name of the second value
                                       , float value1             //!<  second value
                                       ) const;

      //! Specialized callback on warning: PICW_INVALID_VALUE
      /** This warning callback is called, when an invalid value of type int is found in the file.
        * In this base implementatino, a struct InvalidValueInfo is constructed out of the arguments, and the
        * general callback onWarning is called.
        * Overload this callback to implement special handling on incompatible values warnings.  */
      virtual bool onInvalidValue( size_t position          //!<  position in file, where the warning occured
                                 , const std::string &context    //!<  context of the invalid value
                                 , const std::string &valueName  //!<  name of the invalid value
                                 , int value                //!<  invalid value
                                 ) const;

      //! Specialized callback on warning: PICW_INVALID_VALUE
      /** This warning callback is called, when an invalid value of type float is found in the file.
        * In this base implementatino, a struct InvalidValueInfo is constructed out of the arguments, and the
        * general callback onWarning is called.
        * Overload this callback to implement special handling on incompatible values warnings.  */
      virtual bool onInvalidValue( size_t position          //!<  position in file, where the warning occured
                                 , const std::string &context    //!<  context of the invalid value
                                 , const std::string &valueName  //!<  name of the invalid value
                                 , float value              //!<  invalid value
                                 ) const;

      //! Specialized callback on warning: PICW_UNSUPPORTED_TOKEN
      /** This warning callback is called, when an unsupported token is found in the file.
        * In this base implementatino, a struct UnsupportedTokenInfo is constructed out of the arguments, and the
        * general callback onWarning is called.
        * Overload this callback to implement special handling on unsupported token warnings.  */
      virtual bool onUnsupportedToken( size_t position        //!<  position in file, where the warning occured
                                     , const std::string &context  //!<  context of the unsupported token
                                     , const std::string &token    //!<  unsupported token
                                     ) const;

      //! Error callback on PICE_FILE_ACCESS_FAILED error
      /** This callback should be invoked from within a plug-in if a file access error occured. 
        * The default implementation simply calls onError.\n
        * It is advisable to overload this function if more customized error handling is requested
        * for file access errors. */
      virtual void onFileAccessFailed( const std::string& file  //!< Specifies the name of the file
                                     , unsigned int systemSpecificErrorCode //!< Specifies a system specific error code.
                                     ) const;

      //! Error callback on PICE_FILE_MAPPING_FAILED error
      /** This callback should be invoked from within a plug-in if a file mapping error occured. 
        * The default implementation simply calls onError.\n
        * It is advisable to overload this function if more customized error handling is requested
        * for file mapping errors. */
      virtual void onFileMappingFailed( unsigned int systemSpecificErrorCode //!< Specifies a system specific error code.
                                      ) const; 

      //! Error callback on PICE_INCOMPATIBLE_FILE error
      /** This callback should be invoked from within a plug-in if a file is detected to be incompatible. 
        * The default implementation simply calls onError.\n
        * It is advisable to overload this function if more customized error handling is requested
        * when detecting incompatible files. */
      virtual void onImcompatibleFile( const std::string& file //!< Specifies the name of the file.
                                     , const std::string& context //!< Specifies the context of detected incompatibility.
                                     , unsigned int expectedVersion //!< Specifies the expected version. The high-order 16-bits specify
                                                                    //!< the major version; the low-order 16-bits specify the compatibility level.
                                     , unsigned int detectedVersion //!< Specifies the detected version. The high-order 16-bits specify
                                                                    //!< the major version; the low-order 16-bits specify the compatibility level.
                                     ) const; 

      //! Error callback on PICE_INVALID_FILE error
      /** This callback should be invoked from within a plug-in if a file is detected to be invalid. 
        * The default implementation simply calls onError.\n
        * It is advisable to overload this function if more customized error handling is requested
        * when detecting invalid files. */
      virtual void onInvalidFile( const std::string& file //!< Specifies the name of the file.
                                , const std::string& context //!< Specifies the context of detected invalidity.
                                ) const;

    protected: 
      //! Protected destructor to prevent explicit creation on stack.
      virtual ~PlugInCallback();

    private:
      bool  m_throwExceptionOnError;
  };

  inline  PlugInCallback::PlugInCallback()
    : m_throwExceptionOnError(true)
  {
  }

  inline  PlugInCallback::~PlugInCallback()
  {
  }

  inline  void  PlugInCallback::setThrowExceptionOnError( bool set )
  {
    m_throwExceptionOnError = set;
  }

  inline  void  PlugInCallback::onError( PIC_ERROR eid, const void *info ) const
  {
    if ( m_throwExceptionOnError )
    {
      throw( eid );
    }
  }

  inline  bool  PlugInCallback::onWarning( PIC_WARNING wid, const void *info ) const
  {
    return( true );
  }


  inline  void  PlugInCallback::onUnexpectedEndOfFile( size_t position ) const
  {
    onError( PICE_UNEXPECTED_EOF, &position );
  }

  inline  void  PlugInCallback::onUnexpectedToken( size_t position, const std::string &expected, const std::string &encountered ) const
  {
    UnexpectedTokenInfo uti;
    uti.position    = position;
    uti.expected    = expected;
    uti.encountered = encountered;
    onError( PICE_UNEXPECTED_TOKEN, &uti );
  }

  inline  void  PlugInCallback::onUnknownToken( size_t position, const std::string &context, const std::string &token ) const
  {
    UnknownTokenInfo  uti;
    uti.position  = position;
    uti.context   = context;
    uti.token     = token;
    onError( PICE_UNKNOWN_TOKEN, &uti );
  }


  inline  bool  PlugInCallback::onEmptyToken( size_t position, const std::string &context, const std::string &token ) const
  {
    EmptyTokenInfo  uti;
    uti.position  = position;
    uti.context   = context;
    uti.token     = token;
    return( onWarning( PICW_EMPTY_TOKEN, &uti ) );
  }

  inline  bool  PlugInCallback::onFileEmpty( const std::string &file ) const
  {
    return( onWarning( PICW_FILE_EMPTY, &file ) );
  }

  inline  bool  PlugInCallback::onFileNotFound( const std::string &file ) const
  {
    return( onWarning( PICW_FILE_NOT_FOUND, &file ) );
  }

  inline  bool  PlugInCallback::onFilesNotFound( const std::vector<std::string> &files ) const
  {
    return( onWarning( PICW_FILES_NOT_FOUND, &files ) );
  }

  inline  bool  PlugInCallback::onUndefinedToken( size_t position, const std::string &context, const std::string &token ) const
  {
    EmptyTokenInfo  uti;
    uti.position  = position;
    uti.context   = context;
    uti.token     = token;
    return( onWarning( PICW_UNDEFINED_TOKEN, &uti ) );
  }

  inline  bool  PlugInCallback::onIncompatibleValues( size_t position, const std::string &context, const std::string &value0Name,
                                                      int value0, const std::string &value1Name, int value1 ) const
  {
    IncompatibleValueInfo ivi;
    ivi.position    = position;
    ivi.context     = context;
    ivi.valueType   = PICT_INT;
    ivi.value0Name  = value0Name;
    ivi.value0      = &value0;
    ivi.value1Name  = value1Name;
    ivi.value1      = &value1;
    return( onWarning( PICW_INCOMPATIBLE_VALUES, &ivi ) );
  }

  inline  bool  PlugInCallback::onIncompatibleValues( size_t position, const std::string &context, const std::string &value0Name,
                                                      float value0, const std::string &value1Name, float value1 ) const
  {
    IncompatibleValueInfo ivi;
    ivi.position    = position;
    ivi.context     = context;
    ivi.valueType   = PICT_FLOAT;
    ivi.value0Name  = value0Name;
    ivi.value0      = &value0;
    ivi.value1Name  = value1Name;
    ivi.value1      = &value1;
    return( onWarning( PICW_INCOMPATIBLE_VALUES, &ivi ) );
  }


  inline  bool  PlugInCallback::onInvalidValue( size_t position, const std::string &context,
                                                const std::string &valueName, int value ) const
  {
    InvalidValueInfo  ivi;
    ivi.position  = position;
    ivi.context   = context;
    ivi.valueName = valueName;
    ivi.valueType = PICT_INT;
    ivi.value     = &value;
    return( onWarning( PICW_INVALID_VALUE, &ivi ) );
  }

  inline  bool  PlugInCallback::onInvalidValue( size_t position, const std::string &context,
                                                const std::string &valueName, float value ) const
  {
    InvalidValueInfo  ivi;
    ivi.position  = position;
    ivi.context   = context;
    ivi.valueName = valueName;
    ivi.valueType = PICT_FLOAT;
    ivi.value     = &value;
    return( onWarning( PICW_INVALID_VALUE, &ivi ) );
  }

  inline  bool  PlugInCallback::onUnsupportedToken( size_t position, const std::string &context, const std::string &token ) const
  {
    UnsupportedTokenInfo  uti;
    uti.position  = position;
    uti.context   = context;
    uti.token     = token;
    return( onWarning( PICW_UNSUPPORTED_TOKEN, &uti ) );
  }

  inline  void  PlugInCallback::onFileAccessFailed( const std::string& file, unsigned int systemSpecificErrorCode ) const
  {
    FileAccessFailedInfo fafi;
    fafi.file = file;
    fafi.systemSpecificErrorCode = systemSpecificErrorCode;
    onError(PICE_FILE_ACCESS_FAILED, &fafi);
  }

  inline  void  PlugInCallback::onFileMappingFailed( unsigned int systemSpecificErrorCode ) const
  {
    FileMappingFailedInfo fmfi;
    fmfi.systemSpecificErrorCode = systemSpecificErrorCode;
    onError(PICE_FILE_MAPPING_FAILED, &fmfi);
  }

  inline  void  PlugInCallback::onImcompatibleFile( const std::string& file, const std::string& context
                                                  , unsigned int expectedVersion, unsigned int detectedVersion) const
  {
    IncompatibleFileInfo ifi;
    ifi.file = file;
    ifi.context = context;
    ifi.expectedVersion = expectedVersion;
    ifi.detectedVersion = detectedVersion;
    onError(PICE_INCOMPATIBLE_FILE, &ifi);
  }

  inline  void  PlugInCallback::onInvalidFile( const std::string& file, const std::string& context ) const
  {
    InvalidFileInfo ifi;
    ifi.file = file;
    ifi.context = context;
    onError(PICE_INVALID_FILE, &ifi);
  }
}
