// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#if defined(_WIN32)
# include "nvsgwindefs.h"
# ifndef NOMINMAX
#  define NOMINMAX // avoid problems with Microsoft definitions in windef.h and min/max from stl
#  include <windows.h>
#  include <winbase.h>
#  undef NOMINMAX
# else
#  include <windows.h>
#  include <winbase.h>
# endif
#endif

#if defined(LINUX)
# include <pthread.h>  
#endif

namespace nvutil {

#if !defined(DOXYGEN_IGNORE)
// Single write, multiple read synchronization interface.
class ISWMRSync
{
public:
  NVSG_API virtual bool waitToWrite(size_t milliseconds = 0xFFFFFFFF) const = 0;
  NVSG_API virtual void doneWriting()                                 const = 0;
  NVSG_API virtual bool waitToRead(size_t milliseconds = 0xFFFFFFFF)  const = 0;
  NVSG_API virtual void doneReading()                                 const = 0;
};
#endif

//! Single write, multiple read synchronization class.
/** The SWMRSync class is based on the ideas from Jeffrey Richter "Advanced Windows" \n
  * We use three handles to kernel synchronize our objects:\par
  * \b m_hMutexNoWriter is used to indicate whether a writer thread has access to the data. \n
  * When the mutex is signaled, we know that no writer thread is currently \n 
  * accessing the data. \par
  * \b m_hEventNoReaders indicates whether any reader threads are currently accessing the data. \n
  * If no reader threads are accessing the data, this manual-reset event \n
  * object is signaled. \par
  * \b m_hSemNumReaders indicates the number of reader threads that are currently accessing the data. \n
  * This semaphore is only used as a counter. If the counter goes to 0 \n
  * m_hEventMoReaders event can be signaled. */  
class SWMRSync
{
public:
#if !defined(DOXYGEN_IGNORE)
  // Enables locking. This should be called once at initialization time of the nvsg library.
  // Sell also nvsg::nvsgInitialize
  static void enableLocking();
#endif
  
  //! Default constructor.
  SWMRSync(void);
  //! Default destructor.
  virtual ~SWMRSync(void);

  //! Init sync mechanism.
  /** \return true - init went fine, false - init went wrong*/
  NVSG_API bool initialize(const char* name = NULL   //!< The name of the object. Pass NULL if you 
                                            //!<  do not want to share the object (default).
                          ); 

  //! Wait for write access available.
  /** This function checks the current state of the internal write-synchronization object.
    * If that object's state is nonsignaled, the calling thread enters the wait state. 
    * It uses no processor time while waiting for the object state to become signaled or 
    * the time-out interval to elapse. The function returns, if either the write-synchronization
    * object is signaled or the time-out interval elapses. 
    * \remarks If the write-synchronization object was signaled, the function changes the state 
    * of both, the internal write-synchronization object and the internal read-synchronization object 
    * to nonsignaled before it returns. The state of both objects remains nonsignaled until the calling 
    * thread signals write/read access available through a call to doneWriting. 
    * \returns \c true if the calling thread could gain write access, \c false otherwise. */
  NVSG_API bool waitToWrite( size_t milliseconds //!< Time-out interval in milliseconds.
                           , size_t& status //!< currently ignored
                           ) const;
  
  //! Signal write/read access available again.
  /** This function signals both, the internal write-synchronisation object and the internal 
    * read-synchronization object, so another thread currently waiting for either write or read 
    * access, can be released now. */
  NVSG_API void doneWriting() const;

  //! Wait for read access available.
  /** This function checks the current state of the internal read-synchronization object.
    * If that object's state is nonsignaled, the calling thread enters the wait state. 
    * It uses no processor time while waiting for the object state to become signaled or 
    * the time-out interval to elapse. The function returns, if either the read-synchronization
    * object is signaled or the time-out interval elapses. 
    * \remarks If the read-synchronization object was signaled, the function changes the state 
    * of the internal write-synchronization object to nonsignaled before it returns, so that threads
    * that subsequently begin waiting for write access, will enter the wait state. On the other hand,
    * the state of the read-synchronization object will not be changed, so if the read-synchronization 
    * object was signaled, it remains signaled and threads that subsequently begin waiting for read
    * access, immediately will be released. If the calling thread could gain read access, the 
    * syncronization object recognizes one more reader. The state of the write-synchronization object
    * will not earlier change to signaled until all readers finished reading through a call to
    * doneReading.
    * \returns \c true if the calling thread could gain write access, \c false otherwise. */
  NVSG_API bool waitToRead( size_t milliseconds //!< Time-out interval in millsecond
                          , size_t& status //!< currently ignored
                          ) const;
  
  //! Signal one less reading.
  /** Through a call to this function, the synchronization object recognizes one less thread reading.
    * If the number of threads currently reading reaches zero, the function will change the state of the
    * internal write-synchronization object to signaled, so a thread that is currently waiting for write
    * access can be released now. */
  NVSG_API void  doneReading() const;

private:

#if defined(MULTI_PROCESS_AWARE)
#if defined(_WIN32)
  /* This mutex guards access to the other objects
   * managed by this data structure and also indicates 
   * whether any writer threads are writing. */
  HANDLE m_hMutexNoWriter;

  /* This manual-reset event is signaled when
   * no reader threads are reading. */
  HANDLE m_hEventNoReaders;

  /* This semaphore is used simply as a counter that is
   * accessible between multiple processes. It is NOT
   * used for thread synchronization.
   * The count is the number of reader threads reading. */
  HANDLE m_hSemNumReaders;

  /* Helper method to create a name for the semaphores and 
   * mutexes, so we can use them over process boundaries */
  LPCTSTR constructObjName( LPCTSTR lpszPrefix, LPCTSTR lpszSuffix, LPTSTR lpszFullName, size_t cbFullName, bool &bOk) ;
#endif // _WIN32
#else // MULTI_PROCESS_AWARE

  void enterCriticalSection() const;
  void leaveCriticalSection() const;
  void yield() const;

#if defined(_WIN32)
  typedef CRITICAL_SECTION SWMRCriticalSection;
#elif defined(LINUX)
  typedef pthread_mutex_t  SWMRCriticalSection;
#endif

  mutable SWMRCriticalSection m_cs; // thread synchronization object
  mutable int m_sharedLockCount; // accessed only when inside critical section
#endif

  static bool m_lockingEnabled; // set via enableLocking; false by default
};

}

