// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#include "QtRAWnd.h"
#include "nvsg/Camera.h"
#include "nvtraverser/GLTraverser.h"

// enable memory leak detection
#include "nvutil/DbgNew.h"

using namespace nvui;
using namespace nvgl;
using nvsg::Camera;
using nvtraverser::GLTraverser;

QtRAWnd::QtRAWnd(void)
: QGLWidget( QGL::DoubleBuffer | QGL::Rgba | QGL::DepthBuffer )
, m_hglrc(NULL)
, m_dpy(NULL)
, m_drawable(NULL)
{
  // do not clear the background before calling the paint event handler
  QWidget::setBackgroundMode(Qt::NoBackground);
}

QtRAWnd::~QtRAWnd(void)
{
}


void QtRAWnd::initializeGL()
{
  GLXContext ctx = m_hglrc = GLXGetCurrentContext();
  GLXDrawable drw = m_drawable = GLXGetCurrentDrawable();
  Display* dpy = m_dpy = glXGetCurrentDisplay(); 

  // introduce the GL render context to nvsg!
  GLXAttachContext(m_dpy, m_hglrc);   // this must be called if you do not create the OpenGL context through our GLXCreateContext() wrapper!!
                                    // (e.g. when a GUI library calls the regular glXCreateContext() in its implementation)

  // must be current for the creation of the render traverser
  // (see remarks in makeCurrent() routine!)
  makeCurrent();

  if ( RenderArea::init())
  {         
    // create traverser render traverser
    __ASSERT(m_renderTraverser == NULL);
    __VERIFY(m_renderTraverser = new GLTraverser);   // have to create a GLTraverser
    m_renderTraverser->addRef();
  }          

}

void QtRAWnd::resizeGL( int w, int h )
{
  RenderArea::setViewportSize( w, h );

  // must be current now before doing any GL calls or render traversal through NVSG
  // (see remarks in makeCurrent() routine!)  
  makeCurrent();
  GLViewport( 0, 0, (GLsizei)w, (GLsizei)h );
}

void QtRAWnd::paintGL()
{    
  // must be current now before doing any GL calls or render traversal through NVSG
  // (see remarks in makeCurrent() routine!)
  makeCurrent();

  if ( m_scene ) 
  {
    GLClearColor(m_scene->getBackColor()[0], m_scene->getBackColor()[1], m_scene->getBackColor()[2], 0.0f);
  }
  GLClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  
  // Then initiate cull/app/render traversal through base class.
  RenderArea::renderScene();

}

bool QtRAWnd::init( RenderArea * shareArea )
{    
  // Attach the CWnd's window handle to the render area and initialize OpenGL and traversers.
  __ASSERT(winId());
  return true;
}

void QtRAWnd::destroy()
{  
  // Call base class before we destroy the window for OpenGL and traverser cleanup.
  RenderArea::destroy();  

  if (m_renderTraverser)
  {
    m_renderTraverser->removeRef();
    m_renderTraverser = NULL;
  }

  GLXDetachContext(m_hglrc);   // this must be called if you do not destroy the OpenGL context through our GLXDestroyContext() wrapper!!                             // (e.g. when a GUI library calls the regular glXDestroyContext() in its implementation)

  QWidget::destroy();
}

void QtRAWnd::triggerRedraw()
{
  // call base class
  RenderArea::triggerRedraw();

  updateGL();  // schedules a repaint
}

void QtRAWnd::mousePressEvent( QMouseEvent * event)
{
  switch(event->button())
  {
    case Qt::LeftButton:
      mouseButtonDown(CameraManipulator::_LEFT_MOUSE_BTN, event->x(), event->y());
      break;

    case Qt::MidButton:
      mouseButtonDown(CameraManipulator::_MIDDLE_MOUSE_BTN, event->x(), event->y());
      break;

    case Qt::RightButton: 
      if(m_viewState->getCamera())
      {
        Camera* pNCCamera = beginEdit(m_camManip->getAttachedCamera());
        pNCCamera->zoom(m_scene->getRootNode()->getBoundingSphere());
        triggerRedraw();
        endEdit(pNCCamera);
      }
      break;
  }

  QWidget::mousePressEvent(event);
}

void QtRAWnd::mouseReleaseEvent( QMouseEvent * event)
{
  switch(event->button())
  {
    case Qt::LeftButton:
      mouseButtonUp(CameraManipulator::_LEFT_MOUSE_BTN, event->x(), event->y());
      break;

    case Qt::MidButton:
      mouseButtonUp(CameraManipulator::_MIDDLE_MOUSE_BTN, event->x(), event->y());
      break;
  }

  QWidget::mouseReleaseEvent(event);
}

void QtRAWnd::mouseMoveEvent( QMouseEvent * event)
{
  mouseMotion( event->x(), event->y());
  
  if( m_camManip->getButtonState() != CameraManipulator::_NO_BTN_KEY
      || m_camManip->getKeyState() != CameraManipulator::_NO_BTN_KEY)
  {
    triggerRedraw();
  }
  

  QWidget::mouseMoveEvent(event);
}
  
