#include "./SimpleScene.h"

#include "nvmath/nvmath.h"
#include "nvmath/Vec3f.h"
#include "nvsg/Transform.h"
#include "nvsg/Node.h"
#include "nvsg/Face.h"
#include "nvsg/Material.h"
#include "nvsg/Triangles.h"
#include "nvsg/GeoNode.h"
#include "nvsg/FaceAttribute.h"

#include "nvutil/DbgNew.h" // this must be the last include

using namespace nvsg;
using namespace nvmath;

SimpleScene::SimpleScene(void)
{  
  setBackColor(Vec3f(1,0,0)); 

  // Create the root node and apply a rotation.
  // Note: 
  // We do not increment the ref count here, 
  // because we throw the root into the scene 
  // via setRoot(), and setRoot() will increment 
  // it for us.
  Transform * pRoot = beginEdit( Transform::create() );
  pRoot->setOrientation( Quatf(Vec3f(0.0f, 1.0f, 0.0f ), float(PI_QUARTER)) );
  // pRoot->setTranslation( Vec3f( 0.0f, 0.0f, 0.0f ) ); // obsolete here, because 0,0,0 is default for translation

  // positioning the shapes 
  Transform * pShapeTrafo = beginEdit( Transform::create() );
  pShapeTrafo->setOrientation( Quatf(Vec3f(1.0f, 0.0f, 0.0f ), float(PI_QUARTER)) );
  // pShapeTrafo->setTranslation( Vec3f( 0.0f, 0.0f, 0.0f ) ); // obsolete here, because 0,0,0 is default for translation

  Transform * pShapeTrafo2 = beginEdit( Transform::create() );
  pShapeTrafo2->setOrientation( Quatf(Vec3f(0.0f, 1.0f, 0.0f ), float(PI_QUARTER)) );
  // pShapeTrafo2->setTranslation( Vec3f( 0.0f, 0.0f, 0.0f ) ); // obsolete here, because 0,0,0 is default for translation

  Transform * pShape2Trafo = beginEdit( Transform::create() );
  pShape2Trafo->setOrientation( Quatf(Vec3f(1.0f, 0.0f, 0.0f ), 0.0f) );
  pShape2Trafo->setTranslation( Vec3f( -1.0f, 0.0f, 0.0f ) ); 

  Transform * pShape2Trafo2 = beginEdit( Transform::create() );
  pShape2Trafo2->setOrientation( Quatf(Vec3f(0.0f, 1.0f, 0.0f ), float(PI_QUARTER)) );
  // pShape2Trafo2->setTranslation( Vec3f( 0.0f, 0.0f, 0.0f ) ); // obsolete here, because 0,0,0 is default for translation

  Transform * pShape3Trafo = beginEdit( Transform::create() );
  pShape3Trafo->setOrientation( Quatf(Vec3f(1.0f, 0.0f, 0.0f ), 0.0f) );
  pShape3Trafo->setTranslation( Vec3f( 1.0f, 0.0f, 0.0f ) ); 

  // setup vertices and faces  
  Vec3f vertices[8] = { Vec3f(-0.2f, -0.2f,  0.2f)
    , Vec3f(-0.2f,  0.2f,  0.2f)
    , Vec3f( 0.2f,  0.2f,  0.2f)
    , Vec3f( 0.2f, -0.2f,  0.2f)
    , Vec3f( 0.2f, -0.2f, -0.2f)
    , Vec3f( 0.2f,  0.2f, -0.2f)
    , Vec3f(-0.2f,  0.2f, -0.2f)
    , Vec3f(-0.2f, -0.2f, -0.2f) };

  Face3 faces[12]     = { {0,2,1}, {2,0,3}
  , {3,5,2}, {5,3,4}
  , {4,6,5}, {6,4,7}
  , {7,1,6}, {1,7,0}
  , {0,4,3}, {4,0,7}
  , {1,5,6}, {5,1,2} };

  // calculating face normals
  Vec3f v[36];
  Vec3f n[36];
  Face3 f[12];
  Vec3f v0, v1, v2, d0, d1, fn;
  for ( int kf=0, kv=0; kf<12; kf++, kv+=3 )
  {
    v0 = vertices[faces[kf][0]];
    v1 = vertices[faces[kf][1]];
    v2 = vertices[faces[kf][2]];

    d0 = v1 - v0;
    d1 = v2 - v0;

    fn = d0^d1;
    fn.normalize();

    f[kf][0] = kv;
    f[kf][1] = kv+1;
    f[kf][2] = kv+2;

    v[kv]    = v0;
    v[kv+1]  = v1;
    v[kv+2]  = v2;

    n[kv]    = fn; 
    n[kv+1]  = fn; 
    n[kv+2]  = fn; 
  }

  // Create a GeoSet, in this case triangles.
  Triangles * pShape = beginEdit( Triangles::create() );
  pShape->setVertices(v, 36);
  pShape->setNormals(n,36);
  pShape->setFaces(f, 12);

  // Create a material.
  Material * pMaterial = beginEdit( Material::create() );
  pMaterial->setAmbientColor( Vec3f(0.f, 0.f, 1.f) );
  pMaterial->setDiffuseColor( Vec3f(0.4f, 0.4f, 0.4f) );
  pMaterial->setEmissiveColor( Vec3f(0.f, 0.f, 0.f) );
  pMaterial->setSpecularColor( Vec3f(1.f, 1.f, 1.f) );
  pMaterial->setSpecularExponent( 10.f );
  pMaterial->setOpacity( 1.0f );

  // Create a face attribute. The face attribute is a state attribute 
  // that determines how the following faces will be rendered.
  const FaceAttribute * pFaceAttribute = FaceAttribute::create();

  // Create state set. Add material and render mode.
  StateSet * pStateSet = beginEdit( StateSet::create() );
  pStateSet->addAttribute( endEdit( pMaterial ) );
  pStateSet->addAttribute( pFaceAttribute );

  // Setup the whole scene graph
  GeoNode * pGeoNode = beginEdit( GeoNode::create() );
  pGeoNode->addGeometry( endEdit( pShape ), endEdit( pStateSet ) );

  const GeoNode * pGeoNode2 = pGeoNode->clone();

  // Note: 
  // AddChild() will increment the reference count 
  // of the added node.
  pShapeTrafo2->addChild( endEdit( pGeoNode ) );
  pShapeTrafo->addChild( endEdit( pShapeTrafo2 ) );
  pShape2Trafo2->addChild( pGeoNode2 );
  pShape2Trafo->addChild( endEdit( pShape2Trafo2 ) );
  pShape3Trafo->addChild( pGeoNode2 );
  pRoot->addChild( endEdit( pShapeTrafo ) );
  pRoot->addChild( endEdit( pShape2Trafo ) );
  pRoot->addChild( endEdit( pShape3Trafo ) );

  setRootNode( endEdit( pRoot ) );
}

const SimpleScene * SimpleScene::create( void )
{
  return( new SimpleScene );
}
